#include <linux/fs.h>
#include "f2dfs_fs.h"
#include <crypto/hash.h>
#include <crypto/skcipher.h>
#include <linux/crypto.h>
#include <linux/workqueue.h>
#include <linux/radix-tree.h>
#include <linux/rcupdate.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/freezer.h>
#include <linux/sched/signal.h>

#include "f2fs.h"
#include "segment.h"


// Structure to define basic hash information for the SHA-1 algorithm
struct f2dfs_basic_hash_info sha1_info = {
	.name =			"sha1",
	.simple_name =	"SHA-1",
	.full_name =	"Secure Hash Algorithm",
	.len =			FP_LEN_MAX,            // Fingerprint length (20 bytes)
	.zero_prefix =	0x1cea,
	.zero_result =	"\x1c\xea\xf7\x3d"     // Predefined zero hash result
					"\xf4\x0e\x53\x1d"
					"\xf3\xbf\xb2\x6b"
					"\x4f\xb7\xcd\x95"
					"\xfb\x7b\xff\x1d",
	.null_result =	"\x00\x00\x00\x00"     // Null hash result (all zeros)
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00",
	.fp_update =	crypto_shash_update,   // Function to update hash computation
};

// Structure to define basic hash information for the SHA-256 algorithm
struct f2dfs_basic_hash_info sha256_info = {
	.name =			"sha256",
	.simple_name =	"SHA-2", 
	.full_name =	"Secure Hash Algorithm",
	.len =			FP_LEN_MAX,            // Fingerprint length (32 bytes)
	.zero_prefix =	0xad7f,               
	.zero_result =	"\xad\x7f\xac\xb2"     // Predefined zero hash result
					"\x58\x6f\xc6\xe9"
					"\x66\xc0\x04\xd7"
					"\xd1\xd1\x6b\x02"
					"\x4f\x58\x05\xff"
					"\x7c\xb4\x7c\x7a"
					"\x85\xda\xbd\x8b"
					"\x48\x89\x2c\xa7",
	.null_result =	"\x00\x00\x00\x00"     // Null hash result (all zeros)
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00"
					"\x00\x00\x00\x00",
	.fp_update =	crypto_shash_update,   // Function to update hash computation
};

// Initializes a descriptor for hash computation.
// Allocates memory for shash descriptor and associates it with the given algorithm.
static inline struct sdesc *init_sdesc(struct crypto_shash *alg)
{
	struct sdesc *sdesc;
	int size;

	// Calculate required memory size for the descriptor
	size = sizeof(struct shash_desc) + crypto_shash_descsize(alg);
	sdesc = kmalloc(size, GFP_KERNEL);

	// Return error if memory allocation fails
	if (!sdesc)
		return ERR_PTR(-ENOMEM);

	sdesc->shash.tfm = alg; // Assign the algorithm to the descriptor
	return sdesc;
}

// Computes the fingerprint (hash) for a page using the specified hash algorithm.
// Outputs the result in the digest buffer.
static inline int f2dfs_compute_fingerprint(struct f2fs_sb_info *sbi,
				__u8 digest[], struct page *page, int page_len,
				struct sdesc *sdesc)
{
	crypto_shash_init(&sdesc->shash); // Initialize the hash computation
	sbi->s_fingerprint_hash->basic_hash_info->fp_update(
					&sdesc->shash, page_address(page), page_len); // Update hash with page data
	crypto_shash_final(&sdesc->shash, digest); // Finalize and store the hash result
	return FP_HASH_SUCCESS;
}

// Computes the fingerprint for a single page.
// A wrapper around f2dfs_compute_fingerprint with a fixed page length.
static inline int do_fingerprint_of_page(struct f2fs_sb_info *sbi,
				struct page *page, struct sdesc *sdesc, __u8 digest[])
{
	return f2dfs_compute_fingerprint(sbi, digest, page, PAGE_SIZE, sdesc);
}

// Allocates and initializes the necessary hash resources dynamically.
int f2dfs_page_fingerprint_update(struct f2fs_sb_info *sbi,
				struct page *page, __u8 digest[])
{
	struct crypto_shash *alg; // Hash algorithm instance
	struct sdesc *sdesc;      // Descriptor for hash computation
	int ret;

	// Handle invalid page input
	if ((!page) || IS_ERR(page)) {
		memcpy(digest, sbi->s_fingerprint_hash->basic_hash_info->zero_result,
						FP_LEN_MAX); // Assign zero result to digest
		return FP_HASH_FAILURE;
	}

	// Allocate hash algorithm
	alg = crypto_alloc_shash(sbi->s_fingerprint_hash->basic_hash_info->name,
					0, 0);
	if (IS_ERR(alg)) {
		crypto_free_shash(alg); // Free algorithm if allocation fails
		memcpy(digest, sbi->s_fingerprint_hash->basic_hash_info->zero_result,
						FP_LEN_MAX); // Assign zero result to digest
		return FP_HASH_FAILURE; // Return failure status
	}

	// Initialize hash descriptor
	sdesc = init_sdesc(alg);
	if (IS_ERR(sdesc)) {
		kfree(sdesc); // Free descriptor memory
		crypto_free_shash(alg); // Free algorithm
		memcpy(digest, sbi->s_fingerprint_hash->basic_hash_info->zero_result,
						FP_LEN_MAX); // Assign zero result to digest
		return FP_HASH_FAILURE;
	}

	// Compute the fingerprint
	ret = do_fingerprint_of_page(sbi, page, sdesc, digest);
	kfree(sdesc);         // Free descriptor memory
	crypto_free_shash(alg); // Free algorithm instance
	return ret;
}

// Extracts the fingerprint prefix from a given fingerprint array.
unsigned int f2dfs_get_fingerprint_prefix(__u8 fp[])
{
	unsigned int sum = 0;
	int bytes, i;

	bytes = FP_PREFIX_LENGTH_BYTE;
	for (i = 0; i < bytes; ++i) {
		sum = sum << 8;
		sum += le32_to_cpu(fp[i]);
	}

	sum &= FP_PREFIX_MASK_BIT;
	sum >>= ((FP_PREFIX_LENGTH_BYTE << 3) - FP_PREFIX_LENGTH_BIT);
	return sum;
}

// Compares two fingerprints for equality.
int f2dfs_fp_is_equal(__u8 fp1[], __u8 fp2[])
{
	return memcmp(fp1, fp2, FP_LEN_MAX);
}

// Checks if a fingerprint is the zero fingerprint based on its prefix.
bool f2dfs_fp_is_zero(struct f2fs_sb_info *sbi, __u8 fp[], unsigned int prefix)
{
	if (prefix != ZERO_FP_PREFIX)
		return false;

	return (f2dfs_fp_is_equal(fp,
					sbi->s_fingerprint_hash->basic_hash_info->zero_result)
					== 0);
}

// Checks if a fingerprint is the null fingerprint.
bool f2dfs_fp_is_null(struct f2fs_sb_info *sbi, __u8 fp[])
{
	return (f2dfs_fp_is_equal(fp,
					sbi->s_fingerprint_hash->basic_hash_info->null_result)
					== 0);
}

// Checks if a fingerprint is marked for removal.
bool f2dfs_fp_is_remove(__u8 fp[])
{
	__u8 fp_remove[FP_LEN_MAX];

	memset(fp_remove, 0, FP_LEN_MAX);
	memcpy(fp_remove, fp, FP_PREFIX_LENGTH_BYTE);
	fp_remove[FP_PREFIX_LENGTH_BYTE - 1] &= FP_PREFIX_MASK_BIT_PART;

	return (f2dfs_fp_is_equal(fp, fp_remove) == 0);
}

// Constructs a "remove" fingerprint based on a prefix.
void f2dfs_get_fp_remove(__u8 fp_remove[], unsigned int fp_prefix)
{
	int i = 0;

	fp_prefix <<= ((FP_PREFIX_LENGTH_BYTE << 3) - FP_PREFIX_LENGTH_BIT);
	memset(fp_remove, 0, FP_LEN_MAX);

	for (i = 0; i < FP_PREFIX_LENGTH_BYTE; ++i)
		fp_remove[FP_PREFIX_LENGTH_BYTE - i - 1] =
						(unsigned char)(((fp_prefix) >> (i * 8)) & 0x00ff);
}

// Retrieves the pointer field from the fingerprint footer of a page.
nid_t get_fpfooter_pointer(struct page *p)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return le32_to_cpu(rn->fn.fp_footer.fp_pointer);
}

// Retrieves specific tags from the fingerprint footer of a page.
bool get_fpfooter_tags(struct page *p, int type)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return (le32_to_cpu(rn->fn.fp_footer.tags) & type);
}

// Sets the pointer field in the fingerprint footer of a page.
int set_fpfooter_pointer(struct page *p, nid_t pointer)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.fp_footer.fp_pointer = cpu_to_le32(pointer);
	return set_page_dirty(p);
}

// Sets specific tags in the fingerprint footer of a page.
int set_fpfooter_tags(struct page *p, int type)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.fp_footer.tags |= type;
	return set_page_dirty(p);
}

// Clears specific tags in the fingerprint footer of a page.
int clear_fpfooter_tags(struct page *p, int type)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.fp_footer.tags &= ~type;
	return set_page_dirty(p);
}

// Allocates and initializes a single fingerprint bucket.
struct f2dfs_fp_bucket *f2dfs_alloc_single_bucket(
				struct f2fs_sb_info *sbi, struct slist_head *head)
{
	// Allocate memory for the bucket from the bucket cache.
	struct f2dfs_fp_bucket *bucket = kmem_cache_alloc(
					sbi->s_fingerprint_bucket->bucket_cache, GFP_KERNEL);

	// Initialize bucket fields to default values.
	memset(bucket, 0, sizeof(struct f2dfs_fp_bucket));
	bucket->valid_count = 0;
	bucket->remove_count = 0;
	bucket->atime = jiffies - ((DEF_IDLE_FP_BUCKET_INTERVAL + 1) * HZ);

	// Add the bucket to the list and update the bucket count.
	slist_add(&bucket->b_head, head);
	++sbi->s_fingerprint_bucket->current_number;

	return bucket;
}

// Frees a single fingerprint bucket and updates the bucket count.
static inline void f2dfs_free_single_bucket(
				struct f2fs_sb_info *sbi, struct slist_head *head,
				struct f2dfs_fp_bucket *bucket)
{
	// Remove the bucket from the list and release its memory.
	slist_del(&bucket->b_head, head);
	kmem_cache_free(sbi->s_fingerprint_bucket->bucket_cache, bucket);
	--sbi->s_fingerprint_bucket->current_number;
}

// Allocates and initializes a single fingerprint buffer.
struct f2dfs_fp_buffer *f2dfs_alloc_single_buffer(
				struct f2fs_sb_info *sbi, struct slist_head *head)
{
	// Allocate memory for the buffer from the buffer cache.
	struct f2dfs_fp_buffer *buffer = kmem_cache_alloc(
					sbi->s_fingerprint_buffer->buffer_cache, GFP_KERNEL);

	// Initialize buffer fields to default values.
	memset(buffer, 0, sizeof(struct f2dfs_fp_buffer));
	buffer->valid_count = 0;
	buffer->atime = jiffies - ((DEF_IDLE_FP_BUFFER_INTERVAL + 1) * HZ);

	// Add the buffer to the list and update the buffer count.
	slist_add(&buffer->b_head, head);
	++sbi->s_fingerprint_buffer->current_number;

	return buffer;
}

// Frees a single fingerprint buffer and updates the buffer count.
static inline void f2dfs_free_single_buffer(struct f2fs_sb_info *sbi,
				struct slist_head *head, struct f2dfs_fp_buffer *buffer)
{
	// Remove the buffer from the list and release its memory.
	slist_del(&buffer->b_head, head);
	kmem_cache_free(sbi->s_fingerprint_buffer->buffer_cache, buffer);
	--sbi->s_fingerprint_buffer->current_number;
}

/*
 * There are following five operations that F2DFS offline deduplication thread
 * needs to perform conditionally:
 * 1. Persist buckets cache to unstable lists (single bucket over threshold)
 * 2. Stabilize unstable lists to stable ones (system idle and after 1)
 * 3. Persist buffers cache to journal (sum over four pages)
 * 4. Offline dedupe (system idle and after 1~3)
 * 5. Free buckets and buffers cache (at night or long idle time)
 */

// Frees fingerprint bucket caches in the background thread.
static inline bool f2dfs_free_buckets_cache_by_thread(
				struct f2fs_sb_info *sbi)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	bool ret = false;
	int i;

	// Iterate through all fingerprint bucket lists.
	for (i = 0; i < FP_PREFIX_NUMBER; ++i) {
		head = &sbi->s_fingerprint_bucket->pointers[i];
		if (!head)
			return ret;

		// Traverse the list for each bucket.
		slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[i]) {
			if (!head)
				return ret;

			// Retrieve the current bucket.
			bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
			if (!bucket)
				return ret;

			// Check if the bucket should be freed.
			if (should_free_single_bucket(bucket)) {
				// Persist bucket data if needed.
				if (should_persist_single_bucket_normal(bucket))
					f2dfs_sync_single_bucket_foreground(sbi, bucket, i);
				// Free the bucket and mark operation as performed.
				f2dfs_free_single_bucket(sbi, head, bucket);
				ret = true;
				break;
			}
		}
	}

	return ret;
}

// Frees fingerprint buffer caches in the background thread.
static inline bool f2dfs_free_buffers_cache_by_thread(
				struct f2fs_sb_info *sbi)
{
	struct slist_head *head;
	struct f2dfs_fp_buffer *buffer;
	bool ret = false;
	int i;

	// Iterate through all fingerprint buffer lists.
	for (i = 0; i < FP_BUFFER_NUMBER; ++i) {
		head = &sbi->s_fingerprint_buffer->pointers[i];
		if (!head)
			return ret;

		// Traverse the list for each buffer.
		slist_for_each(head, &sbi->s_fingerprint_buffer->pointers[i]) {
			if (!head)
				return ret;

			// Retrieve the current buffer.
			buffer = slist_entry(head, struct f2dfs_fp_buffer, b_head);
			if (!buffer)
				return ret;

			// Check if the buffer should be freed.
			if (should_free_single_buffer(buffer)) {
				// Persist buffer data if needed.
				if (should_persist_single_buffer_umount(buffer))
					f2dfs_sync_single_buffer_foreground(sbi, buffer);
				// Free the buffer and mark operation as performed.
				f2dfs_free_single_buffer(sbi, head, buffer);
				ret = true;
				break;
			}
		}
	}

	return ret;
}

// Frees both bucket and buffer caches in the background thread and resets cache metrics.
static inline void f2dfs_free_cache_by_thread(
				struct f2fs_sb_info *sbi,
				int *cache_rounds, unsigned int *cache_ms)
{
	// Free buckets and buffers cache.
	f2dfs_free_buckets_cache_by_thread(sbi);
	f2dfs_free_buffers_cache_by_thread(sbi);

	// Reset cache statistics.
	(*cache_rounds) = 0;
	(*cache_ms) = 0;
}

// Determines if normal I/O operations should be avoided during specific conditions.
static inline bool f2dfs_avoid_noraml_io(struct f2fs_sb_info *sbi)
{
	// Avoid normal I/O if garbage collection is in urgent mode.
	if ((sbi->gc_mode == GC_URGENT_HIGH) || (sbi->gc_mode == GC_URGENT_LOW))
		return false;

	// Check for ongoing read/write or direct I/O operations.
	if (get_pages(sbi, F2FS_RD_DATA) || get_pages(sbi, F2FS_WB_DATA) ||
				get_pages(sbi, F2FS_DIO_READ) ||
				get_pages(sbi, F2FS_DIO_WRITE) ||
				get_pages(sbi, F2FS_WB_CP_DATA))
		return false;

	// Avoid I/O if there are queued flushes in the flush command control (FCC).
	if (SM_I(sbi) && SM_I(sbi)->fcc_info &&
				atomic_read(&SM_I(sbi)->fcc_info->queued_flush))
		return false;

	// Allow normal I/O otherwise.
	return true;
}

// Checks if the GC (Garbage Collection) lock can be avoided.
static inline bool f2dfs_avoid_gclock(struct f2fs_sb_info *sbi)
{
	// If GC mode is urgent, the GC lock cannot be avoided.
	if ((sbi->gc_mode == GC_URGENT_HIGH) || (sbi->gc_mode == GC_URGENT_LOW))
		return false;

	// Try acquiring the GC lock in write mode without blocking.
	if (!down_write_trylock(&sbi->gc_lock))
		return false;

	// Release the GC lock immediately after checking.
	up_write(&sbi->gc_lock);
	return true;
}

// Performs offline deduplication by the background thread if conditions are met.
static inline bool f2dfs_do_offline_deduplication_by_thread(
				struct f2fs_sb_info *sbi,
				struct f2dfs_fp_kthread *fp_th,
				unsigned int *wait_ms, unsigned int *offline_rounds,
				unsigned int *offline_ms, bool offline_urgent)
{
	bool ret = false;

	// Check if normal I/O operations should take precedence.
	if (!offline_urgent && !f2dfs_avoid_noraml_io(sbi)) {
		// Reset offline deduplication counters and log the reason.
		(*offline_rounds) = 0;
		(*offline_ms) = 0;
		stat_io_skip_bgfp_count(sbi);
		f2dfs_printk(sbi, KERN_INFO "Debug:skip offline deduplication cause "
				"business of foreground IO, wait:%ums",
				(*wait_ms));
		return false;
	}

	// Attempt to perform offline deduplication.
	ret = f2dfs_do_offline_deduplication(sbi, false, false, offline_urgent);

	if (!ret) {
		// Increment skip counter and log the reason if deduplication fails.
		stat_other_skip_bgfp_count(sbi);
		f2dfs_printk(sbi, KERN_INFO "Debug:skip offline deduplication cause "
				"urgent space left, wait:%ums",
				(*wait_ms));

		// Adjust wait time for urgent deduplication cases.
		if (offline_urgent)
			(*wait_ms) = fp_th->urgent_fp_sleep_time;
	}
	else {
		// Reset offline deduplication counters after successful deduplication.
		(*offline_rounds) = 0;
		(*offline_ms) = 0;
	}

	return ret;
}

// Coordinates the persistence and stabilization of caches and deduplication.
static inline bool f2dfs_do_persistence_by_thread(
				struct f2fs_sb_info *sbi,
				struct f2dfs_fp_kthread *fp_th, unsigned int *wait_ms,
				int *cache_rounds, unsigned int *cache_ms,
				int *stable_rounds, unsigned int *stable_ms,
				int *offline_rounds, unsigned int *offline_ms,
				bool offline_urgent)
{
	bool ret = true, could_cache = false, could_stable = false,
			could_offline = false;

	// Determine if offline deduplication can proceed based on urgency or thresholds.
	could_offline = (offline_urgent || could_do_offline_deduplication(
					(*offline_rounds), (*offline_ms)));

	// Determine if stabilization can proceed based on deduplication status or thresholds.
	could_stable = (could_offline || could_do_stabilization(
					(*stable_rounds), (*stable_ms)));

	// Determine if cache cleanup is required based on thresholds.
	could_cache = could_do_free_cache((*cache_rounds), (*cache_ms));

	// Perform cache cleanup if required.
	if (could_cache)
		f2dfs_free_cache_by_thread(sbi, cache_rounds, cache_ms);

	// Attempt persistence and stabilization.
	ret = f2dfs_do_persistence_and_stabilization(
					sbi, false, false, could_stable);

	if (could_stable) {
		// Reset stabilization counters after successful operation.
		(*stable_rounds) = 0;
		(*stable_ms) = 0;
	}

	if (!ret) {
		// Log a message if persistence fails due to urgent space conditions.
		stat_other_skip_bgfp_count(sbi);
		f2dfs_printk(sbi, KERN_INFO "Debug:skip persistence cause urgent "
				"space left, wait:%ums",
				(*wait_ms));
	}
	else {
		// Log a success message if persistence is completed.
		f2dfs_printk(sbi, KERN_INFO "Debug:do persistence successfully, "
				"wait:%ums",
				(*wait_ms));
	}

	// If persistence fails or offline deduplication cannot proceed, return the result.
	if ((!ret) || (!could_offline))
		return ret;

	// Attempt offline deduplication.
	ret = f2dfs_do_offline_deduplication_by_thread(sbi, fp_th, wait_ms,
					offline_rounds, offline_ms, offline_urgent);

	return ret;
}

static int fp_thread_func(void *data)
{
	struct f2fs_sb_info *sbi = data;
	struct f2dfs_fp_kthread *fp_th = sbi->fp_thread;
	wait_queue_head_t *wq = &sbi->fp_thread->fp_wait_queue_head;
	unsigned int wait_ms, cache_ms = 0, stable_ms = 0, offline_ms = 0;
	int cache_rounds = 0, stable_rounds = 0, offline_rounds = 0;
	bool offline_urgent = false;

	// wait_ms = fp_th->min_fp_sleep_time;
	wait_ms = fp_th->no_fp_sleep_time;

	set_freezable();
	do {
		wait_event_interruptible_timeout(*wq,
						(kthread_should_stop() || freezing(current)),
						msecs_to_jiffies(wait_ms));

		++cache_rounds;
		++stable_rounds;
		++offline_rounds;
		cache_ms += wait_ms;
		stable_ms += wait_ms;
		offline_ms += wait_ms;
		offline_urgent = should_dedupe_offline_urgent(sbi);

		if (try_to_freeze()) {
			stat_other_skip_bgfp_count(sbi);
			f2dfs_printk(sbi, KERN_INFO "Debug:skip persistence and offline "
					"deduplication cause try_to_freeze, wait:%ums",
					wait_ms);
			continue;
		}

		if (kthread_should_stop())
			break;

		if (sbi->sb->s_writers.frozen >= SB_FREEZE_WRITE) {
			// increase_fp_sleep_time(fp_th, &wait_ms);
			stat_other_skip_bgfp_count(sbi);
			f2dfs_printk(sbi, KERN_INFO "Debug:skip background FP operations "
					"cause FREEZE-writer, wait:%ums",
					wait_ms);
			continue;
		}

		if (time_to_inject(sbi, FAULT_CHECKPOINT)) {
			f2fs_show_injection_info(sbi, FAULT_CHECKPOINT);
			f2fs_stop_checkpoint(sbi, false);
		}

		if (!sb_start_write_trylock(sbi->sb)) {
			stat_other_skip_bgfp_count(sbi);
			f2dfs_printk(sbi, KERN_INFO "Debug:skip background FP operations "
					"cause sblock, wait:%ums",
					wait_ms);
			continue;
		}

		/*
		 * Background offline deduplication is triggered with the following
		 * conditions:
		 * 1. There are a number of blocks (40% of existed blocks, 20% of space
		 * utilization on device, 16GB logical data blocks) wait to do offline
		 * deduplication with one hundred percentages.
		 * 2. There are enough CPU and RAM resources, and power energy.
		 * 3. It is better to skip offline duplication during foreground file
		 * operation and background garbage collection.
		 */
		// if (offline_urgent) {
		// 	wait_ms = fp_th->urgent_fp_sleep_time;
		// 	down_write(&sbi->fp_lock);
		// 	goto do_fp;
		// }

		// if (!f2dfs_avoid_gclock(sbi)) {
		// 	// increase_fp_sleep_time(fp_th, &wait_ms);
		// 	stat_other_skip_bgfp_count(sbi);
		// 	f2dfs_printk(sbi, KERN_INFO "Debug:skip background FP operations "
		// 					"cause garbage collection, wait:%ums", wait_ms);
		// 	goto next;
		// }

		if (!down_write_trylock(&sbi->fp_lock)) {
			stat_other_skip_bgfp_count(sbi);
			f2dfs_printk(sbi, KERN_INFO "Debug:skip background FP operations "
					"cause fplock, wait:%ums",
					wait_ms);
			goto next;
		}

		if (!is_idle(sbi, FP_PERSIST_TIME)) {
			// increase_fp_sleep_time(fp_th, &wait_ms);
			up_write(&sbi->fp_lock);
			stat_io_skip_bgfp_count(sbi);
			f2dfs_printk(sbi, KERN_INFO "Debug:skip background FP operations "
					"cause not idle, wait:%ums",
					wait_ms);
			goto next;
		}

// do_fp:
// 		f2dfs_do_persistence_by_thread(sbi, fp_th, &wait_ms,
// 				&cache_rounds, &cache_ms, &stable_rounds, &stable_ms,
// 				&offline_rounds, &offline_ms, offline_urgent);

		up_write(&sbi->fp_lock);

next:
		sb_end_write(sbi->sb);
	} while (!kthread_should_stop());

	f2dfs_printk(sbi, KERN_INFO "Debug:%s:fp_persistence_and_offline_"
			"deduplication_return",
			__func__);
	return 0;
}

// Initializes the fingerprint hash structure for the filesystem.
static inline void f2dfs_init_fingerprint_hash(struct f2fs_sb_info *sbi)
{
	int i;

	// Allocate memory for the fingerprint hash structure.
	sbi->s_fingerprint_hash = f2fs_kvmalloc(sbi,
					(sizeof(struct f2dfs_fingerprint_hash)), GFP_NOFS);
	// Set the basic hash info and allocate the hash algorithm.
	sbi->s_fingerprint_hash->basic_hash_info = &sha1_info;
	sbi->s_fingerprint_hash->basic_hash_info->alg = crypto_alloc_shash(
					sbi->s_fingerprint_hash->basic_hash_info->name, 0, 0);
	// Allocate memory for the hash jobs structure.
	sbi->s_fingerprint_hash->hash_jobs = f2fs_kvmalloc(sbi,
					(sizeof(struct f2dfs_single_hash_job) *
					F2DFS_HASH_JOBS_NUMBER), GFP_NOFS);
	// Create a workqueue for parallel hashing.
	sbi->s_fingerprint_hash->parallel_hash_workqueue = alloc_workqueue(
					"f2dfs_fingerprint_hash_wq", WQ_UNBOUND | WQ_HIGHPRI,
					num_online_cpus());

	// Initialize mutexes for each hash job.
	for (i = 0; i < F2DFS_HASH_JOBS_NUMBER; i++)
		mutex_init(&sbi->s_fingerprint_hash->hash_jobs[i].work_mutex);
}

// Frees the memory and cleans up the fingerprint hash structure.
static inline void f2dfs_exit_fingerprint_hash(struct f2fs_sb_info *sbi)
{
	int i;

	// Lock and unlock each hash job's mutex before freeing.
	for (i = 0; i < F2DFS_HASH_JOBS_NUMBER; i++) {
		mutex_lock(&sbi->s_fingerprint_hash->hash_jobs[i].work_mutex);
		mutex_unlock(&sbi->s_fingerprint_hash->hash_jobs[i].work_mutex);
	}

	// Free the allocated resources.
	crypto_free_shash(sbi->s_fingerprint_hash->basic_hash_info->alg);
	destroy_workqueue(sbi->s_fingerprint_hash->parallel_hash_workqueue);
	kvfree(sbi->s_fingerprint_hash->hash_jobs);
	kvfree(sbi->s_fingerprint_hash);
}

// Initializes the fingerprint bucket cache for the filesystem.
static inline void f2dfs_init_buckets_cache(struct f2fs_sb_info *sbi)
{
	int index_cnt = FP_PREFIX_NUMBER, i = 0;

	// Allocate memory for the fingerprint bucket structure.
	sbi->s_fingerprint_bucket = f2fs_kvmalloc(sbi,
					(sizeof(struct f2dfs_fingerprint_bucket)), GFP_NOFS);
	sbi->s_fingerprint_bucket->current_number = 0;
	// Allocate memory for the list of pointers to the fingerprint buckets.
	sbi->s_fingerprint_bucket->pointers = f2fs_kvmalloc(sbi,
					(index_cnt * sizeof(struct slist_head)), GFP_NOFS);
	// Create the memory cache for the fingerprint bucket.
	sbi->s_fingerprint_bucket->bucket_cache =
					kmem_cache_create("f2dfs_fp_buckets_cache",
					sizeof(struct f2dfs_fp_bucket), 0, 0, NULL);

	// Initialize each list and allocate the fingerprint buckets.
	for (i = 0; i < index_cnt; ++i) {
		slist_head_init(&sbi->s_fingerprint_bucket->pointers[i]);
		f2dfs_alloc_single_bucket(sbi, &sbi->s_fingerprint_bucket->pointers[i]);
	}

	// Allocate and initialize stable bitmap for fingerprint storage.
	sbi->s_fingerprint_bucket->s_stable_bitmap = f2fs_kvzalloc(sbi,
					array_size(F2DFS_STABLE_BITMAP_SIZE,
					sizeof(__le32)), GFP_NOFS);
	for (i = 0; i < F2DFS_STABLE_BITMAP_SIZE; ++i)
		*(sbi->s_fingerprint_bucket->s_stable_bitmap + i) = 0;
}

// Frees the memory and cleans up the fingerprint bucket cache.
static inline void f2dfs_exit_buckets_cache(struct f2fs_sb_info *sbi)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	int i = 0;

	// Free each fingerprint bucket.
	for (i = 0; i < FP_PREFIX_NUMBER; ++i) {
		head = &sbi->s_fingerprint_bucket->pointers[i];
		while (!slist_empty(head)) {
			bucket = slist_first_entry(head, struct f2dfs_fp_bucket, b_head);
			f2dfs_free_single_bucket(sbi, head, bucket);
		}
	}

	// Destroy the memory cache and free the resources.
	kmem_cache_destroy(sbi->s_fingerprint_bucket->bucket_cache);
	kvfree(sbi->s_fingerprint_bucket->pointers);
	kvfree(sbi->s_fingerprint_bucket->s_stable_bitmap);
	kvfree(sbi->s_fingerprint_bucket);
}

// Initializes the fingerprint buffer cache for the filesystem.
static inline void f2dfs_init_buffers_cache(struct f2fs_sb_info *sbi)
{
	int i = 0;

	// Allocate memory for the fingerprint buffer structure.
	sbi->s_fingerprint_buffer = f2fs_kvmalloc(sbi,
					(sizeof(struct f2dfs_fingerprint_buffer)), GFP_NOFS);
	sbi->s_fingerprint_buffer->current_number = 0;
	// Allocate memory for the list of pointers to the fingerprint buffers.
	sbi->s_fingerprint_buffer->pointers = f2fs_kvmalloc(sbi,
					(FP_BUFFER_NUMBER * sizeof(struct slist_head)), GFP_NOFS);
	// Create the memory cache for the fingerprint buffer.
	sbi->s_fingerprint_buffer->buffer_cache =
					kmem_cache_create("f2dfs_fp_buffers_cache",
					sizeof(struct f2dfs_fp_buffer), 0, 0, NULL);

	// Initialize each list for the fingerprint buffers.
	for (i = 0; i < FP_BUFFER_NUMBER; ++i)
		slist_head_init(&sbi->s_fingerprint_buffer->pointers[i]);
}

// Frees the memory and cleans up the fingerprint buffer cache.
static inline void f2dfs_exit_buffers_cache(struct f2fs_sb_info *sbi)
{
	struct slist_head *head;
	struct f2dfs_fp_buffer *buffer;
	int i = 0;

	// Free each fingerprint buffer.
	for (i = 0; i < FP_BUFFER_NUMBER; ++i) {
		head = &sbi->s_fingerprint_buffer->pointers[i];
		while (!slist_empty(head)) {
			buffer = slist_first_entry(head, struct f2dfs_fp_buffer, b_head);
			f2dfs_free_single_buffer(sbi, head, buffer);
		}
	}

	// Destroy the memory cache and free the resources.
	kmem_cache_destroy(sbi->s_fingerprint_buffer->buffer_cache);
	kvfree(sbi->s_fingerprint_buffer->pointers);
	kvfree(sbi->s_fingerprint_buffer);
}

// Starts a fingerprint thread for the filesystem.
int f2dfs_start_fp_thread(struct f2fs_sb_info *sbi)
{
	struct f2dfs_fp_kthread *fp_th;
	dev_t dev = sbi->sb->s_bdev->bd_dev;
	int err = 0;

	// Return early if there is an error (this is currently a placeholder).
	return err;

	// Allocate memory for the fingerprint thread structure.
	fp_th = f2fs_kmalloc(sbi, sizeof(struct f2dfs_fp_kthread), GFP_KERNEL);
	if (!fp_th) {
		err = -ENOMEM;
		goto out;
	}

	// Set default values for the fingerprint thread's sleep times.
	fp_th->urgent_fp_sleep_time = DEF_FP_URGENT_TIME;
	fp_th->min_fp_sleep_time = DEF_FP_MIN_SLEEP_TIME;
	fp_th->max_fp_sleep_time = DEF_FP_MAX_SLEEP_TIME;
	fp_th->no_fp_sleep_time = DEF_FP_NO_SLEEP_TIME;

	// Print debug information.
	f2dfs_printk(sbi, KERN_INFO "Debug:%s:start fp thread",
			__func__);

	// Assign the fingerprint thread to the filesystem.
	sbi->fp_thread = fp_th;

	// Initialize the wait queue for the fingerprint thread.
	init_waitqueue_head(&sbi->fp_thread->fp_wait_queue_head);
	// Create the kernel thread for fingerprint processing.
	sbi->fp_thread->fp_task = kthread_run(fp_thread_func,
					sbi, "f2dfs_fp-%u:%u", MAJOR(dev), MINOR(dev));
	if (IS_ERR(fp_th->fp_task)) {
		err = PTR_ERR(fp_th->fp_task);
		kvfree(fp_th);
		sbi->fp_thread = NULL;
		goto out;
	}

out:
	return err;
}

// Stops the fingerprint thread for the filesystem and frees associated resources.
void f2dfs_stop_fp_thread(struct f2fs_sb_info *sbi)
{
	struct f2dfs_fp_kthread *fp_th;

	// Early return if no fingerprint thread exists.
	return;
	fp_th = sbi->fp_thread;
	if (!fp_th)
		return;

	// Stops the fingerprint kernel thread and frees resources.
	kthread_stop(fp_th->fp_task);
	kvfree(fp_th);
	sbi->fp_thread = NULL;

	// Print debug information indicating the thread has been stopped.
	f2dfs_printk(sbi, KERN_INFO "Debug:%s:stop fp thread", __func__);
}

// Returns the free space available in the filesystem based on the free sections and valid blocks.
unsigned int f2dfs_filesystem_free_space(struct f2fs_sb_info *sbi)
{
	// Calculates the minimum between free sections and valid user blocks.
	return min((unsigned int)((FREE_I(sbi)->free_sections) *
					sbi->blocks_per_seg * sbi->segs_per_sec),
			(unsigned int)((SM_I(sbi)->main_segments) *
					sbi->blocks_per_seg - valid_user_blocks(sbi)));
}

// Computes the fingerprint for a single page and stores it in the provided digest.
static inline void f2dfs_compute_single_fingerprint(struct f2fs_sb_info *sbi,
				struct sdesc *sdesc, struct page *page, __u8 digest[])
{
	crypto_shash_init(&sdesc->shash);
	sbi->s_fingerprint_hash->basic_hash_info->fp_update(
					&sdesc->shash, page_address(page), PAGE_SIZE);
	crypto_shash_final(&sdesc->shash, digest);
}

// Executes the fingerprint hashing work for a single page in the background.
static void f2dfs_do_single_fp_hash_work(struct work_struct *work)
{
	struct f2dfs_single_hash_job *job = container_of(work,
					struct f2dfs_single_hash_job, page_work);
	struct f2fs_sb_info *sbi = job->sbi;

	// Initializes the descriptor for the hashing job.
	job->sdesc = init_sdesc(sbi->s_fingerprint_hash->basic_hash_info->alg);
	f2dfs_compute_single_fingerprint(sbi, job->sdesc, job->page,
					job->search_result.fingerprint);
	kfree(job->sdesc);
	mutex_unlock(&job->work_mutex);
}

// Enqueues a fingerprint hashing job for a single page and starts the work in the background.
void f2dfs_enqueue_single_fp_hash_work(struct f2fs_sb_info *sbi,
				int job_no, struct page *page)
{
	struct f2dfs_single_hash_job *job =
					&sbi->s_fingerprint_hash->hash_jobs[job_no];

	// Locks the mutex to avoid race conditions during job setup.
	mutex_lock(&job->work_mutex);
	// Initializes the fingerprint job with the page information.
	job->sbi = sbi;
	memset(&job->search_result.fingerprint, 0, FP_LEN_MAX);
	job->index = (page->index);
	job->page = page;
	// Initializes the work function to process the fingerprint hashing.
	INIT_WORK(&job->page_work, f2dfs_do_single_fp_hash_work);
	// Enqueues the work to the workqueue for background processing.
	queue_work(sbi->s_fingerprint_hash->parallel_hash_workqueue,
					&job->page_work);
}

// Initializes the deduplication system for the filesystem.
void f2dfs_init_deduplication(struct f2fs_sb_info *sbi)
{
	// Initializes the fingerprint hash algorithm.
	f2dfs_init_fingerprint_hash(sbi);
	// Initializes the inline deduplication (fingerprint buckets).
	f2dfs_init_buckets_cache(sbi);
	// Initializes the offline deduplication (fingerprint buffers).
	f2dfs_init_buffers_cache(sbi);
}

// Exits and cleans up the deduplication system for the filesystem.
void f2dfs_exit_deduplication(struct f2fs_sb_info *sbi)
{
	// The commented-out code is meant for performing persistence and stabilization tasks during unmount.
	// struct cp_control cpc_force = {
	// 	.reason = CP_UMOUNT,
	// };
	// mutex_lock(&sbi->umount_mutex);
	// f2dfs_do_persistence_and_stabilization(sbi, true, false, true);
	// f2dfs_do_offline_deduplication(sbi, true, false, false);
	// f2fs_write_checkpoint(sbi, &cpc_force);
	// mutex_unlock(&sbi->umount_mutex);

	f2dfs_exit_buffers_cache(sbi);
	f2dfs_exit_buckets_cache(sbi);
	f2dfs_exit_fingerprint_hash(sbi);
}
