﻿#include <linux/fs.h>

#include "f2dfs_fs.h"
#include "f2fs.h"
#include "segment.h"

static inline void fpbucket_print_normal(
				struct f2dfs_fp_bucket *bucket, unsigned int index)
{
	// Print the details of a normal fingerprint bucket.
	int i = 0, j = 0;
	__u8 bucket_fp[FP_LEN_MAX];
	printk("--------------------------------------------------------%u", index);
	for (i = 0; i < bucket->valid_count; ++i) {
		// Copy the fingerprint data and print it in hexadecimal format.
		memcpy(bucket_fp, &bucket->entries[i].ne.fingerprint, FP_LEN_MAX);
		printk(" ");
		for(j = 0; j < FP_LEN_MAX; ++j)
			printk(KERN_CONT "%02X", bucket_fp[j]);
		// Print the associated virtual address.
		printk(KERN_CONT " %u", bucket->entries[i].ne.virtual_addr);
	}
}

static inline void fpbucket_print_remove(
				struct f2dfs_fp_bucket *bucket, unsigned int index)
{
	// Print the details of a remove fingerprint bucket.
	int i = 0, j = 0;
	printk("--------------------------------------------------------%u", index);
	for (i = FP_ENTRIES_PER_BUCKET - 1;
				i >= FP_ENTRIES_PER_BUCKET - bucket->remove_count; --i) {
		printk(" ");
		// Print all remove addresses within the current entry.
		for(j = 0; j < FP_REMOVE_PER_ENTRY; ++j)
			printk(KERN_CONT "%u ", bucket->entries[i].re.remove_addr[j]);
	}
}

static inline bool f2dfs_is_bucket_empty(struct f2dfs_fp_bucket *bucket)
{
	// Check if a bucket is empty (no valid or removed entries).
	return ((bucket->valid_count == 0) && (bucket->remove_count == 0));
}

static inline bool f2dfs_is_bucket_full(struct f2dfs_fp_bucket *bucket)
{
	// Check if a bucket is full (sum of valid and removed entries reaches the limit).
	return ((bucket->valid_count + bucket->remove_count) == FP_ENTRIES_PER_BUCKET);
}

static inline int f2dfs_search_single_fingerprint_in_single_bucket(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_bucket *bucket,
				struct f2dfs_bucket_change *search_result,
				virtual_t source_vaddr,
				struct f2dfs_global_counter *rc_changes)
{
	// Search for a single fingerprint in a bucket.
	int i = 0, count = bucket->valid_count;
	__u8 temp_fp[FP_LEN_MAX], temp_tags;

	// Initialize search result.
	search_result->virtual_addr = NULL_ADDR;
	search_result->is_stable = false;
	search_result->is_dupl = false;
	search_result->is_persist = false;

	if (f2dfs_is_bucket_empty(bucket))
		return FP_BUCKET_SEARCH_UNIQUE;

	for (i = 0; i < count; ++i) {
		// Compare the fingerprint with existing entries.
		memcpy(temp_fp, &bucket->entries[i].ne.fingerprint, FP_LEN_MAX);
		if (f2dfs_fp_is_equal(search_result->fingerprint, temp_fp) == 0) {
			goto hit;
		}
	}

	return FP_BUCKET_SEARCH_UNIQUE;

hit:
	// Fingerprint found. Update search result.
	search_result->virtual_addr = bucket->entries[i].ne.virtual_addr;
	temp_tags = bucket->entries[i].tags;

	// Move the matching entry to the top of the bucket.
	if (i != 0) {
		memmove(&bucket->entries[1], &bucket->entries[0],
					(sizeof(struct fp_bucket_entry) * i));
		memcpy(&bucket->entries[0].ne.fingerprint, search_result->fingerprint, FP_LEN_MAX);
		bucket->entries[0].ne.virtual_addr = search_result->virtual_addr;
		bucket->entries[0].tags = temp_tags;
	}

	// Update reference count if source address differs from the found address.
	if (source_vaddr != search_result->virtual_addr) {
		f2dfs_change_reference_count_inline(sbi, search_result->virtual_addr, 1, rc_changes);
		fp_set_dupl_tag(&bucket->entries[0]);
	}

	// Set the result tags based on the current entry.
	search_result->is_stable = (fp_is_stable_tag(&bucket->entries[0]) ? true : false);
	search_result->is_dupl = (fp_is_dupl_tag(&bucket->entries[0]) ? true : false);
	search_result->is_persist = (fp_is_persist_tag(&bucket->entries[0]) ? true : false);

	return FP_BUCKET_SEARCH_DUPL;
}

static inline int f2dfs_insert_single_fingerprint_into_single_bucket(
				struct f2dfs_fp_bucket *bucket,
				struct f2dfs_bucket_change *insert_change)
{
	// Insert a new fingerprint into a bucket.
	if (f2dfs_is_bucket_full(bucket))
		return FP_BUCKET_INSERT_FAILURE;

	// Shift existing entries to make room for the new one.
	memmove(&bucket->entries[1], &bucket->entries[0],
					(sizeof(struct fp_bucket_entry) * bucket->valid_count));
	memcpy(&bucket->entries[0].ne.fingerprint, insert_change->fingerprint, FP_LEN_MAX);
	bucket->entries[0].ne.virtual_addr = insert_change->virtual_addr;

	// Initialize tags for the new entry.
	init_fp_tag(&bucket->entries[0]);
	++bucket->valid_count;

	return FP_BUCKET_INSERT_SUCCESS;
}

static inline int f2dfs_remove_single_vaddr_from_single_bucket(
				struct f2dfs_fp_bucket *bucket,
				struct f2dfs_bucket_change *remove_change)
{
	// Remove a virtual address from a bucket.
	int i = 0, j = 0, count = bucket->valid_count;
	bool remove_and_insert = false, pos = false;

	// Search for the entry to remove.
	for (i = 0; i < count; ++i) {
		if (bucket->entries[i].ne.virtual_addr == remove_change->virtual_addr) {
			goto hit;
		}
	}

insert:
	// Insert the removed virtual address into the remove section if needed.
	if (f2dfs_is_bucket_full(bucket))
		return FP_BUCKET_REMOVE_FAILURE;

	if (bucket->remove_count == 0) {
		++bucket->remove_count;
		j = 0;
		goto fill;
	}

	// Find an empty slot in the remove section.
	for (j = 0; j < FP_REMOVE_PER_ENTRY; ++j) {
		if (bucket->entries[FP_ENTRIES_PER_BUCKET -
					bucket->remove_count].re.remove_addr[j] == NULL_ADDR) {
			pos = true;
			break;
		}
	}

	if (!pos) {
		++bucket->remove_count;
		j = 0;
	}

fill:
	if (j == 0) {
		memset(&bucket->entries[FP_ENTRIES_PER_BUCKET - bucket->remove_count],
						0 , sizeof(struct fp_bucket_entry));
		init_fp_tag(&bucket->entries[FP_ENTRIES_PER_BUCKET - bucket->remove_count]);
		fp_set_remove_tag(&bucket->entries[FP_ENTRIES_PER_BUCKET - bucket->remove_count]);
		fp_set_persist_tag(&bucket->entries[FP_ENTRIES_PER_BUCKET - bucket->remove_count]);
	}

	// Add the removed address to the appropriate slot.
	bucket->entries[FP_ENTRIES_PER_BUCKET - bucket->remove_count].re.remove_addr[j] =
					remove_change->virtual_addr;

	return FP_BUCKET_REMOVE_SUCCESS;

hit:
	// If the entry has persistence, flag for re-insertion.
	if (fp_is_persist_tag(&bucket->entries[FP_ENTRIES_PER_BUCKET - bucket->remove_count]))
		remove_and_insert = true;

	// Shift entries to remove the matching entry.
	if (i != bucket->valid_count - 1) {
		memmove(&bucket->entries[i], &bucket->entries[i + 1],
					(sizeof(struct fp_bucket_entry) *
					(bucket->valid_count - i - 1)));
	}

	--bucket->valid_count;
	if (remove_and_insert)
		goto insert;

	return FP_BUCKET_REMOVE_SUCCESS;
}

// Check if a single bucket can be balanced offline by removing entries
// with specific conditions (non-stable and/or non-persistent entries).
static inline bool balance_single_bucket_by_offline(
				struct f2dfs_fp_bucket *bucket)
{
	int i;
	bool hit = false;

	// Try to find a non-stable persistent entry to remove.
	for (i = bucket->valid_count - 1; i >= 0; --i) {
		if (fp_is_persist_tag(&bucket->entries[i]) &&
					(!fp_is_stable_tag(&bucket->entries[i]))) {
			hit = true;
			goto move;
		}
	}

	// If not found, try to find a stable persistent entry to remove.
	for (i = bucket->valid_count - 1; i >= 0; --i) {
		if (fp_is_persist_tag(&bucket->entries[i]) &&
					fp_is_stable_tag(&bucket->entries[i])) {
			hit = true;
			goto move;
		}
	}

	return false;

move:
	// Remove the selected entry and shift remaining entries.
	memmove(&bucket->entries[i], &bucket->entries[i + 1],
					(sizeof(struct fp_bucket_entry) *
					(bucket->valid_count - i - 1)));
	--bucket->valid_count;

	return true;
}

// Search for a stable fingerprint in a single bucket and return the result.
// Returns either FP_BUCKET_SEARCH_STABLE or FP_BUCKET_SEARCH_UNSTABLE.
static inline int f2dfs_get_single_fingerprint_stable_of_single_bucket(
				struct f2dfs_fp_bucket *bucket,
				struct f2dfs_bucket_change *stable_result)
{
	int i = 0, count = bucket->valid_count;
	__u8 temp_fp[FP_LEN_MAX], temp_tags;

	// Initialize the result structure.
	stable_result->virtual_addr = NULL_ADDR;
	stable_result->is_stable = false;
	stable_result->is_persist = false;

	// If the bucket is empty, return unstable.
	if (f2dfs_is_bucket_empty(bucket))
		return FP_BUCKET_SEARCH_UNSTABLE;

	// Search for the fingerprint in the bucket.
	for (i = 0; i < count; ++i) {
		memcpy(temp_fp, &bucket->entries[i].ne.fingerprint, FP_LEN_MAX);
		if (f2dfs_fp_is_equal(stable_result->fingerprint, temp_fp) == 0) {
			goto hit;
		}
	}

	return FP_BUCKET_SEARCH_UNSTABLE;

hit:
	// Update the result structure with the found entry's information.
	stable_result->virtual_addr = bucket->entries[i].ne.virtual_addr;
	temp_tags = bucket->entries[i].tags;
	if (i == 0)
		goto out;

	// Move the found entry to the front of the bucket.
	memmove(&bucket->entries[1], &bucket->entries[0],
					(sizeof(struct fp_bucket_entry) * i));
	memcpy(&bucket->entries[0].ne.fingerprint,
					stable_result->fingerprint, FP_LEN_MAX);
	bucket->entries[0].ne.virtual_addr = stable_result->virtual_addr;
	bucket->entries[0].tags = temp_tags;

out:
	// Update stability and persistence flags in the result.
	stable_result->is_stable =
					(fp_is_stable_tag(&bucket->entries[0]) ? true : false);
	stable_result->is_persist =
					(fp_is_persist_tag(&bucket->entries[0]) ? true : false);

	return (stable_result->is_stable ?
					FP_BUCKET_SEARCH_STABLE : FP_BUCKET_SEARCH_UNSTABLE);
}

// Set a fingerprint in a bucket as stable.
// Returns success or failure depending on bucket capacity and conditions.
static inline int f2dfs_set_single_fingerprint_stable_of_single_bucket(
				struct f2dfs_fp_bucket *bucket,
				struct f2dfs_bucket_change *stable_change)
{
	int i = 0, count = bucket->valid_count;
	__u8 temp_fp[FP_LEN_MAX];

	// Search for the fingerprint in the bucket.
	for (i = 0; i < count; ++i) {
		memcpy(temp_fp, &bucket->entries[i].ne.fingerprint, FP_LEN_MAX);
		if (f2dfs_fp_is_equal(stable_change->fingerprint, temp_fp) == 0)
			goto hit;
	}

	// If the bucket is full, attempt to evict entries offline.
	if (f2dfs_is_bucket_full(bucket)) {
		if (balance_single_bucket_by_offline(bucket))
			return FP_BUCKET_STABLE_FAILURE; // Bucket filled with non-persistent entries.
		i = bucket->valid_count;
	}
	++bucket->valid_count;

hit:
	// Insert or update the fingerprint as stable in the bucket.
	memmove(&bucket->entries[1], &bucket->entries[0],
					(sizeof(struct fp_bucket_entry) * i));
	memcpy(&bucket->entries[0].ne.fingerprint,
					stable_change->fingerprint, FP_LEN_MAX);
	bucket->entries[0].ne.virtual_addr = stable_change->virtual_addr;
	init_fp_tag(&bucket->entries[0]);
	fp_set_dupl_tag(&bucket->entries[0]);
	fp_set_persist_tag(&bucket->entries[0]);
	fp_set_stable_tag(&bucket->entries[0]);
	stable_change->is_stable = true;
	stable_change->is_dupl = true;
	stable_change->is_persist = true;

	return FP_BUCKET_STABLE_SUCCESS;
}

// Balance a bucket using the CLOCK algorithm for eviction.
// Evict entries with the "remove" tag based on specific conditions.
static inline void balance_single_bucket_by_CLOCK(
				struct f2dfs_fp_bucket *bucket)
{
	int times = bucket->valid_count - FP_BUCKET_EVICT_END, i, move_step = 0;

	// First pass: Evict entries without the "duplicate" tag.
	for (i = bucket->valid_count - 1; i >= 0; --i) {
		if (times == 0)
			break;

		if (fp_is_dupl_tag(&bucket->entries[i]))
			fp_clear_dupl_tag(&bucket->entries[i]);
		else {
			fp_set_remove_tag(&bucket->entries[i]);
			--times;
		}
	}

	// Second pass: Evict remaining entries without the "remove" tag.
	if (times > 0) {
		for (i = bucket->valid_count - 1; i >= 0; --i) {
			if (times == 0)
				break;
			if (fp_is_remove_tag(&bucket->entries[i]))
				continue;

			else {
				fp_set_remove_tag(&bucket->entries[i]);
				--times;
			}
		}
	}

	// Shift all entries to remove gaps caused by evictions.
	for (i = 0; i < bucket->valid_count; ++i) {
		if (fp_is_remove_tag(&bucket->entries[i])) {
			++move_step;
			continue;
		}

		if (move_step > i)
			break;
		else if (move_step > 0)
			memmove(&bucket->entries[i - move_step], &bucket->entries[i],
							sizeof(struct fp_bucket_entry));
	}
}

// Balances a single bucket based on the configured eviction strategy (LRU or CLOCK).
static inline void f2dfs_balance_single_bucket(struct f2dfs_fp_bucket *bucket)
{
	switch(FP_BUCKET_EVICTION_CRA) {
		case FP_BUCKET_EVICTION_LRU: // LRU strategy: limit bucket's valid entries to the threshold
			if (bucket->valid_count > FP_BUCKET_EVICT_END)
				bucket->valid_count = FP_BUCKET_EVICT_END;
			break;

		case FP_BUCKET_EVICTION_CLOCK: // CLOCK strategy: evict entries using CLOCK algorithm
			if (bucket->valid_count > FP_BUCKET_EVICT_END) {
				balance_single_bucket_by_CLOCK(bucket);
				bucket->valid_count = FP_BUCKET_EVICT_END;
			}
			break;

		default:
			break; // No eviction strategy defined
	}
}

// Searches for a fingerprint in a specific prefix of the fingerprint table.
static inline int f2dfs_search_single_fingerprint_in_single_prefix(
				struct f2fs_sb_info *sbi,
				struct f2dfs_bucket_change *search_result,
				virtual_t source_vaddr,
				struct f2dfs_global_counter *rc_changes)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	int ret = FP_BUCKET_SEARCH_UNIQUE, index = search_result->fp_prefix;

	// Iterate through all buckets in the target prefix
	slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[index]) {
		if (!head)
			return ret;

		bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
		if (!bucket)
			return ret;

		// Search the fingerprint within the current bucket
		ret = f2dfs_search_single_fingerprint_in_single_bucket(sbi, bucket,
						search_result, source_vaddr, rc_changes);
		if (ret == FP_BUCKET_SEARCH_DUPL) { // Found a duplicate
			bucket->atime = jiffies; // Update access time
			return ret;
		}
	}

	return ret;
}

// Inserts a fingerprint into a specific prefix in the fingerprint table.
static inline int f2dfs_insert_single_fingerprint_into_single_prefix(
				struct f2fs_sb_info *sbi,
				struct f2dfs_bucket_change *insert_change)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	int ret = FP_BUCKET_INSERT_FAILURE, index = insert_change->fp_prefix;

insert_redo:
	// Iterate through all buckets in the target prefix
	slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[index]) {
		if (!head)
			goto bucket_alloc;

		bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
		if (!bucket)
			goto bucket_alloc;

		// Try inserting the fingerprint into the current bucket
		ret = f2dfs_insert_single_fingerprint_into_single_bucket(
						bucket, insert_change);
		if (ret == FP_BUCKET_INSERT_FAILURE) { // Handle full bucket
			// Attempt to shrink the bucket and retry
			f2dfs_shrink_single_bucket(sbi, bucket);
			goto insert_redo;
		}
		bucket->atime = jiffies; // Update access time

		return ret;
	}

bucket_alloc:
	// Allocate a new bucket if necessary
	bucket = f2dfs_alloc_single_bucket(
					sbi, &sbi->s_fingerprint_bucket->pointers[index]);
	ret = f2dfs_insert_single_fingerprint_into_single_bucket(
					bucket, insert_change);
	bucket->atime = jiffies; // Update access time

	return ret;
}

// Removes a virtual address entry from a specific prefix in the fingerprint table.
static inline int f2dfs_remove_single_vaddr_from_single_prefix(
				struct f2fs_sb_info *sbi,
				struct f2dfs_bucket_change *remove_change)
{
	struct slist_head * head;
	struct f2dfs_fp_bucket *bucket;
	int ret = FP_BUCKET_REMOVE_FAILURE, index = remove_change->fp_prefix;

remove_redo:
	// Iterate through all buckets in the target prefix
	slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[index]) {
		if (!head)
			goto bucket_alloc;

		bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
		if (!bucket)
			goto bucket_alloc;

		// Try removing the virtual address from the current bucket
		ret = f2dfs_remove_single_vaddr_from_single_bucket(
						bucket, remove_change);
		if (ret == FP_BUCKET_REMOVE_FAILURE) { // Handle removal failure
			// Attempt to shrink the bucket and retry
			f2dfs_shrink_single_bucket(sbi, bucket);
			goto remove_redo;
		}
		bucket->atime = jiffies; // Update access time

		return ret;
	}

bucket_alloc:
	// Allocate a new bucket if necessary
	bucket = f2dfs_alloc_single_bucket(
					sbi, &sbi->s_fingerprint_bucket->pointers[index]);
	ret = f2dfs_remove_single_vaddr_from_single_bucket(bucket, remove_change);
	bucket->atime = jiffies; // Update access time

	return ret;
}

// Retrieves the stable fingerprint from a specific prefix in the fingerprint table.
static inline int f2dfs_get_single_fingerprint_stable_of_single_prefix(
				struct f2fs_sb_info *sbi,
				struct f2dfs_bucket_change *stable_result)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	int ret = FP_BUCKET_SEARCH_UNSTABLE, index = stable_result->fp_prefix;

	// Iterate through all buckets in the target prefix
	slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[index]) {
		if (!head)
			return ret;

		bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
		if (!bucket)
			return ret;

		// Try retrieving the stable fingerprint from the current bucket
		ret = f2dfs_get_single_fingerprint_stable_of_single_bucket(
						bucket, stable_result);
		if (ret == FP_BUCKET_SEARCH_STABLE) { // Found a stable fingerprint
			bucket->atime = jiffies; // Update access time
			return ret;
		}
	}

	return ret;
}

// Sets a fingerprint as stable in a specific prefix of the fingerprint table.
static inline int f2dfs_set_single_fingerprint_stable_of_single_prefix(
				struct f2fs_sb_info *sbi,
				struct f2dfs_bucket_change *stable_change)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	int ret = FP_BUCKET_STABLE_FAILURE, index = stable_change->fp_prefix;

	// Iterate through all buckets in the target prefix
	slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[index]) {
		if (!head)
			goto bucket_alloc;

		bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
		if (!bucket)
			goto bucket_alloc;

		// Try setting the fingerprint as stable in the current bucket
		ret = f2dfs_set_single_fingerprint_stable_of_single_bucket(
						bucket, stable_change);
		if (ret == FP_BUCKET_STABLE_FAILURE)
			return ret;

		bucket->atime = jiffies; // Update access time
		return ret;
	}

bucket_alloc:
	// Allocate a new bucket if necessary
	bucket = f2dfs_alloc_single_bucket(
					sbi, &sbi->s_fingerprint_bucket->pointers[index]);
	ret = f2dfs_set_single_fingerprint_stable_of_single_bucket(
					bucket, stable_change);
	bucket->atime = jiffies; // Update access time

	return ret;
}

// Appends a single buffer entry with a specified inode NID, parent NID, and reference count change.
// If the parent NID already exists, update its reference count. If the entry count exceeds the buffer limit,
// return an append failure. The function maintains the most recently updated entry at the head of the buffer.
static inline int f2dfs_append_single_buffer_single_nid(
				struct f2dfs_fp_buffer *buffer, nid_t inode_nid,
				nid_t parent_nid, int rc_change)
{
	int count = buffer->valid_count, i = 0, temp_rc, temp_nid, temp_ino;
	bool hit = false;

	// Search for the parent NID in the buffer and update its reference count if found
	for (i = 0; i < count; ++i) {
		if (buffer->entries[i].parent_nid == parent_nid) {
			buffer->entries[i].number += rc_change;
			hit = true;
			if (buffer->entries[i].number == 0)
				goto subtract; // Remove the entry if the reference count becomes zero
			break;
		}
	}

	// If the parent NID is not found, append it as a new entry at the head of the buffer
	if (!hit) {
		if (buffer->valid_count == FP_ENTRIES_PER_BUFFER)
			return FP_BUFFER_APPEND_NID_FAILURE; // Return failure if buffer is full

		memmove(&buffer->entries[1], &buffer->entries[0],
						(sizeof(struct fp_buffer_entry) * buffer->valid_count));
		buffer->entries[0].inode_nid = inode_nid;
		buffer->entries[0].parent_nid = parent_nid;
		buffer->entries[0].number = rc_change;
		++buffer->valid_count;
	}
	else {
		// Reorganize the buffer to move the updated entry to the head
		temp_ino = buffer->entries[i].inode_nid;
		temp_nid = buffer->entries[i].parent_nid;
		temp_rc = buffer->entries[i].number;
		memmove(&buffer->entries[1], &buffer->entries[0],
						(sizeof(struct fp_buffer_entry) * i));
		buffer->entries[0].inode_nid = temp_ino;
		buffer->entries[0].parent_nid = temp_nid;
		buffer->entries[0].number = temp_rc;
	}

	return FP_BUFFER_APPEND_NID_SUCCESS;

subtract:
	// Remove an entry from the buffer when its reference count reaches zero
	memmove(&buffer->entries[i], &buffer->entries[i + 1],
					(sizeof(struct fp_buffer_entry) *
					(buffer->valid_count - i - 1)));
	--buffer->valid_count;

	return FP_BUFFER_APPEND_NID_SUCCESS;
}

// Loads a single buffer entry for a specified inode NID, parent NID, and reference count change.
// Similar to the append function but may load data from persistent storage if required.
static inline int f2dfs_load_single_buffer_single_nid(
				struct f2dfs_fp_buffer *buffer, nid_t inode_nid,
				nid_t parent_nid, int rc_change)
{
	int count = buffer->valid_count, i = 0, temp_rc, temp_nid, temp_ino;
	bool hit = false;

	// Search for the parent NID in the buffer and update its reference count if found
	for (i = 0; i < count; ++i) {
		if (buffer->entries[i].parent_nid == parent_nid) {
			buffer->entries[i].number += rc_change;
			hit = true;
			if (buffer->entries[i].number == 0)
				goto subtract; // Remove the entry if the reference count becomes zero
			break;
		}
	}

	// Append a new entry if the parent NID is not found
	if (!hit) {
		if (buffer->valid_count == FP_ENTRIES_PER_BUFFER)
			return FP_BUFFER_LOAD_NID_FAILURE; // Return failure if buffer is full

		memmove(&buffer->entries[1], &buffer->entries[0],
						(sizeof(struct fp_buffer_entry) * buffer->valid_count));
		buffer->entries[0].inode_nid = inode_nid;
		buffer->entries[0].parent_nid = parent_nid;
		buffer->entries[0].number = rc_change;
		++buffer->valid_count;
	}
	else {
		// Reorganize the buffer to move the updated entry to the head
		temp_ino = buffer->entries[i].inode_nid;
		temp_nid = buffer->entries[i].parent_nid;
		temp_rc = buffer->entries[i].number;
		memmove(&buffer->entries[1], &buffer->entries[0],
						(sizeof(struct fp_buffer_entry) * i));
		buffer->entries[0].inode_nid = temp_ino;
		buffer->entries[0].parent_nid = temp_nid;
		buffer->entries[0].number = temp_rc;
	}

	return FP_BUFFER_LOAD_NID_SUCCESS;

subtract:
	// Remove an entry when its reference count reaches zero
	memmove(&buffer->entries[i], &buffer->entries[i + 1],
					(sizeof(struct fp_buffer_entry) *
					(buffer->valid_count - i - 1)));
	--buffer->valid_count;

	return FP_BUFFER_LOAD_NID_SUCCESS;
}

// Appends a single NID entry to the cache with the specified parameters.
// Allocates and synchronizes buffers if necessary.
static inline int f2dfs_append_buffers_cache_single_nid(
				struct f2fs_sb_info *sbi, nid_t inode_nid,
				nid_t parent_nid, int rc_change)
{
	struct slist_head *head;
	struct f2dfs_fp_buffer *buffer;
	int ret = FP_BUFFER_APPEND_NID_FAILURE,
			index = (inode_nid & FP_SUFFIX_MASK_BIT);

insert_redo:
	// Iterate over buffer list to find a suitable buffer for insertion
	slist_for_each(head, &sbi->s_fingerprint_buffer->pointers[index]) {
		if (!head)
			goto buffer_alloc;

		buffer = slist_entry(head, struct f2dfs_fp_buffer, b_head);
		if (!buffer)
			goto buffer_alloc;

		ret = f2dfs_append_single_buffer_single_nid(
						buffer, inode_nid, parent_nid, rc_change);
		if (ret == FP_BUFFER_APPEND_NID_FAILURE) {
			// Synchronize and shrink the buffer if insertion fails
			printk("f2dfs_sync_single_buffer_foreground");
			f2dfs_sync_single_buffer_foreground(sbi, buffer);
			goto insert_redo;
		}
		buffer->atime = jiffies;

		return ret;
	}

buffer_alloc:
	// Allocate a new buffer if no suitable buffer is found
	buffer = f2dfs_alloc_single_buffer(
					sbi, &sbi->s_fingerprint_buffer->pointers[index]);
	ret = f2dfs_append_single_buffer_single_nid(
					buffer, inode_nid, parent_nid, rc_change);
	buffer->atime = jiffies;

	return ret;
}

// Load a single buffer from the fingerprint cache based on inode_nid and parent_nid.
static inline int f2dfs_load_buffers_cache_single_nid(
                struct f2fs_sb_info *sbi, nid_t inode_nid,
                nid_t parent_nid, int rc_change)
{
    struct slist_head *head;
    struct f2dfs_fp_buffer *buffer;
    int ret = FP_BUFFER_APPEND_NID_FAILURE,
        index = (inode_nid & FP_SUFFIX_MASK_BIT); // Compute hash index for buffer.

    // Iterate through the buffer list at the computed index.
    slist_for_each(head, &sbi->s_fingerprint_buffer->pointers[index]) {
        if (!head) // If the list head is invalid, proceed to buffer allocation.
            goto buffer_alloc;

        buffer = slist_entry(head, struct f2dfs_fp_buffer, b_head);
        if (!buffer) // If buffer entry is invalid, proceed to buffer allocation.
            goto buffer_alloc;

        // Attempt to load the buffer with the given nid and rc_change.
        ret = f2dfs_load_single_buffer_single_nid(
                        buffer, inode_nid, parent_nid, rc_change);
        return ret; // Return the result of the buffer load operation.
    }

buffer_alloc:
    // Allocate a new buffer if no matching buffer is found.
    buffer = f2dfs_alloc_single_buffer(
                    sbi, &sbi->s_fingerprint_buffer->pointers[index]);
    ret = f2dfs_load_single_buffer_single_nid(
                    buffer, inode_nid, parent_nid, rc_change);

    return ret; // Return the result of the load operation.
}

// Synchronize the fingerprint bucket cache by resetting valid counts.
bool f2dfs_ioc_sync_buckets_cache(struct f2fs_sb_info *sbi)
{
    struct slist_head *head;
    struct f2dfs_fp_bucket *bucket;
    int i;

    // Loop through all fingerprint bucket lists.
    for (i = 0; i < FP_PREFIX_NUMBER; ++i) {
        head = &sbi->s_fingerprint_bucket->pointers[i];
        if (!head) // If the list head is invalid, synchronization fails.
            return false;

        // Iterate through each bucket in the list.
        slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[i]) {
            if (!head) // Invalid bucket list entry.
                return false;

            bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
            if (!bucket) // Invalid bucket entry.
                return false;

            bucket->valid_count = 0; // Reset the valid count of the bucket.
        }
    }

    return true; // Return success.
}

// Shrink a single bucket by resetting its remove count and balancing it.
void f2dfs_shrink_single_bucket(struct f2fs_sb_info *sbi,
                    struct f2dfs_fp_bucket *bucket)
{
    bucket->remove_count = 0; // Reset the remove count of the bucket.
    f2dfs_balance_single_bucket(bucket); // Balance the bucket to maintain consistency.
}

// Append an NID to the fingerprint cache.
int f2dfs_append_nid_into_cache(struct f2fs_sb_info *sbi, nid_t inode_nid,
                nid_t parent_nid, int rc_change)
{
    int result = FP_BUFFER_APPEND_NID_SUCCESS;

    // Validate the parameters. Return failure if any are invalid.
    if ((rc_change == 0) || (inode_nid < F2DFS_START_FREE_NID) ||
                (f2fs_check_nid_range(sbi, inode_nid) != 0) ||
                (parent_nid < F2DFS_START_FREE_NID) ||
                (f2fs_check_nid_range(sbi, parent_nid) != 0))
        return FP_BUFFER_APPEND_NID_FAILURE;

    // Attempt to append the NID to the cache.
    result = f2dfs_append_buffers_cache_single_nid(
                    sbi, inode_nid, parent_nid, rc_change);
    return result;
}

// Load an NID into the fingerprint cache.
int f2dfs_load_nid_into_cache(struct f2fs_sb_info *sbi, nid_t inode_nid,
                nid_t parent_nid, int rc_change)
{
    int result = FP_BUFFER_LOAD_NID_SUCCESS;

    // Validate the parameters. Return success if any are invalid.
    if ((rc_change == 0) || (inode_nid < F2DFS_START_FREE_NID) ||
                (f2fs_check_nid_range(sbi, inode_nid) != 0) ||
                (parent_nid < F2DFS_START_FREE_NID) ||
                (f2fs_check_nid_range(sbi, parent_nid) != 0))
        return FP_BUFFER_LOAD_NID_SUCCESS;

    // Attempt to load the NID into the cache.
    result = f2dfs_load_buffers_cache_single_nid(
                    sbi, inode_nid, parent_nid, rc_change);
    return result;
}

// Process a single fingerprint operation on the bucket cache.
int f2dfs_process_single_fingerprint_for_bucket_cache(
                struct f2fs_sb_info *sbi,
                struct f2dfs_bucket_change *bucket_info,
                virtual_t source_vaddr,
                struct f2dfs_global_counter *rc_changes, int mode)
{
    int ret = FP_BUCKET_CACHE_FOR_NOTHING;

    // Validate the fingerprint prefix and operation mode.
    if ((bucket_info->fp_prefix >= DIRECT_FP_PREFIX) ||
                (bucket_info->fp_prefix < 0))
        return ret;

    // Ensure the operation mode matches the intended operation.
    if ((bucket_info->is_remove ? true : false) !=
                (mode == FP_BUCKET_CACHE_FOR_REMOVE ? true : false))
        return ret;

    switch (mode) {
        case FP_BUCKET_CACHE_FOR_SEARCH:
            // Handle fingerprint search operations.
            bucket_info->virtual_addr = NULL_ADDR;
            bucket_info->is_stable = false;
            bucket_info->is_dupl = false;
            bucket_info->is_persist = false;
            if (f2dfs_fp_is_zero(sbi, bucket_info->fingerprint,
                        bucket_info->fp_prefix)) {
                // Handle zero fingerprint cases.
                bucket_info->virtual_addr = ZERO_ADDR;
                bucket_info->is_stable = true;
                bucket_info->is_dupl = true;
                bucket_info->is_persist = true;
                f2dfs_change_reference_count_inline(sbi,
                                ZERO_ADDR, 1, rc_changes);
                ret = FP_BUCKET_SEARCH_ZERO;
            }
            else if (f2dfs_fp_is_remove(bucket_info->fingerprint)) {
                // Mark the fingerprint as unique.
                bucket_info->fp_prefix = DIRECT_FP_PREFIX;
                ret = FP_BUCKET_SEARCH_UNIQUE;
            }
            else {
                // Search the fingerprint in the prefix.
                ret = f2dfs_search_single_fingerprint_in_single_prefix(
                                sbi, bucket_info, source_vaddr, rc_changes);
            }
            break;

        case FP_BUCKET_CACHE_FOR_INSERT:
            // Handle fingerprint insert operations.
            if ((!__is_valid_data_blkaddr(bucket_info->virtual_addr)) ||
                        f2dfs_fp_is_remove(bucket_info->fingerprint)) {
                bucket_info->fp_prefix = DIRECT_FP_PREFIX;
                ret = FP_BUCKET_INSERT_FAILURE;
            }
            else if ((bucket_info->virtual_addr == ZERO_ADDR) ||
                        f2dfs_fp_is_zero(sbi, bucket_info->fingerprint,
                        bucket_info->fp_prefix)) {
                ret = FP_BUCKET_INSERT_SUCCESS;
            }
            else if (!is_real_dnode_physical(bucket_info->virtual_addr)) {
                ret = f2dfs_insert_single_fingerprint_into_single_prefix(
                                sbi, bucket_info);
            }
            break;

        case FP_BUCKET_CACHE_FOR_REMOVE:
            // Handle fingerprint removal operations.
            if (!__is_valid_data_blkaddr(bucket_info->virtual_addr)) {
                ret = FP_BUCKET_REMOVE_FAILURE;
            }
            else if (bucket_info->virtual_addr == ZERO_ADDR) {
                ret = FP_BUCKET_REMOVE_SUCCESS;
            }
            else if (!is_real_dnode_physical(bucket_info->virtual_addr)) {
                ret = f2dfs_remove_single_vaddr_from_single_prefix(
                                sbi, bucket_info);
            }
            break;

        case FP_BUCKET_CACHE_FOR_GET_STABLE:
            // Handle operations to get the stable state of a fingerprint.
            bucket_info->virtual_addr = NULL_ADDR;
            bucket_info->is_stable = false;
            bucket_info->is_dupl = false;
            bucket_info->is_persist = false;
            if (f2dfs_fp_is_zero(sbi, bucket_info->fingerprint,
                        bucket_info->fp_prefix)) {
                bucket_info->virtual_addr = ZERO_ADDR;
                bucket_info->is_stable = true;
                bucket_info->is_dupl = true;
                bucket_info->is_persist = true;
                ret = FP_BUCKET_SEARCH_STABLE;
            }
            else if (!is_real_dnode_physical(bucket_info->virtual_addr)) {
                ret = f2dfs_get_single_fingerprint_stable_of_single_prefix(
                                sbi, bucket_info);
            }
            break;

        case FP_BUCKET_CACHE_FOR_SET_STABLE:
            // Handle operations to set the stable state of a fingerprint.
            if (!__is_valid_data_blkaddr(bucket_info->virtual_addr)) {
                ret = FP_BUCKET_STABLE_FAILURE;
            }
            else if (f2dfs_fp_is_zero(sbi, bucket_info->fingerprint,
                        bucket_info->fp_prefix)) {
                ret = FP_BUCKET_STABLE_SUCCESS;
            }
            else if (!is_real_dnode_physical(bucket_info->virtual_addr)) {
                ret = f2dfs_set_single_fingerprint_stable_of_single_prefix(
                                sbi, bucket_info);
            }
            break;

        default:
            BUG(); // Invalid operation mode; trigger a kernel bug.
    }

    return ret;
}
