﻿#include <linux/fs.h>
#include "f2dfs_fs.h"
#include <linux/mpage.h>
#include <linux/backing-dev.h>
#include <linux/blkdev.h>
#include <linux/pagevec.h>
#include <linux/swap.h>

#include "f2fs.h"
#include "node.h"
#include "segment.h"
#include "xattr.h"
#include "trace.h"
#include <trace/events/f2fs.h>


/**
 * Converts a 16-bit number to a 32-bit number with sign extension.
 * 
 * @param number The 16-bit number to be converted.
 * @return The converted 32-bit number.
 */
static inline int number_16_to_32(int number)
{
	return ((number < (1 << 15)) ? number : (number - (1 << 16)));
}

/**
 * Retrieves the inode number from a given block entry in a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param off The offset of the block entry in the journal.
 * @return The inode number (nid_t) of the block entry.
 */
static inline nid_t get_fpjournal_ino(struct page *p, int off)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return le32_to_cpu(rn->fn.bes[off].inode_nid);
}

/**
 * Retrieves the parent node ID from a given block entry in a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param off The offset of the block entry in the journal.
 * @return The parent node ID (nid_t) of the block entry.
 */
static inline nid_t get_fpjournal_nid(struct page *p, int off)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return le32_to_cpu(rn->fn.bes[off].parent_nid);
}

/**
 * Retrieves the block number from a given block entry in a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param off The offset of the block entry in the journal.
 * @return The block number of the block entry.
 */
static inline int get_fpjournal_num(struct page *p, int off)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return le32_to_cpu(rn->fn.bes[off].number);
}

/**
 * Retrieves the counter from the footer of a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @return The counter value from the footer.
 */
static inline int get_fpfooter_counter(struct page *p)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return le32_to_cpu(rn->fn.fp_footer.prefix);
}

/**
 * Sets the inode number for a given block entry in a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param off The offset of the block entry in the journal.
 * @param inode_nid The new inode number to be set.
 * @return The result of setting the page dirty.
 */
static inline int set_fpjournal_ino(struct page *p, int off, nid_t inode_nid)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.bes[off].inode_nid = cpu_to_le32(inode_nid);
	return set_page_dirty(p);
}

/**
 * Sets the parent node ID for a given block entry in a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param off The offset of the block entry in the journal.
 * @param parent_nid The new parent node ID to be set.
 * @return The result of setting the page dirty.
 */
/**
 * Sets the parent node ID for a given block entry in a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param off The offset of the block entry in the journal.
 * @param parent_nid The new parent node ID to be set.
 * @return The result of setting the page dirty.
 */
static inline int set_fpjournal_nid(struct page *p, int off, nid_t parent_nid)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.bes[off].parent_nid = cpu_to_le32(parent_nid);
	return set_page_dirty(p);
}

/**
 * Sets the block number for a given block entry in a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param off The offset of the block entry in the journal.
 * @param number The new block number to be set.
 * @return The result of setting the page dirty.
 */
static inline int set_fpjournal_num(struct page *p, int off, int number)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.bes[off].number = cpu_to_le32(number);
	return set_page_dirty(p);
}

/**
 * Sets the counter in the footer of a full page journal.
 * 
 * @param p The page containing the full page journal.
 * @param count The new counter value to be set.
 * @return The result of setting the page dirty.
 */
static inline int set_fpfooter_counter(struct page *p, unsigned int count)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.fp_footer.prefix = cpu_to_le32(count);
	return set_page_dirty(p);
}

/**
 * Retrieves a page for reading data, handling various scenarios including extent cache lookup and data block address translation.
 * 
 * @param sbi The file system super block information.
 * @param inode The inode for which the data page is being retrieved.
 * @param index The page index in the inode's address space.
 * @param virtual_addr A pointer to store the virtual address of the retrieved page.
 * @param stable_result A structure to store the result of the stable bucket change.
 * @return The retrieved page or an error pointer if the operation fails.
 */
static inline struct page *f2dfs_get_read_data_page(struct f2fs_sb_info *sbi,
				struct inode *inode, pgoff_t index, virtual_t *virtual_addr,
				struct f2dfs_bucket_change *stable_result)
{
	struct address_space *mapping = inode->i_mapping;
	struct dnode_of_data dn;
	struct page *page;
	struct extent_info ei = {0, 0, 0};
	int err;
	unsigned int fp_prefix = DIRECT_FP_PREFIX;
	bool is_stable = false;

	// Attempt to grab a cached page from the inode's address space.
	page = f2fs_grab_cache_page(mapping, index, false);
	if (!page)
		return ERR_PTR(-ENOMEM);

	// Check if the page is in the extent cache and translate the block address accordingly.
	if (f2fs_lookup_extent_cache(inode, index, &ei)) {
		dn.virtual_addr = ei.blk + index - ei.fofs;
		dn.data_blkaddr = f2dfs_address_translation_offline_path(
						sbi, dn.virtual_addr, &fp_prefix, &is_stable);
		if (!f2fs_is_valid_blkaddr(sbi, dn.data_blkaddr,
					DATA_GENERIC_ENHANCE_READ)) {
			err = -EFSCORRUPTED;
			goto put_err;
		}
		goto got_it;
	}

	// Initialize a new dnode and attempt to get the dnode for the data page.
	set_new_dnode(&dn, inode, NULL, NULL, 0);
	err = f2dfs_get_dnode_of_data_offline_path(&dn, index,
					LOOKUP_NODE, &fp_prefix, &is_stable);
	if (err)
		goto put_err;

	f2fs_put_dnode(&dn);

	// Check if the data block address is valid.
	if (unlikely(dn.data_blkaddr == NULL_ADDR)) {
		err = -ENOENT;
		goto put_err;
	}

	if ((dn.data_blkaddr != NEW_ADDR) &&
				(!f2fs_is_valid_blkaddr(sbi, dn.data_blkaddr,
						DATA_GENERIC_ENHANCE))) {
		err = -EFSCORRUPTED;
		goto put_err;
	}

got_it:
	// Store the virtual address in the provided pointer.
	(*virtual_addr) = dn.virtual_addr;

	// Update the stable result structure based on the physicality of the dnode.
	if (is_real_dnode_physical(dn.virtual_addr)) {
		stable_result->fp_prefix = DIRECT_FP_PREFIX;
		stable_result->is_stable = false;
	}
	else {
		stable_result->fp_prefix = fp_prefix;
		stable_result->is_stable = is_stable;
		if (stable_result->is_stable &&
					(stable_result->fp_prefix != DIRECT_FP_PREFIX) &&
					(stable_result->fp_prefix != INVALID_FP_PREFIX))
			goto put_err;
		else
			stable_result->is_stable = false;
	}

	// If the page is up-to-date, unlock and return it.
	if (PageUptodate(page)) {
		unlock_page(page);
		return page;
	}

	// If the data block address is NEW_ADDR, zero out the page and mark it up-to-date.
	if (dn.data_blkaddr == NEW_ADDR) {
		zero_user_segment(page, 0, PAGE_SIZE);
		if (!PageUptodate(page))
			SetPageUptodate(page);
		unlock_page(page);
		return page;
	}

	// Submit the page for reading from the data block address.
	err = f2fs_submit_page_read(inode, page, dn.data_blkaddr, 0, false);

	if (err)
		goto put_err;
	return page;

put_err:
	// On error, release the page and return an error pointer.
	f2fs_put_page(page, (PageLocked(page) ? 1 : 0));
	return ERR_PTR(err);
}
/**
 * Retrieves a locked data page for the given inode and page index.
 * This function is used to get a page for read operations with additional locking.
 *
 * @param sbi The file system super block information.
 * @param inode The inode for which the data page is being retrieved.
 * @param index The page index in the inode's address space.
 * @param virtual_addr A pointer to store the virtual address of the retrieved page.
 * @param stable_result A structure to store the result of the stable bucket change.
 * @return The locked page if successful, or NULL if the stable result is not as expected.
 */
static inline struct page *f2dfs_get_lock_data_page(struct f2fs_sb_info *sbi,
				struct inode *inode, pgoff_t index, virtual_t *virtual_addr,
				struct f2dfs_bucket_change *stable_result)
{
	struct address_space *mapping = inode->i_mapping;
	struct page *page;

repeat:
	page = f2dfs_get_read_data_page(sbi, inode, index, virtual_addr, stable_result);

	if (stable_result->is_stable &&
				(stable_result->fp_prefix != DIRECT_FP_PREFIX) &&
				(stable_result->fp_prefix != INVALID_FP_PREFIX))
		return NULL;
	else
		stable_result->is_stable = false;

	if (IS_ERR(page))
		return page;

	lock_page(page);
	if (unlikely(page->mapping != mapping)) {
		f2fs_put_page(page, (PageLocked(page) ? 1 : 0));
		goto repeat;
	}

	if (unlikely(!PageUptodate(page))) {
		f2fs_put_page(page, (PageLocked(page) ? 1 : 0));
		return ERR_PTR(-EIO);
	}

	return page;
}

/**
 * Retrieves the fingerprint for a given logical block address.
 * This function is used to get the fingerprint of a page by its logical address.
 *
 * @param sbi The file system super block information.
 * @param inode The inode for which the fingerprint is being retrieved.
 * @param bidx The logical block index.
 * @param virtual_addr A pointer to store the virtual address of the retrieved page.
 * @param stable_result A structure to store the result of the stable bucket change.
 * @return True if the fingerprint is successfully retrieved, false otherwise.
 */
static inline bool f2dfs_get_fingerprint_by_logical_addr(
				struct f2fs_sb_info *sbi, struct inode *inode,
				block_t bidx, virtual_t *virtual_addr,
				struct f2dfs_bucket_change *stable_result)
{
	struct page *data_page;
	bool ret = false;

	stable_result->is_stable = false;
	data_page = f2dfs_get_read_data_page(sbi, inode, bidx, virtual_addr, stable_result);

	if (stable_result->is_stable &&
				(stable_result->fp_prefix != DIRECT_FP_PREFIX) &&
				(stable_result->fp_prefix != INVALID_FP_PREFIX)) {
		ret = true;
		goto out;
	}
	else
		stable_result->is_stable = false;

	if (IS_ERR(data_page))
		goto out;

	f2fs_put_page(data_page, 0); // put_page(page is already unlocked)
	data_page = NULL;

	data_page = f2dfs_get_lock_data_page(sbi, inode, bidx, virtual_addr, stable_result);

	if (stable_result->is_stable &&
				(stable_result->fp_prefix != DIRECT_FP_PREFIX) &&
				(stable_result->fp_prefix != INVALID_FP_PREFIX)) {
		ret = true;
		goto out;
	}
	else
		stable_result->is_stable = false;

	if (IS_ERR(data_page))
		goto out;

	if (f2dfs_page_fingerprint_update(sbi, data_page, stable_result->fingerprint) == FP_HASH_SUCCESS)
		ret = true;

	f2fs_put_page(data_page, (PageLocked(data_page) ? 1 : 0));

out:
	return ret;
}

/**
 * Checks if the offline bitmap is consistent with the given node information.
 * This function is used to verify the consistency of the offline bitmap.
 *
 * @param sbi The file system super block information.
 * @param ino The inode number.
 * @param nid The node ID.
 * @param count The expected count of bitmap entries.
 * @param number The actual count of bitmap entries.
 * @return True if the offline bitmap is consistent, false otherwise.
 */
static inline bool is_offline_bitmap_consistent(struct f2fs_sb_info *sbi,
				nid_t ino, nid_t nid, int count, int number)
{
	if (count != number) {
		f2fs_info(sbi, "offline deduplictaion not match in node %u "
				"(file %u): %u != %u", nid, ino, count, number);
		return false;
	}
	return true;
}

/**
 * Initializes the node page bitmap for offline deduplication.
 * This function sets up the bitmap for a node page based on the offline deduplication information.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @param node_page The node page to initialize the bitmap for.
 * @param number A pointer to store the count of set bits in the bitmap.
 * @param times The number of times to iterate over the node's address array.
 */
static inline void f2dfs_init_node_page_bitmap(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_offline *ofi,
				struct page *node_page, int *number, int times)
{
	struct f2fs_node *raw_node;
	__le32 *addr_array;
	bool is_inode;
	int count = 0, base = 0, i = 0;

	raw_node = F2FS_NODE(node_page);
	is_inode = IS_INODE(node_page);

	// Determine the base offset for the address array based on the inode information.
	if (is_inode) {
		if (!ofi->inode)
			base = offset_in_addr(&raw_node->i);
		else if (f2fs_has_extra_attr(ofi->inode))
			base = get_extra_isize(ofi->inode);
	}

	addr_array = blkaddr_in_node(raw_node);
	memset(&ofi->node_bitmap, 0, (FP_NODE_PAGE_BITMAP_SIZE * sizeof(__le32)));

	// Iterate over the address array and set bits in the bitmap for offline dnodes.
	for (i = 0; i < times; ++i) {
		if (!is_dnode_offline(addr_array[base + i]))
			continue;

		f2fs_set_bit(i, (char *)ofi->node_bitmap);
		++count;
	}

	// Update the number of set bits if the bitmap is not consistent.
	if (!is_offline_bitmap_consistent(sbi, ino_of_node(node_page),
				nid_of_node(node_page), count, (*number)))
		(*number) = count;
}

/**
 * Updates data block addresses using offline duplicates.
 * This function updates the block addresses in a node page with new virtual addresses from offline duplicates.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @param parent_nid The parent node ID.
 * @param new_vaddrs An array of new virtual addresses.
 * @param times The number of times to iterate over the new virtual addresses.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_update_data_blkaddr_by_offline_duplicate(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_offline *ofi,
				nid_t parent_nid, virtual_t *new_vaddrs, int times, bool umount)
{
	struct f2fs_node *raw_node;
	struct dnode_of_data dn;
	struct page *node_page;
	__le32 *addr_array;
	bool is_inode;
	int base = 0, i = 0;

	// Perform read-ahead operations for metadata and node pages.
	f2fs_ra_meta_pages(sbi, NAT_BLOCK_OFFSET(parent_nid), 1, META_NAT, true);
	f2fs_ra_node_page(sbi, parent_nid);

	// Get the node page for the parent node ID.
	node_page = f2fs_get_node_page(sbi, parent_nid);
	if (IS_ERR(node_page))
		return;

	// Check if the node page is valid and belongs to the expected parent node.
	if ((parent_nid != nid_of_node(node_page)) ||
				(ofi->inode_nid != ino_of_node(node_page))) {
		f2fs_put_page(node_page, 1);
		return;
	}

	// Initialize a new dnode and set its properties.
	set_new_dnode(&dn, ofi->inode, NULL, node_page, parent_nid);
	dn.node_changed = false;
	dn.ofs_in_node = 0;

	is_inode = IS_INODE(node_page);
	raw_node = F2FS_NODE(node_page);
	if (is_inode) {
		if (!ofi->inode)
			base = offset_in_addr(&raw_node->i);
		else if (f2fs_has_extra_attr(ofi->inode))
			base = get_extra_isize(ofi->inode);
	}
	addr_array = blkaddr_in_node(raw_node);

	// Iterate over the bitmap and update block addresses with new virtual addresses.
	for (i = 0; i < times; ++i) {
		if (!f2fs_test_bit(i, (char *)ofi->node_bitmap))
			continue;

		dn.ofs_in_node = i;
		// TODO: extent update use function f2fs_update_extent_cache_range
		f2fs_update_data_blkaddr(&dn, new_vaddrs[i], true);
	}

	// Clear the offline flag in the address array.
	for (i = 0; i < times; ++i)
		addr_array[base + i] = clear_dnode_offline(addr_array[base + i]);

	// Mark the node page as dirty and move it to the appropriate free node list.
	set_page_dirty(node_page);
	f2fs_move_node_page(node_page, (umount ? FG_GC : BG_GC));
}

/**
 * Performs offline deduplication for a single block.
 * This function finds duplicates for a single block and updates the block address if a duplicate is found.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @param start_bidx The starting block index.
 * @param parent_nid The parent node ID.
 * @param ofs_in_node The offset within the node.
 * @param new_vaddrs An array of new virtual addresses.
 * @param rc_changes A pointer to the global counter for reference count changes.
 * @return 1 if a duplicate is found, 0 otherwise.
 */
static inline int f2dfs_do_offline_for_single_block(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_offline *ofi,
				block_t start_bidx, nid_t parent_nid, int ofs_in_node,
				virtual_t *new_vaddrs, struct f2dfs_global_counter *rc_changes)
{
	block_t block_index = start_bidx + ofs_in_node;
	virtual_t dnode_va = NULL_ADDR, search_va = NULL_ADDR;
	struct f2dfs_bucket_change stable_result = {
		.virtual_addr = NULL_ADDR,
		.fp_prefix = DIRECT_FP_PREFIX,
		.is_stable = false,
		.is_persist = false,
		.is_dupl = false,
		.is_remove = false,
	};
	bool is_physical_to_virtual = false;

	new_vaddrs[ofs_in_node] = NULL_ADDR;

	// Try to get the fingerprint for the block index.
	if (!f2dfs_get_fingerprint_by_logical_addr(sbi, ofi->inode, block_index,
				&dnode_va, &stable_result))
		goto clear;

	// Check if the fingerprint is stable and has a valid prefix.
	if (stable_result.is_stable &&
				(stable_result.fp_prefix != DIRECT_FP_PREFIX) &&
				(stable_result.fp_prefix != INVALID_FP_PREFIX))
		goto clear;

	// Get the fingerprint prefix and process it for bucket cache.
	stable_result.fp_prefix =
					f2dfs_get_fingerprint_prefix(stable_result.fingerprint);
	if (f2dfs_process_single_fingerprint_for_bucket_cache(
				sbi, &stable_result, dnode_va, NULL,
				FP_BUCKET_CACHE_FOR_GET_STABLE) == FP_BUCKET_SEARCH_STABLE) {
		search_va = stable_result.virtual_addr;
		goto skip_search_list;
	}

	// Perform offline deduplication for the single fingerprint.
	search_va = f2dfs_do_offline_for_single_fingerprint(
					sbi, stable_result.fingerprint,
					parent_nid, ofs_in_node, dnode_va,
					stable_result.fp_prefix, &is_physical_to_virtual);

skip_search_list:
	// Check if a duplicate is found and update the reference count.
	if ((search_va != dnode_va) && __is_valid_data_blkaddr(search_va)) {
		if (!is_physical_to_virtual)
			f2dfs_change_reference_count_offline_dupl(sbi, dnode_va, search_va,
							stable_result.fp_prefix, rc_changes);
		new_vaddrs[ofs_in_node] = search_va;
		return 1;
	}
	else
		f2dfs_change_reference_count_offline_unique(sbi, dnode_va,
						stable_result.fp_prefix, rc_changes);

clear:
	// Clear the bit in the node bitmap if no duplicate is found.
	f2fs_clear_bit(ofs_in_node, (char *)ofi->node_bitmap);
	return 0;
}
/**
 * Performs offline deduplication for a single data node.
 * This function processes a data node to find and utilize duplicates to save space.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @param parent_nid The parent node ID.
 * @param start_bidx The starting block index.
 * @param times The number of times to iterate over the node's address array.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_do_offline_for_single_dnode(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_offline *ofi,
				nid_t parent_nid, block_t start_bidx,
				int times, bool umount)
{
	int i = 0, offline_logical_free = 0;
	virtual_t *new_vaddrs;
	struct f2dfs_global_counter rc_changes = {
		.valid_rc_change = 0,
		.unique_rc_change = 0,
		.dupl_rc_change = 0,
		.zero_rc_change = 0,
	};

	// Allocate memory for new virtual addresses.
	new_vaddrs = f2fs_kvzalloc(sbi, array_size((times + 1),
					sizeof(virtual_t)), GFP_NOFS);

	// Perform read-ahead for multiple pages.
	f2dfs_offline_mpage_multipages_readahead(sbi, ofi, start_bidx, times);

	// Iterate over the node's address array and process each block.
	for (i = 0; i < times; ++i) {
		if (!f2fs_test_bit(i, (char *)ofi->node_bitmap))
			continue;

		offline_logical_free += f2dfs_do_offline_for_single_block(
						sbi, ofi, start_bidx, parent_nid, i, new_vaddrs,
						&rc_changes);
	}

	// Update data block addresses with new virtual addresses from offline duplicates.
	f2dfs_update_data_blkaddr_by_offline_duplicate(sbi, ofi, parent_nid,
					new_vaddrs, times, umount);
	kvfree(new_vaddrs);

	// Update the count of offline found duplicates.
	ofi->offline_found_count += offline_logical_free;
	f2dfs_change_block_count_offline(sbi, &rc_changes);
}

/**
 * Initializes a single data node for offline deduplication.
 * This function prepares a data node for offline deduplication by loading the node page and initializing the bitmap.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @param nid The node ID.
 * @param number A pointer to store the count of set bits in the bitmap.
 * @param times A pointer to store the number of times to iterate over the node's address array.
 * @param start_bidx A pointer to store the starting block index.
 * @return True if the initialization is successful, false otherwise.
 */
static inline bool f2dfs_do_offline_init_single_dnode(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_offline *ofi,
				nid_t nid, int *number, int *times, int *start_bidx)
{
	struct page *node_page;

	// Perform read-ahead for metadata and node pages.
	f2fs_ra_meta_pages(sbi, NAT_BLOCK_OFFSET(nid), 1, META_NAT, true);
	f2fs_ra_node_page(sbi, nid);

	// Get the node page for the given node ID.
	node_page = f2fs_get_node_page(sbi, nid);
	if (IS_ERR(node_page))
		return false;

	// Check if the node page is valid and belongs to the expected node.
	if ((nid != nid_of_node(node_page)) ||
				(ofi->inode_nid != ino_of_node(node_page))) {
		f2fs_put_page(node_page, 1);
		return false;
	}

	// Set the times and start block index for the node.
	(*times) = ADDRS_PER_PAGE(node_page, ofi->inode);
	(*start_bidx) = f2fs_start_bidx_of_node(ofs_of_node(node_page), ofi->inode);

	// Initialize the node page bitmap for offline deduplication.
	f2dfs_init_node_page_bitmap(sbi, ofi, node_page, number, (*times));
	f2fs_put_page(node_page, 1);

	return true;
}

/**
 * Retrieves the file name for a single inode during offline deduplication.
 * This function loads the inode page and copies the file name for offline processing.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @return True if the file name is successfully retrieved, false otherwise.
 */
static inline bool f2dfs_do_offline_get_single_inode_file_name(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_offline *ofi)
{
	struct page *ipage;
	unsigned char null_name[16] = "<NULL FILENAME>",
			encrypted_name[32] = "<offline hides encrypted file>";

	// Initialize the file name to a null value.
	memcpy(&ofi->file_name, null_name, 16);

	// Get the inode page for the given inode ID.
	ipage = f2fs_get_node_page(sbi, ofi->inode_nid);
	if (IS_ERR(ipage))
		return false;

	// Check if the inode page is valid and belongs to the expected inode.
	if ((ofi->inode_nid != nid_of_node(ipage)) ||
				(ofi->inode_nid != ino_of_node(ipage))) {
		f2fs_put_page(ipage, 1);
		return false;
	}

	// Copy the file name from the inode page, handling encrypted file names.
	if (file_enc_name(ofi->inode))
		memcpy(&ofi->file_name, encrypted_name, 32);
	else
		memcpy(&ofi->file_name, F2FS_INODE(ipage)->i_name, F2FS_NAME_LEN);

	f2fs_put_page(ipage, 1);

	return true;
}

/**
 * Initializes a single inode for offline deduplication.
 * This function loads the inode and optionally retrieves the file name for offline processing.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @param get_name A flag indicating whether to retrieve the file name.
 * @return True if the initialization is successful, false otherwise.
 */
static inline bool f2dfs_do_offline_init_single_inode(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_offline *ofi, bool get_name)
{
	struct super_block *sb = sbi->sb;

	// Perform read-ahead for metadata and node pages.
	f2fs_ra_meta_pages(sbi, NAT_BLOCK_OFFSET(ofi->inode_nid),
					1, META_NAT, true);
	f2fs_ra_node_page(sbi, ofi->inode_nid);
	ofi->inode = f2fs_iget(sb, ofi->inode_nid);

	// Check if the inode is loaded successfully and is not corrupted.
	if (IS_ERR(ofi->inode) || is_bad_inode(ofi->inode)) {
		set_sbi_flag(sbi, SBI_NEED_FSCK);
		return false;
	}

	// Optionally retrieve the file name for the inode.
	if (get_name)
		f2dfs_do_offline_get_single_inode_file_name(sbi, ofi);

	return true;
}
/**
 * Performs offline deduplication for a single file.
 * This function processes a file to find and utilize duplicates to save space.
 *
 * @param sbi The file system super block information.
 * @param ofi The offline fingerprint information.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_do_offline_for_single_file(struct f2fs_sb_info *sbi,
				struct f2fs_fp_offline *ofi, bool umount)
{
	block_t start_bidx;
	nid_t temp_nid;
	int temp_number, times, i = 0;

	// Iterate over the file's nodes and process each one for offline deduplication.
	for (i = 0; i < ofi->nr_pairs; ++i) {
		// Initialize the inode for offline deduplication, skipping if not successful.
		if (!f2dfs_do_offline_init_single_inode(sbi, ofi, (i == 0)))
			continue;

		// Retrieve temporary values for the parent node ID, number of blocks, and other metadata.
		temp_nid = le32_to_cpu(ofi->entries[i].parent_nid);
		temp_number = le16_to_cpu(ofi->entries[i].number);

		// Initialize the single data node for offline deduplication, skipping if not successful.
		if (!f2dfs_do_offline_init_single_dnode(sbi, ofi, temp_nid, &temp_number, &times, &start_bidx))
			goto skip;

		// Log information about the offline deduplication process.
		f2fs_info(sbi, "offline info: file %u, node page %u, bitmap blocks %u",
				ofi->inode_nid, temp_nid, temp_number);
		f2dfs_do_offline_for_single_dnode(sbi, ofi, temp_nid, start_bidx, times, umount);

skip:
		// Release the inode and reset the offline fingerprint information.
		iput(ofi->inode);
		ofi->inode = NULL;
	}
}

/**
 * Creates offline deduplication for a single file.
 * This function sets up the necessary information to perform offline deduplication for a file.
 *
 * @param sbi The file system super block information.
 * @param buffer The fingerprint buffer.
 * @param inode_nid The inode number of the file.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_create_offline_for_single_file(
				struct f2fs_sb_info *sbi, struct f2dfs_fp_buffer *buffer,
				nid_t inode_nid, bool umount)
{
	struct f2dfs_fp_offline *ofi = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2fs_fp_offline)), GFP_NOFS);
	int i = 0;
	char units[8][4] = {{"KB"}, {"MB"}, {"GB"}, {"TB"},
					{"PB"}, {"EB"}, {"ZB"}, {"YB"}};

	// Initialize the offline fingerprint information.
	ofi->offline_found_count = 0;
	ofi->nr_pairs = 0;
	ofi->inode_nid = inode_nid;

	// Organize the buffer by node ID and populate the offline fingerprint entries.
	for (i = 0; i < buffer->valid_count; ++i) {
		if ((buffer->entries[i].inode_nid == ofi->inode_nid) &&
					(buffer->entries[i].number > 0)) {
			ofi->entries[i].parent_nid =
					buffer->entries[buffer->valid_count - i - 1].parent_nid;
			ofi->entries[i].number =
					buffer->entries[buffer->valid_count - i - 1].number;
			++ofi->nr_pairs;
		}
	}

	// Perform offline deduplication for the single file.
	f2dfs_do_offline_for_single_file(sbi, ofi, umount);

	// Log information about the offline deduplication result.
	if (ofi->offline_found_count > 0)
		f2fs_info(sbi, "100%% offline deduplication reduces %u blocks "
				"(%u %2s) for single file \"%s\" (filename) %u (ino)",
				ofi->offline_found_count, f2dfs_dedupe_calculation(
					(unsigned long)ofi->offline_found_count, 0, 0),
				units[f2dfs_dedupe_unit(
					(unsigned long)ofi->offline_found_count, 0)],
				ofi->file_name, ofi->inode_nid);
	f2fs_info(sbi, "offline deduplication is finished for file \"%s\"(%u)",
			ofi->file_name, ofi->inode_nid);
	kvfree(ofi);
}

/**
 * Destroys offline deduplication for a single buffer.
 * This function removes entries from the buffer that correspond to a specific inode.
 *
 * @param buffer The fingerprint buffer.
 * @param inode_nid The inode number of the file.
 */
static inline void f2dfs_destroy_offline_for_single_buffer(
				struct f2dfs_fp_buffer *buffer, nid_t inode_nid)
{
	int count = buffer->valid_count, i = 0, move_step = 0;

	// Shift buffer entries to overwrite the removed entries.
	if (buffer->valid_count == 0)
		return;

	for (i = 0; i < count; ++i) {
		if (move_step > i)
			return;

		if (buffer->entries[i].inode_nid == inode_nid) {
			++move_step;
			continue;
		}

		if (move_step > 0) {
			buffer->entries[i - move_step].inode_nid =
							buffer->entries[i].inode_nid;
			buffer->entries[i - move_step].parent_nid =
							buffer->entries[i].parent_nid;
			buffer->entries[i - move_step].number = buffer->entries[i].number;
		}
	}
	buffer->valid_count -= move_step;
}

/**
 * Performs offline deduplication for a single buffer.
 * This function processes a buffer to find and utilize duplicates to save space.
 *
 * @param sbi The file system super block information.
 * @param buffer The fingerprint buffer.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_do_offline_for_single_buffer(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_buffer *buffer, bool umount)
{
	nid_t inode_nid;

	// Repeat the process for each file in the buffer.
offline_single_file:
	if ((!buffer) || (buffer->valid_count == 0))
		return;

	inode_nid = buffer->entries[0].inode_nid;
	f2dfs_create_offline_for_single_file(sbi, buffer, inode_nid, umount);
	f2dfs_destroy_offline_for_single_buffer(buffer, inode_nid);

	goto offline_single_file;
}

/**
 * Performs offline deduplication for buffers cache.
 * This function processes the buffers cache to find and utilize duplicates to save space.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_do_offline_for_buffers_cache(
				struct f2fs_sb_info *sbi, bool umount)
{
	struct slist_head *head;
	struct f2dfs_fp_buffer *buffer;
	int i;

	// Iterate over the fingerprint buffers and process each one for offline deduplication.
	for (i = 0; i < FP_BUFFER_NUMBER; ++i) {
		head = &sbi->s_fingerprint_buffer->pointers[i];
		if (!head)
			goto skip;

		slist_for_each(head, &sbi->s_fingerprint_buffer->pointers[i]) {
			if (!head)
				goto skip;

			buffer = slist_entry(head, struct f2dfs_fp_buffer, b_head);
			if (!buffer)
				goto skip;

			f2dfs_do_offline_for_single_buffer(sbi, buffer, umount);
			buffer->valid_count = 0;
		}

skip:
		continue;
	}
}

/**
 * Loads journal entries into the fingerprint buffers cache for one round.
 * This function processes a single page of journal entries and loads the NIDs into the cache.
 *
 * @param sbi The file system super block information.
 * @param page The page of journal entries to process.
 * @param load_at_once A pointer to a flag indicating whether to load all entries at once.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_load_journal_to_buffers_cache_one_round(
				struct f2fs_sb_info *sbi, struct page *page, bool *load_at_once)
{
	int i = 0, num = get_fpfooter_counter(page), move_step = 0,
			ret = FP_BUFFER_LOAD_NID_SUCCESS, count = 0;

	// If there are no entries to process, return success.
	if (num == 0)
		return FP_JOURNAL_NODE_REMOVE_SUCCESS;

	// Process each journal entry in the page.
	for (i = 0; i < num; ++i) {
		if (move_step > i)
			return FP_JOURNAL_NODE_REMOVE_FAILURE;

		// Attempt to load the NID into the cache.
		ret = f2dfs_load_nid_into_cache(sbi, get_fpjournal_ino(page, i),
						get_fpjournal_nid(page, i), get_fpjournal_num(page, i));
		if (ret == FP_BUFFER_LOAD_NID_SUCCESS) {
			++move_step;
			*load_at_once = true;
			count += number_16_to_32(get_fpjournal_num(page, i));
			continue;
		}

		// If the cache is full, shift the entries to overwrite the unprocessed ones.
		if (move_step > 0) {
			set_fpjournal_ino(page, i - move_step, get_fpjournal_ino(page, i));
			set_fpjournal_nid(page, i - move_step, get_fpjournal_nid(page, i));
			set_fpjournal_num(page, i - move_step, get_fpjournal_num(page, i));
		}
	}

	// Update the footer counter and adjust the global offline block count.
	if (move_step > 0) {
		set_fpfooter_counter(page, num - move_step);
		f2dfs_global_change_offline_block_count(sbi, count, -1);
	}

	// Return success if all entries have been processed, otherwise return failure.
	return ((get_fpfooter_counter(page) == 0) ?
					FP_JOURNAL_NODE_REMOVE_SUCCESS :
					FP_JOURNAL_NODE_REMOVE_FAILURE);
}

/**
 * Checks if offline deduplication has finished.
 * This function checks if all journal pages have been processed.
 *
 * @param journal The fingerprint journal to check.
 * @return True if offline deduplication has finished, false otherwise.
 */
static inline bool is_offline_finished(struct f2dfs_fp_journal *journal)
{
	int i = 0, count = 0;

	// If the journal is empty or has no pages, return true.
	if ((!journal) || (!journal->pages[0]) || (journal->nr_pages == 0))
		return true;

	// Sum the counters of all journal pages to determine if processing is complete.
	for (i = 0; i < journal->nr_pages; ++i)
		count += get_fpfooter_counter(journal->pages[i]);

	return (count == 0);
}

/**
 * Loads journal entries into the fingerprint buffers cache.
 * This function processes all pages of journal entries and loads the NIDs into the cache.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal to process.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_load_journal_to_buffers_cache(
				struct f2fs_sb_info *sbi,
				struct f2dfs_fp_journal *journal)
{
	int i = journal->nr_pages - 1, ret = FP_JOURNAL_REMOVE_SUCCESS;
	bool finish = true, load_at_once = false;

load_again:
	finish = true;
	load_at_once = false;

	// Process each page of journal entries in reverse order.
	for (i = journal->nr_pages - 1; i >= 0; --i) {
		ret = FP_JOURNAL_NODE_REMOVE_SUCCESS;
		ret = f2dfs_load_journal_to_buffers_cache_one_round(
						sbi, journal->pages[i], &load_at_once);
		if (ret == FP_JOURNAL_NODE_REMOVE_FAILURE)
			finish = false;
	}

	// If all journal pages have been processed, return success.
	if (is_offline_finished(journal))
		return FP_JOURNAL_REMOVE_SUCCESS;

	// If processing is not finished and loading can continue, repeat the process.
	if ((!finish) && load_at_once)
		goto load_again;

	// Return success if processing is finished, otherwise return failure.
	return (finish ? FP_JOURNAL_REMOVE_SUCCESS : FP_JOURNAL_REMOVE_FAILURE);
}
/**
 * Retrieves the journal head page.
 * This function gets the first page of the fingerprint journal.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 * @return The journal head page.
 */
static inline struct page *f2dfs_get_journal_head_page(
				struct f2fs_sb_info *sbi, struct f2fs_fp_journal *journal)
{
	struct page *page;
	nid_t fp_nid = FP_JOURNAL_HEAD_NID;

	// Perform read-ahead for metadata and fingerprint pages.
	f2fs_ra_meta_pages(sbi, NAT_BLOCK_OFFSET(fp_nid), 1, META_NAT, true);
	f2fs_ra_fp_page(sbi, fp_nid);
	page = f2dfs_get_fp_page(sbi, fp_nid);

	// If the page is not available or an error occurs, create a new journal head page.
	if ((!page) || IS_ERR(page)) {
		page = f2dfs_get_new_fp_page_head(sbi, fp_nid);
		set_fpfooter_pointer(page, FP_JOURNAL_NULL);
		set_fpfooter_counter(page, 0);
		clear_fpfooter_tags(page, FP_NODE_STABLE_TAG);
		set_fpfooter_tags(page, FP_NODE_JOURNAL_TAG);
		journal->alloc[0] = 0;		// Indicate that the NID has been allocated.
	}

	return page;
}

/**
 * Retrieves the fingerprint journal.
 * This function loads the fingerprint journal pages into the journal structure.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 */
static inline void f2dfs_get_journal(struct f2fs_sb_info *sbi,
				struct f2fs_fp_journal *journal)
{
	int i = 1;
	nid_t temp_nid;

	// Get the journal head page and store it in the journal structure.
	journal->pages[0] = f2dfs_get_journal_head_page(sbi, journal);
	for (i = 1; i < FP_JOURNAL_MAX_LEN; ++i) {
		// If the current page's next pointer is null, break the loop.
		if (get_fpfooter_pointer(journal->pages[i - 1]) == FP_JOURNAL_NULL)
			break;

		temp_nid = get_fpfooter_pointer(journal->pages[i - 1]);
		f2fs_ra_meta_pages(sbi, NAT_BLOCK_OFFSET(temp_nid), 1, META_NAT, true);
		f2fs_ra_fp_page(sbi, temp_nid);
		journal->pages[i] = f2dfs_get_fp_page(sbi, temp_nid);
	}
	journal->nr_pages = i; // Store the total number of journal pages.
	journal->next_nid = get_fpfooter_pointer(journal->pages[i - 1]);
}

/**
 * Retrieves the journal head.
 * This function initializes the journal structure with the journal head page.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 */
static inline void f2dfs_get_journal_head(struct f2fs_sb_info *sbi,
				struct f2fs_fp_journal *journal)
{
	journal->pages[0] = f2dfs_get_journal_head_page(sbi, journal);
	journal->nr_pages = 1; // Set the number of pages to 1.
}

/**
 * Releases the fingerprint journal.
 * This function releases the journal pages and performs necessary cleanup.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_release_journal(struct f2fs_sb_info *sbi,
				struct f2fs_fp_journal *journal, bool umount)
{
	int i = 0;
	nid_t temp_nid;

	// Iterate over the journal pages and release each one.
	for (i = 0; i < FP_JOURNAL_MAX_LEN; ++i) {
		if (journal->pages[i] == NULL)
			continue;

		temp_nid = nid_of_node(journal->pages[i]);
		if ((i != 0) && (journal->alloc[i] == 1))
			f2fs_alloc_nid_done(sbi, temp_nid);

		// Move the fingerprint page to the appropriate free node list.
		f2fs_move_node_page(journal->pages[i], umount ? FG_GC : BG_GC);
	}
}

/**
 * Releases the journal head.
 * This function releases the journal head pages up to the specified size.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 * @param size The number of pages to release.
 * @param umount A flag indicating whether the operation is during unmount.
 */
static inline void f2dfs_release_journal_head(struct f2fs_sb_info *sbi,
				struct f2fs_fp_journal *journal, int size, bool umount)
{
	int i = 0;
	nid_t temp_nid;

	// Iterate over the specified number of journal head pages and release each one.
	for (i = 0; i < size; ++i) {
		if (journal->pages[i] == NULL)
			continue;

		temp_nid = nid_of_node(journal->pages[i]);
		if ((i != 0) && (journal->alloc[i] == 1))
			f2fs_alloc_nid_done(sbi, temp_nid);

		// Move the fingerprint page to the appropriate free node list.
		f2fs_move_node_page(journal->pages[i], umount ? FG_GC : BG_GC);
	}
}
/**
 * Inserts a single parent entry into the journal head.
 * This function adds an entry to the journal head page.
 *
 * @param sbi The file system super block information.
 * @param page The journal head page.
 * @param inode_nid The inode number of the entry.
 * @param parent_nid The parent node ID of the entry.
 * @param number The block number associated with the entry.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_insert_single_parent_into_journal_head(
				struct f2fs_sb_info *sbi, struct page *page,
				nid_t inode_nid, nid_t parent_nid, int number)
{
	int num = get_fpfooter_counter(page);

	// If the journal head page is full, return failure.
	if (num == FPS_PER_TREE_BLOCK)
		return FP_JOURNAL_NODE_INSERT_FAILURE;

	// Set the entry in the journal head page and increment the counter.
	set_fpjournal_ino(page, num, inode_nid);
	set_fpjournal_nid(page, num, parent_nid);
	set_fpjournal_num(page, num, number);
	set_fpfooter_counter(page, num + 1);

	return FP_JOURNAL_NODE_INSERT_SUCCESS;
}

/**
 * Transfers entries inside the journal.
 * This function moves entries from one journal page to another.
 *
 * @param sbi The file system super block information.
 * @param src_page The source journal page.
 * @param trg_page The target journal page.
 */
static inline void f2dfs_transfer_inside_journal(struct f2fs_sb_info *sbi,
				struct page *src_page, struct page *trg_page)
{
	int times = get_fpfooter_counter(src_page), i = 0,
					ret = FP_JOURNAL_RETURN_NOTHING;

	// Transfer each entry from the source page to the target page.
	for (i = 0; i < times; ++i) {
		ret = f2dfs_insert_single_parent_into_journal_head(sbi, trg_page,
						get_fpjournal_ino(src_page, i),
						get_fpjournal_nid(src_page, i),
						get_fpjournal_num(src_page, i));
	}
	// Reset the counter of the source page.
	set_fpfooter_counter(src_page, 0);
}

/**
 * Allocates a single journal node.
 * This function allocates a new journal page and adds it to the journal.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_alloc_single_journal_node(struct f2fs_sb_info *sbi,
				struct f2fs_fp_journal *journal)
{
	struct page *page;
	int i;

	// Ensure the journal head node ID is correct.
	f2fs_bug_on(sbi, nid_of_node(journal->pages[0]) != FP_JOURNAL_HEAD_NID);

	// If the journal has reached its maximum length, return failure.
	if (journal->nr_pages >= FP_JOURNAL_MAX_LEN)
		return FP_JOURNAL_NODE_ALLOC_FAILURE;
	page = f2dfs_get_new_fp_page(sbi);
	if ((!page) || (PTR_ERR(page) > 0)) {
		page = NULL;
		return FP_JOURNAL_NODE_ALLOC_FAILURE;
	}

	// Increment the journal's page count and add the new page.
	++journal->nr_pages;
	for (i = journal->nr_pages - 1; i >= 1; --i) {
		journal->pages[i] = journal->pages[i - 1];
		journal->alloc[i] = journal->alloc[i - 1];
	}
	journal->pages[1] = page;
	journal->alloc[1] = 1;

	// Set the pointers and counters for the new journal page.
	set_fpfooter_pointer(journal->pages[1],
					get_fpfooter_pointer(journal->pages[0]));
	set_fpfooter_pointer(journal->pages[0], nid_of_node(journal->pages[1]));
	set_fpfooter_counter(journal->pages[1], 0);

	// Transfer entries from the old first page to the new first page.
	f2dfs_transfer_inside_journal(sbi,
			journal->pages[0], journal->pages[1]);

	return FP_JOURNAL_NODE_ALLOC_SUCCESS;
}

/**
 * Inserts a single parent into the journal.
 * This function inserts an entry into the journal, allocating new pages if necessary.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 * @param inode_nid The inode number of the entry.
 * @param parent_nid The parent node ID of the entry.
 * @param number The block number associated with the entry.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_insert_single_parent_into_journal(
				struct f2fs_sb_info *sbi, struct f2fs_fp_journal *journal,
				nid_t inode_nid, nid_t parent_nid, int number)
{
	int ret = FP_JOURNAL_RETURN_NOTHING, i = 0, times = 1;

	// Check for invalid input values and return success if they are not valid.
	if ((number == 0) || (inode_nid < F2DFS_START_FREE_NID) ||
				(f2fs_check_nid_range(sbi, inode_nid) != 0) ||
				(parent_nid < F2DFS_START_FREE_NID) ||
				(f2fs_check_nid_range(sbi, parent_nid) != 0))
		return FP_JOURNAL_INSERT_SUCCESS;

insert_redo:
	times = journal->nr_pages;
	for (i = 0; i < times; ++i) {
		ret = f2dfs_insert_single_parent_into_journal_head(sbi,
						journal->pages[i], inode_nid, parent_nid, number);
		if (ret == FP_JOURNAL_NODE_INSERT_SUCCESS)
			return FP_JOURNAL_INSERT_SUCCESS;
	}

	// If the journal is full, allocate a new page and retry the insertion.
	ret = f2dfs_alloc_single_journal_node(sbi, journal);
	if (ret == FP_JOURNAL_NODE_ALLOC_FAILURE)
		return FP_JOURNAL_INSERT_FAILURE;
	else
		goto insert_redo;

	return FP_JOURNAL_INSERT_FAILURE;
}

/**
 * Checks if the compaction of the fingerprint journal is finished.
 * This function determines if there are enough entries in the journal pages
 * to consider compaction as finished.
 *
 * @param journal The fingerprint journal structure.
 * @return True if compaction is finished, false otherwise.
 */
static inline bool is_compact_finished(struct f2dfs_fp_journal *journal)
{
	int i = 1, count = 0;

	// If there are very few pages, compaction is considered finished.
	if ((journal->nr_pages == 0) || (journal->nr_pages == 1))
		return true;

	// Count the number of entries in all pages except the first one.
	for (i = 1; i < journal->nr_pages; ++i)
		count += get_fpfooter_counter(journal->pages[i]);

	// If the count is more than the total capacity minus two pages, compaction is finished.
	return (count > (journal->nr_pages - 2) * FPS_PER_TREE_BLOCK);
}

/**
 * Frees a single journal node.
 * This function frees a specific journal page and updates the journal structure.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 * @param k The index of the journal page to free.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_free_single_journal_node(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_journal *journal, int k)
{
	int i, err = 0;

	// Ensure the journal head node ID is correct.
	f2fs_bug_on(sbi, nid_of_node(journal->pages[0]) != FP_JOURNAL_HEAD_NID);

	// Check if the journal page index is valid.
	if ((journal->nr_pages <= 0) || (k >= journal->nr_pages))
		return FP_JOURNAL_NODE_FREE_FAILURE;

	// If the journal has only one page and it's the one to be freed, reset the page.
	if ((k == 0) && (journal->nr_pages == 1)) {
		set_fpfooter_counter(journal->pages[0], 0);
		set_fpfooter_pointer(journal->pages[0], FP_JOURNAL_NULL);
		return FP_JOURNAL_NODE_FREE_SUCCESS;
	}

	// If the journal has more than one page and the first page is to be freed,
	// transfer entries from the second page to the first.
	if ((k == 0) && (journal->nr_pages > 1)) {
		f2dfs_transfer_inside_journal(sbi, journal->pages[1], journal->pages[0]);
		k = 1;
	}

	// Update the pointer of the previous page to bypass the freed page.
	set_fpfooter_pointer(journal->pages[k - 1], get_fpfooter_pointer(journal->pages[k]));

	// Truncate the fingerprint page and free it.
	err = f2dfs_truncate_fp_page(sbi, journal->pages[k], journal->alloc[k]);
	if (err)
		return FP_JOURNAL_NODE_FREE_FAILURE;

	// Update the journal structure to remove the freed page.
	journal->pages[k] = NULL;
	for (i = k; i < journal->nr_pages - 1; ++i) {
		journal->pages[i] = journal->pages[i + 1];
		journal->alloc[i] = journal->alloc[i + 1];
	}
	journal->pages[journal->nr_pages - 1] = NULL;
	--journal->nr_pages;

	return FP_JOURNAL_NODE_FREE_SUCCESS;
}

/**
 * Frees zero-count journal nodes.
 * This function frees journal pages that have no entries.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 */
static inline void f2dfs_free_zero_journal_nodes(struct f2fs_sb_info *sbi,
				struct f2fs_fp_journal *journal)
{
	int i = 1;

	// If the journal has one or fewer pages, there's nothing to do.
	if (journal->nr_pages <= 1)
		return;

	// Iterate over the journal pages and free any that have no entries.
	while (i < journal->nr_pages) {
		if (get_fpfooter_counter(journal->pages[i]) == 0) {
			f2dfs_free_single_journal_node(sbi, journal, i);
			continue;
		}
		++i;
	}
}

/**
 * Compacts the fingerprint journal.
 * This function compacts the journal by moving entries and freeing empty pages.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 */
static inline void f2dfs_compact_journal(struct f2fs_sb_info *sbi,
				struct f2fs_fp_journal *journal)
{
	int times, src_index, trg_index, i = 0, ret = FP_JOURNAL_RETURN_NOTHING;

	// Free any zero-count journal nodes.
	f2dfs_free_zero_journal_nodes(sbi, journal);

release_tail:
	// If compaction is finished, return.
	if (is_compact_finished(journal))
		return;

	// Determine the source and target indices for compaction.
	src_index = journal->nr_pages - 1;
	trg_index = journal->nr_pages - 2;
	times = get_fpfooter_counter(journal->pages[src_index]);

	// Move entries from the source page to the target page.
	for (i = 0; i < times; ++i) {
insert_redo:
		ret = FP_JOURNAL_RETURN_NOTHING;
		ret = f2dfs_insert_single_parent_into_journal_head(
						sbi, journal->pages[trg_index],
						get_fpjournal_ino(journal->pages[src_index], i),
						get_fpjournal_nid(journal->pages[src_index], i),
						get_fpjournal_num(journal->pages[src_index], i));
		if (ret != FP_JOURNAL_NODE_INSERT_SUCCESS) {
			--trg_index;
			goto insert_redo;
		}
	}

	// If all entries have been moved, free the source page.
	if (i >= times - 1)
		f2dfs_free_single_journal_node(sbi, journal, src_index);
	else {
		set_fpfooter_counter(journal->pages[src_index], times - i - 1);
		return;
	}

	goto release_tail;
}

/**
 * Persists a single buffer to the journal.
 * This function inserts buffer entries into the journal and updates the offline block count.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 * @param buffer The fingerprint buffer to persist.
 */
static inline void f2dfs_persist_single_buffer_to_journal(
				struct f2fs_sb_info *sbi, struct f2fs_fp_journal *journal,
				struct f2dfs_fp_buffer *buffer)
{
	int ret = FP_JOURNAL_RETURN_NOTHING, i = 0, count = 0;

	// Iterate over the buffer entries and insert them into the journal.
	for (i = 0; i < buffer->valid_count; ++i) {
		ret = FP_JOURNAL_RETURN_NOTHING;
		ret = f2dfs_insert_single_parent_into_journal(sbi, journal,
						buffer->entries[i].inode_nid,
						buffer->entries[i].parent_nid,
						buffer->entries[i].number);
		if (ret == FP_JOURNAL_INSERT_SUCCESS)
			count += number_16_to_32(buffer->entries[i].number);
	}
	// Reset the buffer's valid count and update the offline block count.
	buffer->valid_count = 0;
	f2dfs_global_change_offline_block_count(sbi, count, 1);
}

/**
 * Persists buffers cache to journal.
 * This function iterates over the fingerprint buffers and persists them to the journal.
 *
 * @param sbi The file system super block information.
 * @param journal The fingerprint journal structure.
 */
static inline void f2dfs_persist_buffers_cache_to_journal(
				struct f2fs_sb_info *sbi, struct f2fs_fp_journal *journal)
{
	struct slist_head *head;
	struct f2dfs_fp_buffer *buffer;
	int i;

	// Iterate over the fingerprint buffer cache.
	for (i = 0; i < FP_BUFFER_NUMBER; ++i) {
		head = &sbi->s_fingerprint_buffer->pointers[i];
		if (!head)
			goto skip;

		// Iterate over each buffer in the list and persist it to the journal.
		slist_for_each(head, &sbi->s_fingerprint_buffer->pointers[i]) {
			if (!head)
				goto skip;

			buffer = slist_entry(head, struct f2dfs_fp_buffer, b_head);
			if (!buffer)
				goto skip;

			// Persist the buffer to the journal if it's marked for persistence.
			if (should_persist_single_buffer_umount(buffer)) {
				f2dfs_persist_single_buffer_to_journal(sbi, journal, buffer);
				buffer->valid_count = 0;
			}
		}

skip:
		continue;
	}
}
/**
 * Synchronizes fingerprint buffers cache in the background.
 * This function synchronizes the fingerprint buffers cache to the journal,
 * handling various conditions such as unmount, ioctl, and offline deduplication needs.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 * @param ioctl A flag indicating whether the operation is triggered by an ioctl call.
 * @param need_offline A flag indicating whether offline deduplication is needed.
 * @param finish A pointer to a flag that indicates whether the synchronization is finished.
 * @return True if the synchronization is successful, false otherwise.
 */
static inline bool f2dfs_sync_buffers_cache_background(
				struct f2fs_sb_info *sbi, bool umount,
				bool ioctl, bool need_offline, bool *finish)
{
	struct f2dfs_fp_journal *journal = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2fs_fp_journal)), GFP_NOFS);
	bool ret = false;

	// Allocate memory for journal pages.
	journal->pages = f2fs_kvzalloc(sbi, array_size(FP_JOURNAL_MAX_LEN,
					sizeof(struct page *)), GFP_NOFS);
	journal->nr_pages = 0;
	memset(journal->alloc, 0, FP_JOURNAL_MAX_LEN);

	// If there's not enough space for journal, set finish to true and return false.
	if (has_not_enough_space_for_journal(sbi)) {
		ret = false;
		(*finish) = true;
		goto out;
	}

	// Get the journal and persist buffers cache to journal.
	f2dfs_get_journal(sbi, journal);
	f2dfs_persist_buffers_cache_to_journal(sbi, journal);

	// If unmounting, IOCTL called, or offline deduplication is needed, load journal to buffers cache.
	if (umount || ioctl || need_offline)
		ret = (f2dfs_load_journal_to_buffers_cache(sbi, journal) ==
						FP_JOURNAL_REMOVE_SUCCESS);
	if (is_offline_finished(journal)) {
		f2dfs_global_change_offline_block_count(sbi, 0, 0);
		(*finish) = true;
	}

	// Compact the journal and release it.
	f2dfs_compact_journal(sbi, journal);
	f2dfs_release_journal(sbi, journal, umount);

out:
	kvfree(journal->pages);
	kvfree(journal);

	return ret;
}

/**
 * Performs offline deduplication with journal.
 * This function performs offline deduplication, using the journal to track progress.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 * @param ioctl A flag indicating whether the operation is triggered by an ioctl call.
 * @param urgent A flag indicating whether the operation is urgent.
 * @param need_offline A flag indicating whether offline deduplication is needed.
 * @return True if the offline deduplication is finished, false otherwise.
 */
static inline bool f2dfs_do_offline_deduplication_with_journal(
				struct f2fs_sb_info *sbi, bool umount,
				bool ioctl, bool urgent, bool need_offline)
{
	bool finish = false;

	// If there's not enough space for offline deduplication, return false.
	if (has_not_enough_space_for_offline(sbi))
		return finish;

	// Loop until offline deduplication is finished.
	while (!finish) {
		if (offline_user_blocks(sbi) == 0) {
			finish = true;
			break;
		}
		f2dfs_sync_buffers_cache_background(sbi,
						umount, ioctl, need_offline, &finish);
		f2dfs_do_offline_for_buffers_cache(sbi, umount);
	}

	f2dfs_do_offline_for_buffers_cache(sbi, umount);
	sbi->last_time[FP_OFFLINE_TIME] = jiffies;

	return finish;
}

/**
 * Performs offline deduplication.
 * This function decides whether to perform offline deduplication based on various conditions.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 * @param ioctl A flag indicating whether the operation is triggered by an ioctl call.
 * @param urgent A flag indicating whether the operation is urgent.
 * @return True if the operation is successful, false otherwise.
 */
bool f2dfs_do_offline_deduplication(struct f2fs_sb_info *sbi,
				bool umount, bool ioctl, bool urgent)
{
	bool need_offline = (umount || ioctl || urgent ||
			should_dedupe_offline_normal(sbi));

	// If offline deduplication is not needed and buffers need to be persisted, return true.
	if ((!need_offline) && should_persist_buffers_normal(sbi))
		return true;

	// Flush stable fingerprint nodes before proceeding.
	f2fs_flush_merged_writes(sbi);

	// Perform offline deduplication for buffers cache.
	f2dfs_do_offline_for_buffers_cache(sbi, umount);
	return true;

	// Optionally, perform offline deduplication with journal.
	// return f2dfs_do_offline_deduplication_with_journal(
	// 				sbi, umount, ioctl, urgent, need_offline);
}

/**
 * Synchronizes buffers cache via ioctl.
 * This function synchronizes the buffers cache in response to an ioctl call.
 *
 * @param sbi The file system super block information.
 * @return True if the synchronization is successful, false otherwise.
 */
bool f2dfs_ioc_sync_buffers_cache(struct f2fs_sb_info *sbi)
{
	struct f2dfs_fp_journal *journal = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2fs_fp_journal)), GFP_NOFS);
	bool ret = true;

	// Allocate memory for journal pages.
	journal->pages = f2fs_kvzalloc(sbi, array_size(FP_JOURNAL_MAX_LEN,
					sizeof(struct page *)), GFP_NOFS);
	journal->nr_pages = 0;
	memset(journal->alloc, 0, FP_JOURNAL_MAX_LEN);

	// If there's not enough space for journal, set ret to false and skip synchronization.
	if (has_not_enough_space_for_journal(sbi)) {
		ret = false;
		goto out;
	}

	// Get the journal, persist buffers cache to journal, compact the journal, and release it.
	f2dfs_get_journal(sbi, journal);
	f2dfs_persist_buffers_cache_to_journal(sbi, journal);

	f2dfs_compact_journal(sbi, journal);
	f2dfs_release_journal(sbi, journal, false);

out:
	kvfree(journal->pages);
	kvfree(journal);

	return ret;
}

/**
 * Synchronizes a single buffer in the foreground.
 * This function synchronizes a single buffer to the journal.
 *
 * @param sbi The file system super block information.
 * @param buffer The fingerprint buffer to synchronize.
 */
void f2dfs_sync_single_buffer_foreground(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_buffer *buffer)
{
	struct f2dfs_fp_journal *journal = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2fs_fp_journal)), GFP_NOFS);
	int size = 2;

	// Allocate memory for journal pages.
	journal->pages = f2fs_kvzalloc(sbi, array_size(size,
					sizeof(struct page *)), GFP_NOFS);
	journal->nr_pages = 0;
	memset(journal->alloc, 0, size);

	// If there's not enough space for a single journal, skip synchronization.
	if (has_not_enough_space_for_single_journal(sbi, buffer))
		goto out;

	// Get the journal head and release the journal head.
	f2dfs_get_journal_head(sbi, journal);
	f2dfs_release_journal_head(sbi, journal, size, false);
	buffer->valid_count = 0;

out:
	kvfree(journal->pages);
	kvfree(journal);
}
