#include <linux/fs.h>
#include "f2dfs_fs.h"
#include <linux/mpage.h>
#include <linux/backing-dev.h>
#include <linux/blkdev.h>
#include <linux/pagevec.h>
#include <linux/swap.h>

#include "f2fs.h"
#include "node.h"
#include "segment.h"
#include "xattr.h"
#include "trace.h"
#include <trace/events/f2fs.h>


/**
 * Copies a fingerprint into an array of addresses.
 * This function takes a fingerprint and copies it into an array of __le32 types.
 *
 * @param fp The fingerprint to copy.
 * @param fp_addrs The array of addresses to copy the fingerprint into.
 */
static inline void fingerprint_to_addrs(__u8 fp[], __le32 *fp_addrs)
{
	int i = 0;
	for (i = 0; i < FP_LENGTH_ADDR; ++i)
		memcpy((fp_addrs + i), (fp + (i * sizeof(__le32))), sizeof(__le32));
}

/**
 * Copies an array of addresses into a fingerprint.
 * This function takes an array of addresses and copies it into a fingerprint.
 *
 * @param fp The fingerprint to copy into.
 * @param fp_addrs The array of addresses to copy.
 */
static inline void addrs_to_fingerprint(__u8 fp[], __le32 *fp_addrs)
{
	int i = 0;
	for (i = 0; i < FP_LENGTH_ADDR; ++i)
		memcpy((fp + (i * sizeof(__le32))), (fp_addrs + i), sizeof(__le32));
}

/**
 * Retrieves the footer counter from a page.
 * This function gets the valid count from the footer of a fingerprint page.
 *
 * @param p The page to retrieve the counter from.
 * @return The footer counter.
 */
static inline int get_fpfooter_counter(struct page *p)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return le32_to_cpu(rn->fn.fp_footer.valid_count);
}

/**
 * Retrieves the footer prefix from a page.
 * This function gets the prefix from the footer of a fingerprint page.
 *
 * @param p The page to retrieve the prefix from.
 * @return The footer prefix.
 */
static inline unsigned int get_fpfooter_prefix(struct page *p)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return (le32_to_cpu(rn->fn.fp_footer.prefix) & FP_PREFIX_MASK_BIT);
}

/**
 * Sets the footer counter in a page.
 * This function sets the valid count in the footer of a fingerprint page.
 *
 * @param p The page to set the counter in.
 * @param count The new counter value.
 * @return The result of setting the page dirty.
 */
static inline int set_fpfooter_counter(struct page *p, int count)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.fp_footer.valid_count = cpu_to_le32(count);
	return set_page_dirty(p);
}

/**
 * Sets the footer prefix in a page.
 * This function sets the prefix in the footer of a fingerprint page.
 *
 * @param p The page to set the prefix in.
 * @param fp_prefix The new prefix value.
 * @return The result of setting the page dirty.
 */
static inline int set_fpfooter_prefix(struct page *p, unsigned int fp_prefix)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.fp_footer.prefix = cpu_to_le32(fp_prefix & FP_PREFIX_MASK_BIT);
	return set_page_dirty(p);
}

/**
 * Retrieves the virtual address from a fingerprint list.
 * This function gets the virtual address at a specific offset in a fingerprint list.
 *
 * @param p The page containing the fingerprint list.
 * @param off The offset in the list.
 * @return The virtual address.
 */
static inline virtual_t get_fplist_vaddr(struct page *p, int off)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	return le32_to_cpu(rn->fn.nes[off].virtual_addr);
}

/**
 * Retrieves the fingerprint from a fingerprint list.
 * This function gets the fingerprint at a specific offset in a fingerprint list.
 *
 * @param p The page containing the fingerprint list.
 * @param off The offset in the list.
 * @param fp The fingerprint to copy into.
 */
static inline void get_fplist_fingerprint(struct page *p, int off, __u8 fp[])
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	__le32 *fp_addrs;
	int i;

	fp_addrs = f2fs_kzalloc(F2FS_P_SB(p), FP_LEN_MAX, GFP_NOFS);
	for (i = 0; i < FP_LENGTH_ADDR; ++i)
		(*(fp_addrs + i)) = cpu_to_le32(rn->fn.nes[off].fp_part[i]);
	addrs_to_fingerprint(fp, fp_addrs);
	kfree(fp_addrs);
}

/**
 * Sets the virtual address in a fingerprint list.
 * This function sets the virtual address at a specific offset in a fingerprint list.
 *
 * @param p The page containing the fingerprint list.
 * @param off The offset in the list.
 * @param virtual_addr The new virtual address.
 * @return The result of setting the page dirty.
 */
static inline int set_fplist_vaddr(struct page *p, int off, virtual_t virtual_addr)
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	rn->fn.nes[off].virtual_addr = cpu_to_le32(virtual_addr);
	return set_page_dirty(p);
}

/**
 * Sets the fingerprint in a fingerprint list.
 * This function sets the fingerprint at a specific offset in a fingerprint list.
 *
 * @param p The page containing the fingerprint list.
 * @param off The offset in the list.
 * @param fp The fingerprint to set.
 * @return The result of setting the page dirty.
 */
static inline int set_fplist_fingerprint(struct page *p, int off, __u8 fp[])
{
	struct f2fs_node *rn = F2DFS_FP_NODE(p);
	__le32 *fp_addrs;
	int i;

	fp_addrs = f2fs_kzalloc(F2FS_P_SB(p), FP_LEN_MAX, GFP_NOFS);
	f2fs_wait_on_page_writeback(p, NODE, true, true);
	fingerprint_to_addrs(fp, fp_addrs);
	for (i = 0; i < FP_LENGTH_ADDR; ++i)
		rn->fn.nes[off].fp_part[i] = cpu_to_le32(*(fp_addrs + i));
	kfree(fp_addrs);
	return set_page_dirty(p);
}

/**
 * Prints a fingerprint node.
 * This function prints the fingerprints and virtual addresses in a fingerprint node.
 *
 * @param page The page containing the fingerprint node.
 */
static inline void fptree_node_print(struct page *page)
{
	int times = get_fpfooter_counter(page), i = 0, j = 0;
	__u8 fp[FP_LEN_MAX];
	virtual_t vaddr = NULL_ADDR;

	for (i = 0; i < times; ++i) {
		get_fplist_fingerprint(page, i, fp);
		vaddr = get_fplist_vaddr(page, i);
		printk(" ");
		for(j = 0; j < FP_LEN_MAX; ++j)
			printk(KERN_CONT "%02X",
					fp[j]);
		printk(KERN_CONT " %u",
				vaddr);
	}
}

/**
 * Prints a fingerprint list.
 * This function prints the fingerprint list, including node IDs and fingerprints.
 *
 * @param list The fingerprint list to print.
 * @param is_stable A flag indicating whether the list is stable.
 */
static inline void fptree_list_print(struct f2dfs_fp_list *list, bool is_stable)
{
	int i = 0;

	is_stable ? printk("stable list") : printk("unstable list");
	printk(" ");
	printk(KERN_CONT "nid list: %u",
			nid_of_node(list->pages[0]));
	for (i = 1; i < list->nr_pages; ++i)
		printk(KERN_CONT " -> %u",
				nid_of_node(list->pages[i]));
	for (i = 0; i < list->nr_pages; ++i)
		fptree_node_print(list->pages[i]);
}

/*
 * There are two list for single same prefix:
 * 1. Unstable list: append (remove entry & persist entry) from bucket.
 * 2. Stable list: unique stabilize from unstable list, be used to offline
 * deduplication, before offline process, unstable list must be cleared.
 */

static inline struct page *f2dfs_get_list_head_page(
				struct f2fs_sb_info *sbi,
				struct f2dfs_fp_list *list, bool is_stable)
{
	struct page *page;
	nid_t fp_nid = (is_stable ? FP_NID_STABLE(list->index) :
					FP_NID_UNSTABLE(list->index));

	f2fs_ra_meta_pages(sbi, NAT_BLOCK_OFFSET(fp_nid), 1, META_NAT, true);
	f2dfs_ra_fp_page(sbi, fp_nid);
	page = f2dfs_get_fp_page(sbi, fp_nid);

	if ((!page) || IS_ERR(page)) {
		page = f2dfs_get_new_fp_page_head(sbi, fp_nid);
		set_fpfooter_pointer(page, FP_LIST_NULL);
		set_fpfooter_counter(page, 0);
		if (is_stable)
			set_fpfooter_tags(page, FP_NODE_STABLE_TAG);
		else
			clear_fpfooter_tags(page, FP_NODE_STABLE_TAG);
		clear_fpfooter_tags(page, FP_NODE_JOURNAL_TAG);
		set_fpfooter_prefix(page, list->index);
		list->alloc[0] = 0;		// pass f2fs_alloc_nid(_done)
	}

	return page;
}

static inline void f2dfs_get_single_list(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_list *list, bool is_stable)
{
	int i = 1;
	nid_t temp_nid;

	list->pages[0] = f2dfs_get_list_head_page(sbi, list, is_stable);
	for (i = 1; i < FP_LIST_MAX_LEN; ++i) {
		if (get_fpfooter_pointer(list->pages[i - 1]) == FP_LIST_NULL)
			break;

		temp_nid = get_fpfooter_pointer(list->pages[i - 1]);
		f2fs_ra_meta_pages(sbi, NAT_BLOCK_OFFSET(temp_nid), 1, META_NAT, true);
		f2dfs_ra_fp_page(sbi, temp_nid);
		list->pages[i] = f2dfs_get_fp_page(sbi, temp_nid);
	}
	list->nr_pages = i;
}

static inline void f2dfs_get_single_list_head(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_list *list, bool is_stable)
{
	list->pages[0] = f2dfs_get_list_head_page(sbi, list, is_stable);
	list->nr_pages = 1;
}

static inline void f2dfs_release_single_list(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_list *list, bool is_stable, bool umount)
{
	int i = 0;
	nid_t temp_nid;

	if (!is_stable) {
		if ((list->nr_pages > 0) && (get_fpfooter_counter(list->pages[0]) > 0))
			f2fs_set_bit(list->index,
							(char *)sbi->s_fingerprint_bucket->s_stable_bitmap);
		else
			f2fs_clear_bit(list->index,
							(char *)sbi->s_fingerprint_bucket->s_stable_bitmap);
	}

	for (i = 0; i < FP_LIST_MAX_LEN; ++i) {
		if (list->pages[i] == NULL)
			continue;

		temp_nid = nid_of_node(list->pages[i]);
		if ((i != 0) && (list->alloc[i] == 1))
			f2fs_alloc_nid_done(sbi, temp_nid);

		// f2dfs_move_fp_page(list->pages[i], umount);
		f2fs_move_node_page(list->pages[i], umount ? FG_GC : BG_GC);

		// if (PageDirty(list->pages[i])) {
		// 	f2dfs_move_fp_page(list->pages[i], umount);
		// }
		// else {
		// 	ClearPageUptodate(list->pages[i]);
		// 	if (PageLocked(list->pages[i]))
		// 		f2fs_put_page(list->pages[i], 1);
		// 	invalidate_mapping_pages(FP_MAPPING(sbi), temp_nid, temp_nid);
		// 	invalidate_mapping_pages(NODE_MAPPING(sbi), temp_nid, temp_nid);
		// }
	}
}

static inline void f2dfs_release_single_list_head(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_list *list, int size, bool umount)
{
	int i = 0;
	nid_t temp_nid;

	if ((size > 1) || ((list->nr_pages > 0) &&
				(get_fpfooter_counter(list->pages[0]) > 0)))
		f2fs_set_bit(list->index,
						(char *)sbi->s_fingerprint_bucket->s_stable_bitmap);

	for (i = 0; i < size; ++i) {
		if (list->pages[i] == NULL)
			continue;

		temp_nid = nid_of_node(list->pages[i]);
		if ((i != 0) && (list->alloc[i] == 1))
			f2fs_alloc_nid_done(sbi, temp_nid);


		// f2dfs_move_fp_page(list->pages[i], umount);
		f2fs_move_node_page(list->pages[i], umount ? FG_GC : BG_GC);

		// if (PageDirty(list->pages[i])) {
		// 	f2dfs_move_fp_page(list->pages[i], umount);
		// }
		// else {
		// 	ClearPageUptodate(list->pages[i]);
		// 	if (PageLocked(list->pages[i]))
		// 		f2fs_put_page(list->pages[i], 1);
		// 	invalidate_mapping_pages(FP_MAPPING(sbi), temp_nid, temp_nid);
		// 	invalidate_mapping_pages(NODE_MAPPING(sbi), temp_nid, temp_nid);
		// }
	}
}

/**
 * Removes a single virtual address from a node.
 * This function removes a virtual address from a fingerprint node.
 *
 * @param sbi The file system super block information.
 * @param page The page containing the node.
 * @param remove_vaddr The virtual address to remove.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_remove_single_vaddr_from_single_node(
				struct f2fs_sb_info *sbi, struct page *page,
				virtual_t remove_vaddr)
{
	int num = get_fpfooter_counter(page), i = 0;
	__u8 node_fp[FP_LEN_MAX];
	bool hit = false;

	// If the node is empty, return failure.
	if (num == 0)
		return FP_LIST_NODE_REMOVE_FAILURE;

	// Find the virtual address to remove.
	for (i = 0; i < num; ++i) {
		if (get_fplist_vaddr(page, i) == remove_vaddr) {
			hit = true;
			break;
		}
	}

	// If the virtual address is not found, return failure.
	if (!hit)
		return FP_LIST_NODE_REMOVE_FAILURE;

	// Shift the remaining entries to fill the gap.
	for (; i < num - 1; ++i) {
		get_fplist_fingerprint(page, i + 1, node_fp);
		set_fplist_fingerprint(page, i, node_fp);
		set_fplist_vaddr(page, i, get_fplist_vaddr(page, i + 1));
	}
	set_fpfooter_counter(page, num - 1);

	return FP_LIST_NODE_REMOVE_SUCCESS;
}

/**
 * Inserts a fingerprint into a stable node.
 * This function inserts a fingerprint into a stable fingerprint node.
 *
 * @param sbi The file system super block information.
 * @param page The page containing the node.
 * @param fp The fingerprint to insert.
 * @param insert_vaddr The virtual address to insert.
 * @param fp_prefix The fingerprint prefix.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_insert_single_fingerprint_into_single_stable_node(
				struct f2fs_sb_info *sbi, struct page *page, __u8 fp[],
				virtual_t insert_vaddr, unsigned int fp_prefix)
{
	int left = 0, right = get_fpfooter_counter(page) - 1, middle = 0, i = 0;
	__u8 node_fp[FP_LEN_MAX];
	struct f2dfs_bucket_change stable_change = {
		.virtual_addr = NULL_ADDR,
		.fp_prefix = DIRECT_FP_PREFIX,
		.is_stable = false,
		.is_persist = false,
		.is_dupl = false,
		.is_remove = false,
	};

	// If the node is full, return failure.
	if (right == FPS_PER_TREE_BLOCK - 1)
		return FP_LIST_NODE_INSERT_FAILURE;

	// Find the correct position to insert the fingerprint.
	while (left <= right) {
		middle = (left + right) / 2;
		get_fplist_fingerprint(page, middle, node_fp);

		if (f2dfs_fp_is_equal(fp, node_fp) > 0)
			left = middle + 1;
		else if (f2dfs_fp_is_equal(fp, node_fp) < 0)
			right = middle - 1;
		else
			return FP_LIST_NODE_INSERT_FAILURE;
	}

	// Shift the existing entries to make room for the new fingerprint.
	if (left == (middle + 1))
		++middle;

	for (i = get_fpfooter_counter(page) - 1; i >= middle; --i) {
		get_fplist_fingerprint(page, i, node_fp);
		set_fplist_fingerprint(page, i + 1, node_fp);
		set_fplist_vaddr(page, i + 1, get_fplist_vaddr(page, i));
	}
	set_fplist_fingerprint(page, middle, fp);
	set_fplist_vaddr(page, middle, insert_vaddr);
	set_fpfooter_counter(page, get_fpfooter_counter(page) + 1);

	// Update the stable bucket cache.
	stable_change.is_stable = true;
	stable_change.is_persist = true;
	stable_change.fp_prefix = fp_prefix;
	stable_change.virtual_addr = insert_vaddr;
	memcpy(&stable_change.fingerprint, fp, FP_LEN_MAX);
	f2dfs_process_single_fingerprint_for_bucket_cache(sbi,
					&stable_change, NULL_ADDR, NULL,
					FP_BUCKET_CACHE_FOR_SET_STABLE);

	return FP_LIST_NODE_INSERT_SUCCESS;
}

/**
 * Inserts a fingerprint into an unstable node.
 * This function inserts a fingerprint into an unstable fingerprint node.
 *
 * @param sbi The file system super block information.
 * @param page The page containing the node.
 * @param fp The fingerprint to insert.
 * @param insert_vaddr The virtual address to insert.
 * @param fp_prefix The fingerprint prefix.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_insert_single_fingerprint_into_single_unstable_node(
				struct f2fs_sb_info *sbi, struct page *page, __u8 fp[],
				virtual_t insert_vaddr, unsigned int fp_prefix)
{
	int num = get_fpfooter_counter(page);

	// If the node is full, return failure.
	if (num == FPS_PER_TREE_BLOCK)
		return FP_LIST_NODE_INSERT_FAILURE;

	// Insert the fingerprint at the end of the node.
	set_fplist_fingerprint(page, num, fp);
	set_fplist_vaddr(page, num, insert_vaddr);
	set_fpfooter_counter(page, num + 1);

	return FP_LIST_NODE_INSERT_SUCCESS;
}

/**
 * Inserts a fingerprint into a node.
 * This function inserts a fingerprint into a fingerprint node, handling both stable and unstable nodes.
 *
 * @param sbi The file system super block information.
 * @param page The page containing the node.
 * @param fp The fingerprint to insert.
 * @param insert_vaddr The virtual address to insert.
 * @param fp_prefix The fingerprint prefix.
 * @param is_stable A flag indicating whether the node is stable.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_insert_single_fingerprint_into_single_node(
				struct f2fs_sb_info *sbi, struct page *page, __u8 fp[],
				virtual_t insert_vaddr, unsigned int fp_prefix, bool is_stable)
{
	// Choose the insertion function based on the stability of the node.
	return (is_stable ?
			f2dfs_insert_single_fingerprint_into_single_stable_node(
							sbi, page, fp, insert_vaddr, fp_prefix) :
			f2dfs_insert_single_fingerprint_into_single_unstable_node(
							sbi, page, fp, insert_vaddr, fp_prefix));
}

/**
 * Transfers entries inside a single list.
 * This function transfers entries from one node to another within a list.
 *
 * @param sbi The file system super block information.
 * @param src_page The source page containing the node to transfer from.
 * @param trg_page The target page containing the node to transfer to.
 * @param fp_prefix The fingerprint prefix.
 * @param is_stable A flag indicating whether the list is stable.
 */
static inline void f2dfs_transfer_inside_single_list(
				struct f2fs_sb_info *sbi, struct page *src_page,
				struct page *trg_page, unsigned int fp_prefix, bool is_stable)
{
	__u8 temp_fp[FP_LEN_MAX];
	virtual_t temp_vaddr;
	int times = get_fpfooter_counter(src_page), i = 0,
			ret = FP_LIST_RETURN_NOTHING;

	// Transfer each entry from the source node to the target node.
	for (i = 0; i < times; ++i) {
		get_fplist_fingerprint(src_page, i, temp_fp);
		temp_vaddr = get_fplist_vaddr(src_page, i);
		ret = f2dfs_insert_single_fingerprint_into_single_node(sbi, trg_page,
						temp_fp, temp_vaddr, fp_prefix, is_stable);
	}
	set_fpfooter_counter(src_page, 0);
}
/**
 * Allocates a new node for a fingerprint list.
 * This function allocates a new page for the fingerprint list and initializes it.
 *
 * @param sbi The file system super block information.
 * @param list The fingerprint list structure.
 * @param is_stable A flag indicating whether the list is stable.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_alloc_single_list_node(struct f2fs_sb_info *sbi,
				struct f2fs_fp_list *list, bool is_stable)
{
	struct page *page;
	int i;

	// Check if the list has reached its maximum length.
	if (list->nr_pages >= FP_LIST_MAX_LEN)
		return FP_LIST_NODE_ALLOC_FAILURE;

	// Get a new fingerprint page.
	page = f2dfs_get_new_fp_page(sbi);
	if ((!page) || (PTR_ERR(page) > 0)) {
		page = NULL;
		return FP_LIST_NODE_ALLOC_FAILURE;
	}

	// Increment the list's page count and add the new page.
	++list->nr_pages;
	for (i = list->nr_pages - 1; i >= 1; --i) {
		list->pages[i] = list->pages[i - 1];
		list->alloc[i] = list->alloc[i - 1];
	}
	list->pages[1] = page;
	list->alloc[1] = 1;

	// Set the footer pointer and counter for the new page.
	set_fpfooter_pointer(list->pages[1], get_fpfooter_pointer(list->pages[0]));
	set_fpfooter_pointer(list->pages[0], nid_of_node(list->pages[1]));
	set_fpfooter_counter(list->pages[1], 0);
	if (is_stable)
		set_fpfooter_tags(list->pages[1], FP_NODE_STABLE_TAG);
	else
		clear_fpfooter_tags(list->pages[1], FP_NODE_STABLE_TAG);
	clear_fpfooter_tags(list->pages[1], FP_NODE_JOURNAL_TAG);
	set_fpfooter_prefix(list->pages[1], list->index);

	// Transfer entries from the old first page to the new first page.
	f2dfs_transfer_inside_single_list(sbi, list->pages[0], list->pages[1],
					list->index, is_stable);

	return FP_LIST_NODE_ALLOC_SUCCESS;
}

/**
 * Checks if the compaction of a fingerprint list is finished.
 * This function determines if there are enough entries in the list to consider compaction as finished.
 *
 * @param list The fingerprint list structure.
 * @return True if compaction is finished, false otherwise.
 */
static inline bool is_compact_finished(struct f2fs_fp_list *list)
{
	int i = 1, count = 0;

	// If the list has few pages, compaction is considered finished.
	if ((list->nr_pages == 0) || (list->nr_pages == 1))
		return true;

	// Count the number of entries in all pages except the first one.
	for (i = 1; i < list->nr_pages; ++i)
		count += get_fpfooter_counter(list->pages[i]);

	// If the count is more than the total capacity minus two pages, compaction is finished.
	return (count > ((list->nr_pages - 2) * FPS_PER_TREE_BLOCK));
}

/**
 * Frees a single node from a fingerprint list.
 * This function frees a specific page from the fingerprint list and updates the list structure.
 *
 * @param sbi The file system super block information.
 * @param list The fingerprint list structure.
 * @param k The index of the page to free.
 * @param is_stable A flag indicating whether the list is stable.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_free_single_list_node(struct f2fs_sb_info *sbi,
				struct f2fs_fp_list *list, int k, bool is_stable)
{
	int i, err = 0;

	// Check if the list page index is valid.
	if ((list->nr_pages <= 0) || (k >= list->nr_pages))
		return FP_LIST_NODE_FREE_FAILURE;

	// If the list has only one page and it's the one to be freed, reset the page.
	if ((k == 0) && (list->nr_pages == 1)) {
		set_fpfooter_counter(list->pages[0], 0);
		if (is_stable)
			set_fpfooter_tags(list->pages[0], FP_NODE_STABLE_TAG);
		else
			clear_fpfooter_tags(list->pages[0], FP_NODE_STABLE_TAG);
		set_fpfooter_prefix(list->pages[0], list->index);
		set_fpfooter_pointer(list->pages[0], FP_LIST_NULL);
		return FP_LIST_NODE_FREE_SUCCESS;
	}

	// If the list has more than one page and the first page is to be freed,
	// transfer entries from the second page to the first.
	if ((k == 0) && (list->nr_pages > 1)) {
		f2dfs_transfer_inside_single_list(sbi, list->pages[1], list->pages[0],
						list->index, is_stable);
		k = 1;
	}

	// Update the pointer of the previous page to bypass the freed page.
	set_fpfooter_pointer(list->pages[k - 1], get_fpfooter_pointer(list->pages[k]));

	// Truncate the fingerprint page and free it.
	err = f2dfs_truncate_fp_page(sbi, list->pages[k], list->alloc[k]);
	if (err)
		return FP_LIST_NODE_FREE_FAILURE;

	// Update the list structure to remove the freed page.
	list->pages[k] = NULL;
	for (i = k; i < list->nr_pages - 1; ++i) {
		list->pages[i] = list->pages[i + 1];
		list->alloc[i] = list->alloc[i + 1];
	}
	list->pages[list->nr_pages - 1] = NULL;
	--list->nr_pages;

	return FP_LIST_NODE_FREE_SUCCESS;
}

/**
 * Frees zero-count nodes from a fingerprint list.
 * This function frees pages that have no entries from the fingerprint list.
 *
 * @param sbi The file system super block information.
 * @param list The fingerprint list structure.
 * @param is_stable A flag indicating whether the list is stable.
 */
static inline void f2dfs_free_zero_list_nodes(struct f2fs_sb_info *sbi,
				struct f2fs_fp_list *list, bool is_stable)
{
	int i = 1;

	// If the list has one or fewer pages, there's nothing to do.
	if (list->nr_pages <= 1)
		return;

	// Iterate over the list pages and free any that have no entries.
	while (i < list->nr_pages) {
		if (get_fpfooter_counter(list->pages[i]) == 0) {
			f2dfs_free_single_list_node(sbi, list, i, is_stable);
			continue;
		}
		++i;
	}
}
/**
 * Compacts a single fingerprint list.
 * This function moves entries from the last filled node to the previous ones in the list
 * to keep the list compact and free up space at the end of the list.
 *
 * @param sbi The file system super block information.
 * @param list The fingerprint list structure to be compacted.
 * @param is_stable A flag indicating whether the list is stable or not.
 */
static inline void f2dfs_compact_single_list(struct f2fs_sb_info *sbi,
				struct f2fs_fp_list *list, bool is_stable)
{
	__u8 temp_fp[FP_LEN_MAX]; // Temporary storage for fingerprint data.
	virtual_t temp_vaddr; // Temporary storage for virtual address.
	int times, src_index, trg_index, i = 0, ret = FP_LIST_RETURN_NOTHING;

	// Free nodes in the list that have zero entries.
	f2dfs_free_zero_list_nodes(sbi, list, is_stable);

release_tail:
	// If the compaction is finished, return.
	if (is_compact_finished(list))
		return;

	// Identify the source and target indices for moving entries.
	src_index = list->nr_pages - 1;
	trg_index = list->nr_pages - 2;
	times = get_fpfooter_counter(list->pages[src_index]);

	// Move entries from the source node to the target node.
	for (i = 0; i < times; ++i) {
		get_fplist_fingerprint(list->pages[src_index], i, temp_fp);
		temp_vaddr = get_fplist_vaddr(list->pages[src_index], i);

insert_redo:
		ret = FP_LIST_RETURN_NOTHING;
		// Insert the fingerprint into the target node.
		ret = f2dfs_insert_single_fingerprint_into_single_node(sbi,
						list->pages[trg_index], temp_fp, temp_vaddr,
						list->index, is_stable);
		if (ret != FP_LIST_NODE_INSERT_SUCCESS) {
			--trg_index;
			goto insert_redo;
		}
	}

	// If all entries have been moved, free the source node.
	if (i >= times - 1)
		f2dfs_free_single_list_node(sbi, list, src_index, is_stable);
	else {
		// Update the counter for the source node.
		set_fpfooter_counter(list->pages[src_index], times - i - 1);
		return;
	}

	goto release_tail;
}

/**
 * Inserts a single fingerprint into a single fingerprint list.
 * This function attempts to insert a fingerprint into the list, starting from the first node.
 *
 * @param sbi The file system super block information.
 * @param list The fingerprint list structure to insert into.
 * @param fp The fingerprint data to insert.
 * @param insert_vaddr The virtual address associated with the fingerprint.
 * @param is_stable A flag indicating whether the list is stable or not.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_insert_single_fingerprint_into_single_list(
				struct f2fs_sb_info *sbi, struct f2fs_fp_list *list,
				__u8 fp[], virtual_t insert_vaddr, bool is_stable)
{
	int ret = FP_LIST_RETURN_NOTHING, i = 0, times = 1;

	// If the virtual address is not valid, consider the operation successful.
	if (!__is_valid_data_blkaddr(insert_vaddr))
		return FP_LIST_INSERT_SUCCESS;

	// If the fingerprint is marked for removal and the list is stable, consider the operation successful.
	if (is_stable && (f2dfs_fp_is_remove(fp)))
		return FP_LIST_INSERT_SUCCESS;

insert_redo:
	// Determine the number of times to attempt insertion based on the list stability.
	times = (is_stable ? list->nr_pages : 1); // Unstable lists only append to the head.
	for (i = 0; i < times; ++i) {
		// Attempt to insert the fingerprint into a node within the list.
		ret = f2dfs_insert_single_fingerprint_into_single_node(sbi,
						list->pages[i], fp, insert_vaddr,
						list->index, is_stable);
		if (ret == FP_LIST_NODE_INSERT_SUCCESS)
			return FP_LIST_INSERT_SUCCESS;
	}
	// If insertion fails, allocate a new list node and retry.
	ret = f2dfs_alloc_single_list_node(sbi, list, true);
	if (ret == FP_LIST_NODE_ALLOC_FAILURE)
		return FP_LIST_INSERT_FAILURE;
	else
		goto insert_redo;

	return FP_LIST_INSERT_FAILURE;
}

/**
 * Persists a single bucket to a single unstable list.
 * This function takes a fingerprint bucket and inserts its entries into an unstable fingerprint list.
 *
 * @param sbi The file system super block information.
 * @param list The fingerprint list structure to persist into.
 * @param bucket The fingerprint bucket containing entries to persist.
 */
static inline void f2dfs_persist_single_bucket_to_single_unstable_list(
				struct f2fs_sb_info *sbi, struct f2fs_fp_list *list,
				struct f2dfs_fp_bucket *bucket)
{
	int ret = FP_LIST_RETURN_NOTHING, i = 0, j = 0, r_i = 0;
	__u8 fp_remove[FP_LEN_MAX];

	// Get the fingerprint data for removal based on the list index.
	f2dfs_get_fp_remove(fp_remove, list->index);
	for (i = 1; i <= bucket->remove_count; ++i) {
		r_i = FP_ENTRIES_PER_BUCKET - i;
		// If an entry is not marked for persistence or removal, reset the remove count and break.
		if ((!fp_is_persist_tag(&bucket->entries[r_i])) ||
					(!fp_is_remove_tag(&bucket->entries[r_i]))) {
			bucket->remove_count = 0;
			break;
		}

		// Insert each removal fingerprint into the list.
		for (j = 0; j < FP_REMOVE_PER_ENTRY; ++j) {
			ret = FP_LIST_RETURN_NOTHING;
			ret = f2dfs_insert_single_fingerprint_into_single_list(
							sbi, list, fp_remove,
							bucket->entries[r_i].re.remove_addr[j], false);
		}
		// Reset the entry tag and mark it for removal.
		init_fp_tag(&bucket->entries[r_i]);
		fp_set_remove_tag(&bucket->entries[r_i]);
	}

	// Persist valid entries that are not marked for removal.
	for (i = 0; i < bucket->valid_count; ++i) {
		if (fp_is_persist_tag(&bucket->entries[i]))
			continue;
		if (fp_is_remove_tag(&bucket->entries[i])) {
			bucket->valid_count = 0;
			break;
		}

		ret = FP_LIST_RETURN_NOTHING;
		ret = f2dfs_insert_single_fingerprint_into_single_list(sbi,
						list, bucket->entries[i].ne.fingerprint,
						bucket->entries[i].ne.virtual_addr, false);
		// Mark the entry as persisted.
		fp_set_persist_tag(&bucket->entries[i]);
	}
}
/**
 * Clears a single virtual address from a fingerprint list.
 * This function removes all occurrences of a virtual address from a fingerprint list,
 * starting from a specified node index.
 *
 * @param sbi The file system super block information.
 * @param list The fingerprint list structure.
 * @param remove_vaddr The virtual address to remove.
 * @param k The index of the node to start from.
 * @param is_stable A flag indicating whether the list is stable.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_clear_single_vaddr_from_single_list(
				struct f2fs_sb_info *sbi, struct f2fs_fp_list *list,
				virtual_t remove_vaddr, int k, bool is_stable)
{
	int i = (is_stable ? 0 : (k + 1)), ret = FP_LIST_RETURN_NOTHING;

	// Iterate over the nodes in the list starting from the specified index.
	while (i < list->nr_pages) {
		ret = f2dfs_remove_single_vaddr_from_single_node(sbi,
						list->pages[i], remove_vaddr);
		if (ret == FP_LIST_NODE_REMOVE_SUCCESS)
			return FP_LIST_REMOVE_SUCCESS;
		++i;
	}

	// If no matching virtual address is found in any node, return failure.
	return FP_LIST_REMOVE_FAILURE;
}

/**
 * Clears a single virtual address within a single node.
 * This function removes a virtual address from a single fingerprint node,
 * starting from a specified index.
 *
 * @param page The page containing the node.
 * @param remove_vaddr The virtual address to remove.
 * @param t The index of the entry to start from.
 * @return An integer result code indicating the success or failure of the operation.
 */
static inline int f2dfs_clear_single_vaddr_in_single_node(
				struct page *page, virtual_t remove_vaddr, int t)
{
	__u8 node_fp[FP_LEN_MAX];
	int i = 0, num = get_fpfooter_counter(page), move_step = 0;

	// Get the fingerprint at the starting index.
	get_fplist_fingerprint(page, t, node_fp);
	// If the virtual address does not match or the fingerprint is not marked for removal, return failure.
	if ((get_fplist_vaddr(page, t) != remove_vaddr) ||
				(!f2dfs_fp_is_remove(node_fp)))
		return FP_LIST_REMOVE_FAILURE;

	// Shift entries to fill the gap left by the removed virtual address.
	for (i = 0; i < num; ++i) {
		if ((i < t) && (get_fplist_vaddr(page, i) == remove_vaddr)) {
			++move_step;
			continue;
		}

		if (move_step > 0) {
			get_fplist_fingerprint(page, i, node_fp);
			set_fplist_fingerprint(page, (i - move_step), node_fp);
			set_fplist_vaddr(page, (i - move_step), get_fplist_vaddr(page, i));
		}
	}

	// Update the counter for the node if any entries were moved.
	if (move_step > 0) {
		set_fpfooter_counter(page, num - move_step);
		return FP_LIST_REMOVE_SUCCESS;
	}

	// If no entries were moved, return failure.
	return FP_LIST_REMOVE_FAILURE;
}

/**
 * Clears removal virtual addresses within a single node.
 * This function removes all virtual addresses associated with a specific fingerprint
 * from a single fingerprint node.
 *
 * @param page The page containing the node.
 * @param fp_remove The fingerprint to match for removal.
 */
static inline void f2dfs_clear_remove_vaddrs_in_single_node(
				struct page *page, __u8 fp_remove[])
{
	int i = 0, num = get_fpfooter_counter(page), move_step = 0;
	__u8 temp_fp[FP_LEN_MAX];

	// Iterate over the entries in the node.
	for (i = 0; i < num; ++i) {
		get_fplist_fingerprint(page, i, temp_fp);
		// If the fingerprint matches the one marked for removal, increment the move step.
		if (f2dfs_fp_is_equal(fp_remove, temp_fp) == 0) {
			++move_step;
			continue;
		}

		// Shift the entries to fill the gap left by the removed entry.
		if (move_step > 0) {
			set_fplist_fingerprint(page, (i - move_step), temp_fp);
			set_fplist_vaddr(page, (i - move_step), get_fplist_vaddr(page, i));
		}
	}

	// Update the counter for the node if any entries were moved.
	if (move_step > 0)
		set_fpfooter_counter(page, num - move_step);
}
/**
 * Clears removal virtual addresses from the file system.
 * This function removes all virtual addresses associated with a specific fingerprint
 * from both unstable and stable fingerprint lists.
 *
 * @param sbi The file system super block information.
 * @param list_1 The unstable fingerprint list.
 * @param list_2 The stable fingerprint list.
 */
static inline void f2dfs_clear_remove_vaddrs_from_fs(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_list *list_1, struct f2dfs_fp_list *list_2)
{
	int i = 0, j = 0, num, ret = FP_LIST_RETURN_NOTHING;
	__u8 fp_remove[FP_LEN_MAX], temp_fp[FP_LEN_MAX];
	virtual_t temp_va;

	// Get the fingerprint data for removal based on the list index.
	f2dfs_get_fp_remove(fp_remove, list_1->index);
	for (i = list_1->nr_pages - 1; i >= 0; --i) {
		j = 0;
		num = get_fpfooter_counter(list_1->pages[i]);

		// Iterate over the entries in the current page.
		while (j < num) {
			get_fplist_fingerprint(list_1->pages[i], j, temp_fp);
			// If the fingerprint matches the one marked for removal.
			if (f2dfs_fp_is_equal(temp_fp, fp_remove) == 0) {
				// There should be only one <FP, VA> pair before one remove entry.
				temp_va = get_fplist_vaddr(list_1->pages[i], j);
				// Try to clear the virtual address within the current node.
				ret = f2dfs_clear_single_vaddr_in_single_node(
								list_1->pages[i], temp_va, j);
				if (ret == FP_LIST_REMOVE_SUCCESS) {
					num = get_fpfooter_counter(list_1->pages[i]);
					continue;
				}

				// Try to clear the virtual address from the unstable list.
				ret = f2dfs_clear_single_vaddr_from_single_list(
								sbi, list_1, temp_va, i, false);
				if (ret == FP_LIST_REMOVE_SUCCESS)
					goto skip;

				// Try to clear the virtual address from the stable list.
				ret = f2dfs_clear_single_vaddr_from_single_list(
								sbi, list_2, temp_va, -1, true);
				if (ret == FP_LIST_REMOVE_SUCCESS)
					goto skip;
			}

skip:
			++j;
		}
		// Clear removal virtual addresses within the current node.
		f2dfs_clear_remove_vaddrs_in_single_node(list_1->pages[i], fp_remove);
	}
}

/**
 * Searches for a single fingerprint in a single stable node.
 * This function finds a fingerprint in a stable fingerprint node and returns the associated virtual address.
 *
 * @param sbi The file system super block information.
 * @param page The stable page to search in.
 * @param fp The fingerprint to search for.
 * @param result_vaddr A pointer to store the result virtual address.
 * @param fp_prefix The fingerprint prefix.
 * @return An integer result code indicating whether the node is stable or not.
 */
static inline int f2dfs_search_single_fingerprint_in_single_stable_node(
				struct f2fs_sb_info *sbi, struct page *page, __u8 fp[],
				virtual_t *result_vaddr, unsigned int fp_prefix)
{
	int left = 0, right = get_fpfooter_counter(page) - 1, middle = 0;
	__u8 node_fp[FP_LEN_MAX];
	struct f2dfs_bucket_change stable_change = {
		.virtual_addr = NULL_ADDR,
		.fp_prefix = DIRECT_FP_PREFIX,
		.is_stable = false,
		.is_persist = false,
		.is_dupl = false,
		.is_remove = false,
	};

	(*result_vaddr) = NULL_ADDR;
	// If the node is empty, return unstable.
	if (right == -1)
		return FP_LIST_NODE_IS_UNSTABLE;

	// Check the fingerprint at the left and right boundaries.
	get_fplist_fingerprint(page, left, node_fp);
	if (f2dfs_fp_is_equal(fp, node_fp) < 0)
		return FP_LIST_NODE_IS_UNSTABLE;

	get_fplist_fingerprint(page, right, node_fp);
	if (f2dfs_fp_is_equal(fp, node_fp) > 0)
		return FP_LIST_NODE_IS_UNSTABLE;

	// Perform a binary search for the fingerprint.
	while (left <= right) {
		middle = (left + right) / 2;
		get_fplist_fingerprint(page, middle, node_fp);

		if (f2dfs_fp_is_equal(fp, node_fp) > 0)
			left = middle + 1;
		else if (f2dfs_fp_is_equal(fp, node_fp) < 0)
			right = middle - 1;
		else {
			// If the fingerprint is found, update the stable bucket cache and return stable.
			(*result_vaddr) = get_fplist_vaddr(page, middle);
			stable_change.is_stable = true;
			stable_change.is_persist = true;
			stable_change.fp_prefix = fp_prefix;
			stable_change.virtual_addr = (*result_vaddr);
			memcpy(&stable_change.fingerprint, &node_fp, FP_LEN_MAX);
			f2dfs_process_single_fingerprint_for_bucket_cache(sbi,
							&stable_change, NULL_ADDR, NULL,
							FP_BUCKET_CACHE_FOR_SET_STABLE);
			return FP_LIST_NODE_IS_STABLE;
		}
	}

	// If the fingerprint is not found, return unstable.
	return FP_LIST_NODE_IS_UNSTABLE;
}
/**
 * Synchronizes fingerprint buckets cache in the background.
 * This function synchronizes the fingerprint buckets cache to the persistent storage,
 * handling various conditions such as unmount, ioctl, and the need for stabilization.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 * @param ioctl A flag indicating whether the operation is triggered by an ioctl call.
 * @param need_stable A flag indicating whether stabilization is needed.
 * @return True if the synchronization is successful, false otherwise.
 */
static inline bool f2dfs_sync_buckets_cache_background(struct f2fs_sb_info *sbi,
				bool umount, bool ioctl, bool need_stable)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	int i;

	// Iterate over each prefix index in the fingerprint bucket cache.
	for (i = 0; i < FP_PREFIX_NUMBER; ++i) {
		head = &sbi->s_fingerprint_bucket->pointers[i];
		if (!head)
			return false;

		// Iterate over each bucket in the current prefix index.
		slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[i]) {
			if (!head)
				return false;

			bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
			if (!bucket)
				return false;

			// Check if there's not enough space for persistence and return false if so.
			if (has_not_enough_space_for_single_persist(sbi, bucket))
				return false;

			// If unmounting or an ioctl call, and the bucket needs to be persisted, do so.
			if ((umount || ioctl) &&
						should_persist_single_bucket_umount(bucket)) {
				f2dfs_sync_single_bucket_background(sbi, bucket, i, umount);
				bucket->remove_count = 0;
			}
			// Otherwise, if not unmounting and not an ioctl call, and the bucket needs to be persisted normally, do so.
			else if ((!umount) && (!ioctl) &&
						should_persist_single_bucket_normal(bucket)) {
				f2dfs_sync_single_bucket_background(sbi, bucket, i, umount);
				f2dfs_shrink_single_bucket(sbi, bucket);
			}
		}
	}

	// If unmounting, an ioctl call, or stabilization is needed, stabilize the buckets.
	if (umount || ioctl || need_stable) {
		for (i = 0; i < FP_PREFIX_NUMBER; ++i) {
			// Check if there's not enough space for stabilization and return false if so.
			if (has_not_enough_space_for_single_stable(sbi))
				return false;

			// If the current prefix index is marked for stabilization, stabilize the list.
			if (f2fs_test_bit(i,
						(char *)sbi->s_fingerprint_bucket->s_stable_bitmap))
				f2dfs_stabilize_single_list_background(sbi, i, umount);
		}
	}

	return true;
}

/**
 * Creates a single list head.
 * This function initializes a new fingerprint list head.
 *
 * @param sbi The file system super block information.
 * @param index The index of the list.
 * @param is_stable A flag indicating whether the list is stable or not.
 */
static inline void f2dfs_create_single_list_head(struct f2fs_sb_info *sbi,
				unsigned int index, bool is_stable)
{
	struct f2fs_fp_list *list = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2dfs_fp_list)), GFP_NOFS);
	struct node_info ni;
	int size = 1;

	// Allocate memory for the list's pages.
	list->pages = f2fs_kvzalloc(sbi, array_size(size,
					sizeof(struct page *)), GFP_NOFS);
	list->nr_pages = 0;
	list->index = index;
	memset(list->alloc, 0, FP_LIST_MAX_LEN);

	// Get the node information for the list head.
	if (f2fs_get_node_info(sbi, (is_stable ? FP_NID_STABLE(index) :
				FP_NID_UNSTABLE(index)), &ni))
		goto out;

	// Check if the node information is valid.
	if (__is_valid_data_blkaddr(ni.blk_addr) &&
				f2fs_is_valid_blkaddr(sbi, ni.blk_addr, DATA_GENERIC_ENHANCE))
		goto out;

	// Get the single list head and release it.
	f2dfs_get_single_list_head(sbi, list, is_stable);
	f2dfs_release_single_list_head(sbi, list, size, false);

out:
	// Free the allocated memory.
	kvfree(list->pages);
	kvfree(list);
}

/**
 * Performs persistence and stabilization.
 * This function calls the background synchronization function for persistence and stabilization.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 * @param ioctl A flag indicating whether the operation is triggered by an ioctl call.
 * @param need_stable A flag indicating whether stabilization is needed.
 * @return True if the operation is successful, false otherwise.
 */
bool f2dfs_do_persistence_and_stabilization(struct f2fs_sb_info *sbi,
				bool umount, bool ioctl, bool need_stable)
{
	return f2dfs_sync_buckets_cache_background(sbi, umount, ioctl, need_stable);
}

/**
 * Synchronizes a single bucket in the foreground.
 * This function synchronizes a single fingerprint bucket to the persistent storage.
 *
 * @param sbi The file system super block information.
 * @param bucket The fingerprint bucket to synchronize.
 * @param index The index of the bucket.
 */
void f2dfs_sync_single_bucket_foreground(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_bucket *bucket, unsigned int index)
{
	struct f2dfs_fp_list *list = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2dfs_fp_list)), GFP_NOFS);
	int size = DIV_ROUND_UP(((FP_REMOVE_PER_ENTRY * bucket->remove_count) +
					bucket->valid_count), FPS_PER_TREE_BLOCK) + 1;

	// Allocate memory for the list's pages.
	list->pages = f2fs_kvzalloc(sbi, array_size(size,
					sizeof(struct page *)), GFP_NOFS);
	list->nr_pages = 0;
	list->index = index;
	memset(list->alloc, 0, FP_LIST_MAX_LEN);

	// Check if there's not enough space for persistence and return if so.
	if (has_not_enough_space_for_single_persist(sbi, bucket))
		goto out;

	// Get the single list head and persist the bucket to the unstable list.
	f2dfs_get_single_list_head(sbi, list, false);
	f2dfs_persist_single_bucket_to_single_unstable_list(sbi, list, bucket);
	f2dfs_shrink_single_bucket(sbi, bucket);
	f2dfs_release_single_list_head(sbi, list, size, false);

out:
	// Free the allocated memory.
	kvfree(list->pages);
	kvfree(list);
}

/**
 * Synchronizes fingerprint buckets cache in the background.
 * This function processes each fingerprint bucket to persist data to storage
 * based on conditions like unmount, ioctl, and stabilization needs.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 * @param ioctl A flag indicating whether the operation is triggered by an ioctl call.
 * @param need_stable A flag indicating whether stabilization is needed.
 * @return True if the synchronization is successful, false otherwise.
 */
static inline bool f2dfs_sync_buckets_cache_background(struct f2fs_sb_info *sbi,
				bool umount, bool ioctl, bool need_stable)
{
	struct slist_head *head;
	struct f2dfs_fp_bucket *bucket;
	int i;

	// Iterate over each prefix index in the fingerprint bucket cache.
	for (i = 0; i < FP_PREFIX_NUMBER; ++i) {
		head = &sbi->s_fingerprint_bucket->pointers[i];
		if (!head)
			return false;

		// Iterate over each bucket in the current prefix index.
		slist_for_each(head, &sbi->s_fingerprint_bucket->pointers[i]) {
			if (!head)
				return false;

			bucket = slist_entry(head, struct f2dfs_fp_bucket, b_head);
			if (!bucket)
				return false;

			// If there's not enough space to persist, return false.
			if (has_not_enough_space_for_single_persist(sbi, bucket))
				return false;

			// If unmounting or an ioctl call, and the bucket needs to be persisted, do so.
			if ((umount || ioctl) &&
						should_persist_single_bucket_umount(bucket)) {
				f2dfs_sync_single_bucket_background(sbi, bucket, i, umount);
				bucket->remove_count = 0;
			}
			// If not unmounting and not an ioctl call, and the bucket needs to be persisted normally, do so.
			else if ((!umount) && (!ioctl) &&
						should_persist_single_bucket_normal(bucket)) {
				f2dfs_sync_single_bucket_background(sbi, bucket, i, umount);
				f2dfs_shrink_single_bucket(sbi, bucket);
			}
		}
	}

	// If unmounting, an ioctl call, or stabilization is needed, stabilize the buckets.
	if (umount || ioctl || need_stable) {
		for (i = 0; i < FP_PREFIX_NUMBER; ++i) {
			// If there's not enough space to stabilize, return false.
			if (has_not_enough_space_for_single_stable(sbi))
				return false;

			// If the current prefix index is marked for stabilization, stabilize the list.
			if (f2fs_test_bit(i,
						(char *)sbi->s_fingerprint_bucket->s_stable_bitmap))
				f2dfs_stabilize_single_list_background(sbi, i, umount);
		}
	}

	return true;
}

/**
 * Creates a single list head.
 * This function initializes a new fingerprint list head and checks for its validity.
 *
 * @param sbi The file system super block information.
 * @param index The index of the list.
 * @param is_stable A flag indicating whether the list is stable or not.
 */
static inline void f2dfs_create_single_list_head(struct f2fs_sb_info *sbi,
				unsigned int index, bool is_stable)
{
	struct f2fs_fp_list *list = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2dfs_fp_list)), GFP_NOFS);
	struct node_info ni;
	int size = 1;

	// Allocate memory for the list's pages.
	list->pages = f2fs_kvzalloc(sbi, array_size(size,
					sizeof(struct page *)), GFP_NOFS);
	list->nr_pages = 0;
	list->index = index;
	memset(list->alloc, 0, FP_LIST_MAX_LEN);

	// Get the node information for the list head.
	if (f2fs_get_node_info(sbi, (is_stable ? FP_NID_STABLE(index) :
				FP_NID_UNSTABLE(index)), &ni))
		goto out;

	// If the node information is not valid, skip creating the list head.
	if (__is_valid_data_blkaddr(ni.blk_addr) &&
				f2fs_is_valid_blkaddr(sbi, ni.blk_addr, DATA_GENERIC_ENHANCE))
		goto out;

	// Get the single list head and release it.
	f2dfs_get_single_list_head(sbi, list, is_stable);
	f2dfs_release_single_list_head(sbi, list, size, false);

out:
	// Free the allocated memory.
	kvfree(list->pages);
	kvfree(list);
}

/**
 * Performs persistence and stabilization.
 * This function calls the background synchronization function for persistence and stabilization.
 *
 * @param sbi The file system super block information.
 * @param umount A flag indicating whether the operation is during unmount.
 * @param ioctl A flag indicating whether the operation is triggered by an ioctl call.
 * @param need_stable A flag indicating whether stabilization is needed.
 * @return True if the operation is successful, false otherwise.
 */
bool f2dfs_do_persistence_and_stabilization(struct f2fs_sb_info *sbi,
				bool umount, bool ioctl, bool need_stable)
{
	return f2dfs_sync_buckets_cache_background(sbi, umount, ioctl, need_stable);
}

/**
 * Synchronizes a single bucket in the foreground.
 * This function synchronizes a single fingerprint bucket to the persistent storage.
 *
 * @param sbi The file system super block information.
 * @param bucket The fingerprint bucket to synchronize.
 * @param index The index of the bucket.
 */
void f2dfs_sync_single_bucket_foreground(struct f2fs_sb_info *sbi,
				struct f2dfs_fp_bucket *bucket, unsigned int index)
{
	struct f2dfs_fp_list *list = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2dfs_fp_list)), GFP_NOFS);
	int size = DIV_ROUND_UP(((FP_REMOVE_PER_ENTRY * bucket->remove_count) +
					bucket->valid_count), FPS_PER_TREE_BLOCK) + 1;

	// Allocate memory for the list's pages.
	list->pages = f2fs_kvzalloc(sbi, array_size(size,
					sizeof(struct page *)), GFP_NOFS);
	list->nr_pages = 0;
	list->index = index;
	memset(list->alloc, 0, FP_LIST_MAX_LEN);

	// If there's not enough space to persist, skip synchronization.
	if (has_not_enough_space_for_single_persist(sbi, bucket))
		goto out;

	// Get the single list head and persist the bucket to the unstable list.
	f2dfs_get_single_list_head(sbi, list, false);
	f2dfs_persist_single_bucket_to_single_unstable_list(sbi, list, bucket);
	f2dfs_shrink_single_bucket(sbi, bucket);
	f2dfs_release_single_list_head(sbi, list, size, false);

out:
	// Free the allocated memory.
	kvfree(list->pages);
	kvfree(list);
}
/**
 * Performs offline deduplication for a segment summary area block.
 * This function updates the segment summary block with a new virtual address
 * for a given block address, parent node ID, and offset within the node.
 *
 * @param sbi The file system super block information.
 * @param blkaddr The block address of the segment summary area.
 * @param new_vaddr The new virtual address to update.
 * @param parent_nid The parent node ID.
 * @param ofs_in_node The offset within the node.
 * @return 0 on success, 1 if the update fails due to mismatch.
 */
static inline int f2dfs_do_offline_for_segment_summary_area_block(
				struct f2fs_sb_info *sbi, block_t blkaddr, virtual_t new_vaddr,
				nid_t parent_nid, int ofs_in_node)
{
	unsigned int segno = GET_SEGNO(sbi, blkaddr);
	unsigned short blkoff = GET_BLKOFF_FROM_SEG0(sbi, blkaddr);
	struct f2fs_summary_block *sum_node;
	struct page *sum_page;
	int type = get_seg_entry(sbi, segno)->type;
	struct curseg_info *curseg = CURSEG_I(sbi, type);

	// Lock the current segment mutex for thread safety.
	mutex_lock(&curseg->curseg_mutex);
	// If the segment number matches the current segment, update the summary block.
	if (segno == curseg->segno) {
		if ((curseg->sum_blk->entries[blkoff].nid != parent_nid) ||
					(curseg->sum_blk->entries[blkoff].ofs_in_node
					!= ofs_in_node)) {
			mutex_unlock(&curseg->curseg_mutex);
			return 1;
		}
		curseg->sum_blk->entries[blkoff].nid = new_vaddr;
		curseg->sum_blk->entries[blkoff].ofs_in_node = SSA_VIRTUAL_ADDR;
		mutex_unlock(&curseg->curseg_mutex);
		return 0;
	}
	mutex_unlock(&curseg->curseg_mutex);

	// Get the summary page for the given segment number.
	sum_page = f2fs_get_sum_page(sbi, segno);
	if (IS_ERR(sum_page))
		return PTR_ERR(sum_page);

	// Get the summary node from the page.
	sum_node = (struct f2fs_summary_block *)page_address(sum_page);
	// If the node ID or offset does not match, release the page and return 1.
	if ((sum_node->entries[blkoff].nid != parent_nid) ||
				(sum_node->entries[blkoff].ofs_in_node != ofs_in_node)) {
		f2fs_put_page(sum_page, 1);
		return 1;
	}
	// Update the node ID and offset in the summary block.
	sum_node->entries[blkoff].nid = new_vaddr;
	sum_node->entries[blkoff].ofs_in_node = SSA_VIRTUAL_ADDR;

	// Mark the page as dirty and release it.
	set_page_dirty(sum_page);
	f2fs_put_page(sum_page, 1);
	return 0;
}

/**
 * Performs offline deduplication for a single fingerprint.
 * This function finds duplicates for a single fingerprint and updates
 * the virtual address if a duplicate is found.
 *
 * @param sbi The file system super block information.
 * @param fp The fingerprint to search for.
 * @param parent_nid The parent node ID.
 * @param ofs_in_node The offset within the node.
 * @param vaddr The virtual address of the fingerprint.
 * @param index The index of the fingerprint.
 * @param is_physical_to_virtual A pointer to a flag indicating whether the address is converted from physical to virtual.
 * @return The resulting virtual address after offline deduplication.
 */
virtual_t f2dfs_do_offline_for_single_fingerprint(struct f2fs_sb_info *sbi,
				__u8 fp[], nid_t parent_nid, int ofs_in_node, virtual_t vaddr,
				unsigned int index, bool *is_physical_to_virtual)
{
	struct f2dfs_fp_list *list = f2fs_kvzalloc(sbi, array_size(1,
					sizeof(struct f2dfs_fp_list)), GFP_NOFS);
	virtual_t result_vaddr = NULL_ADDR;
	int ret = FP_LIST_IS_UNSTABLE;
	struct vnode_of_data vn;

	// Allocate memory for the list's pages.
	list->pages = f2fs_kvzalloc(sbi, array_size(FP_LIST_MAX_LEN,
					sizeof(struct page *)), GFP_NOFS);
	list->nr_pages = 0;
	list->index = index;
	memset(list->alloc, 0, FP_LIST_MAX_LEN);
	(*is_physical_to_virtual) = false;

	// If the fingerprint is zero, return the zero address.
	if (f2dfs_fp_is_zero(sbi, fp, index)) {
		result_vaddr = ZERO_ADDR;
		goto out;
	}

	// Get the single list for the fingerprint.
	f2dfs_get_single_list(sbi, list, true);
	// Search for the fingerprint in the stable list.
	ret = f2dfs_search_single_fingerprint_in_single_stable_list(
					sbi, list, fp, &result_vaddr);
	// If the fingerprint is not found or the result address is invalid, handle it.
	if ((ret == FP_LIST_IS_UNSTABLE) ||
				(!__is_valid_data_blkaddr(result_vaddr))) {
		result_vaddr = vaddr;
		// If the virtual address is physical, allocate a new virtual address.
		if (is_real_dnode_physical(vaddr)) {
			if (!f2fs_alloc_nid(sbi, &result_vaddr)) {
				result_vaddr = NULL_ADDR;
				goto release;
			}
			set_new_vnode(&vn, result_vaddr,
							get_dnode_physical(vaddr), index, true, true);
			// Set the virtual address in the file system.
			if (f2dfs_set_virtual_addr(sbi, &vn)) {
				f2fs_alloc_nid_failed(sbi, result_vaddr);
				result_vaddr = NULL_ADDR;
				goto release;
			}
			f2fs_alloc_nid_done(sbi, result_vaddr);
			(*is_physical_to_virtual) = true;

			// Perform offline deduplication for the segment summary area block.
			if (f2dfs_do_offline_for_segment_summary_area_block(sbi,
						get_dnode_physical(vaddr), result_vaddr,
						parent_nid, ofs_in_node) != 0) {
				result_vaddr = NULL_ADDR;
				goto release;
			}
		}
		// Insert the fingerprint into the list.
		f2dfs_insert_single_fingerprint_into_single_list(
						sbi, list, fp, result_vaddr, true);
	}

release:
	// Release the single list.
	f2dfs_release_single_list(sbi, list, true, false);

out:
	// Free the allocated memory.
	kvfree(list->pages);
	kvfree(list);

	return result_vaddr;
}
