#include <linux/fs.h>
#include "f2dfs_fs.h"
#include <linux/mpage.h>
#include <linux/backing-dev.h>
#include <linux/blkdev.h>
#include <linux/pagevec.h>
#include <linux/swap.h>

#include "f2fs.h"
#include "node.h"
#include "segment.h"
#include "xattr.h"
#include "trace.h"
#include "gc.h"
#include <trace/events/f2fs.h>

/* gc_inode pages aops */

/**
 * Clear the page dirty flag for a given virtual address in the garbage collection process.
 * 
 * @param sbi - The super block information of the file system.
 * @param vaddr - The virtual address to clear dirty flag for.
 */
void f2dfs_clear_gc_page_dirty(struct f2fs_sb_info *sbi, virtual_t vaddr)
{
    struct page *page;

    // Early return if the function is not needed
    return;

    // Check if the virtual address is valid for data block address
    if (!__is_valid_data_blkaddr(vaddr))
        goto direct_out;

    // Grab the cache page for the given virtual address
    page = f2fs_grab_cache_page(GC_MAPPING(sbi), vaddr, false);
    if (!page)
        return;

    // Check if the page index and mapping are correct
    if ((page->index != vaddr) ||
            (page->mapping != GC_MAPPING(sbi))) {
        goto not_clear;
    }

    // Clear the cold data flag for the page
    clear_cold_data(page);
    if (PageDirty(page)) {
        // Clear the page cache dirty tag and page dirty for I/O
        f2fs_clear_page_cache_dirty_tag(page);
        clear_page_dirty_for_io(page);
        // Decrement the dirty data page count
        dec_page_count(sbi, F2FS_DIRTY_DATA);
    }
    // Clear the page up-to-date flag
    ClearPageUptodate(page);

not_clear:
    // Release the page and invalidate the mapping pages
    f2fs_put_page(page, 1);
    invalidate_mapping_pages(GC_MAPPING(sbi), vaddr, vaddr);

direct_out:
    return;
}

/**
 * Perform the write operation for a garbage collection page.
 * 
 * @param fio - The file I/O information structure.
 * @param virtual_addr - The virtual address of the page to write.
 * @param is_bggc - Flag indicating if this is a background garbage collection.
 */
static inline void f2dfs_do_write_gc_page(struct f2fs_io_info *fio,
        virtual_t virtual_addr, bool is_bggc)
{
    struct f2fs_summary sum;

    // Set up the summary structure for the page
    set_summary(&sum, virtual_addr, SSA_VIRTUAL_ADDR, 0);
    // Perform the page write operation
    do_write_page(&sum, fio);
    // Update the I/O statistics
    f2fs_update_iostat(fio->sbi, fio->io_type, F2FS_BLKSIZE);
    // Invalidate the mapping pages
    invalidate_mapping_pages(GC_MAPPING(fio->sbi), virtual_addr, virtual_addr);
}

/**
 * Write a garbage collection page to the storage.
 * 
 * @param page - The page structure to write.
 * @param submitted - A pointer to a flag indicating if the page was submitted for write.
 * @param wbc - The writeback control structure.
 * @param io_type - The type of I/O operation.
 * @param virtual_addr - The virtual address of the page to write.
 * @param is_bggc - Flag indicating if this is a background garbage collection.
 * @return - The result of the write operation.
 */
static inline int __write_gc_page(
                struct page *page, bool *submitted,
                struct writeback_control *wbc,
                enum iostat_type io_type,
                virtual_t virtual_addr, bool is_bggc)
{
    struct f2fs_sb_info *sbi = F2FS_P_SB(page);
    struct node_info ni;
    struct f2fs_io_info fio = {
        .sbi = sbi,
        .ino = F2FS_GC_INO(sbi),
        .type = DATA,
        .temp = COLD,
        .op = REQ_OP_WRITE,
        .op_flags = wbc_to_write_flags(wbc),
        .page = page,
        .encrypted_page = NULL,
        .submitted = false,
        .io_type = io_type,
        .io_wbc = wbc,
    };

    // For background garbage collection, use the page index as virtual address
    if (is_bggc)
        virtual_addr = le32_to_cpu(page->index);

    // Trace the write page operation
    trace_f2fs_writepage(page, DATA);

    // Check for file system corruption and other conditions that may require redirtying the page
    if (unlikely(f2fs_cp_error(sbi))) {
        if (is_sbi_flag_set(sbi, SBI_IS_CLOSE)) {
            ClearPageUptodate(page);
            dec_page_count(sbi, F2FS_DIRTY_DATA);
            unlock_page(page);
            return 0;
        }
        goto redirty_out;
    }

    // Check if the file system is undergoing power-on recovery
    if (unlikely(is_sbi_flag_set(sbi, SBI_POR_DOING)))
        goto redirty_out;

    // Get the old block address of the node page
    if (f2dfs_get_virtual_address_info(sbi, virtual_addr, &ni))
        goto redirty_out;

    // Check if the page has been truncated
    if (unlikely(ni.blk_addr == NULL_ADDR)) {
        ClearPageUptodate(page);
        dec_page_count(sbi, F2FS_DIRTY_DATA);
        unlock_page(page);
        return 0;
    }

    // Validate the block address
    if (__is_valid_data_blkaddr(ni.blk_addr) &&
        !f2fs_is_valid_blkaddr(sbi, ni.blk_addr, DATA_GENERIC_ENHANCE))
        goto redirty_out;

    // Set the page as writeback and clear the error flag
    set_page_writeback(page);
    ClearPageError(page);

    // Set the old block address and perform the write operation
    fio.old_blkaddr = ni.blk_addr;
    f2dfs_do_write_gc_page(&fio, virtual_addr, is_bggc);
    set_virtual_blkaddr(sbi, &ni, fio.new_blkaddr, 0, false, false);
    dec_page_count(sbi, F2FS_DIRTY_DATA);

    // Handle merged write for reclaim
    if (wbc->for_reclaim) {
        f2fs_submit_merged_write_cond(sbi, NULL, page, 0, DATA);
        submitted = NULL;
    }

    // Unlock the page
    unlock_page(page);

    // Check for file system corruption after write
    if (unlikely(f2fs_cp_error(sbi))) {
        f2fs_submit_merged_write(sbi, DATA);
        submitted = NULL;
    }
    if (submitted)
        *submitted = fio.submitted;
    return 0;

redirty_out:
    // Redirty the page for writepage
    redirty_page_for_writepage(wbc, page);
    return AOP_WRITEPAGE_ACTIVATE;
}

/**
 * Write a single garbage collection page.
 * 
 * @param page - The page to be written.
 * @param wbc - Writeback control structure.
 * @return - The result of the write operation.
 */
static int f2dfs_write_gc_page(struct page *page, struct writeback_control *wbc) {
    // Call the internal write function with specific parameters
    return __write_gc_page(page, NULL, wbc, FS_DATA_IO, le32_to_cpu(page->index), true);
}

/**
 * Synchronize garbage collection pages.
 * 
 * @param sbi - File system super block information.
 * @param wbc - Writeback control structure.
 * @param io_type - Type of I/O operation.
 * @return - The result of the synchronization operation.
 */
int f2dfs_sync_gc_pages(struct f2fs_sb_info *sbi, struct writeback_control *wbc, enum iostat_type io_type) {
    pgoff_t index;
    struct pagevec pvec;
    int step = 0;
    int nwritten = 0;
    int ret = 0;
    int nr_pages, done = 0;

    // Initialize the pagevec structure for page collection
    pagevec_init(&pvec);

    // Label for the next step in the loop
next_step:
    index = 0;

    // Loop through pages with the DIRTY tag
    while (!done && (nr_pages = pagevec_lookup_tag(&pvec, GC_MAPPING(sbi), &index, PAGECACHE_TAG_DIRTY))) {
        int i;

        for (i = 0; i < nr_pages; i++) {
            struct page *page = pvec.pages[i];
            bool submitted = false;

            // Prioritize WB_SYNC threads
            if (atomic_read(&sbi->wb_sync_req[DATA]) &&
                    wbc->sync_mode == WB_SYNC_NONE) {
                done = 1;
                break;
            }
            // Lock the page based on sync mode
            if (wbc->sync_mode == WB_SYNC_ALL)
                lock_page(page);
            else if (!trylock_page(page))
                continue;

            // Check if the page mapping is valid
            if (unlikely(page->mapping != GC_MAPPING(sbi))) {
                goto continue_unlock;
            }

            // If the page is not dirty, skip it
            if (!PageDirty(page)) {
                /* someone wrote it for us */
                goto continue_unlock;
            }

            // Wait for the page writeback to complete
            f2fs_wait_on_page_writeback(page, DATA, true, true);

            // Clear the page dirty flag if possible
            if (!clear_page_dirty_for_io(page))
                goto continue_unlock;

            // Write the page using the internal write function
            ret = __write_gc_page(page, &submitted, wbc, io_type, le32_to_cpu(page->index), true);
            if (ret)
                unlock_page(page);
            else if (submitted)
                nwritten++;

            // Break the loop if the writeback control limit is reached
            if (--wbc->nr_to_write == 0)
                break;
        }
        // Release the pagevec and allow for rescheduling
        pagevec_release(&pvec);
        cond_resched();

        // Break the loop if the writeback control limit is reached
        if (wbc->nr_to_write == 0) {
            step = 2;
            break;
        }
    }

    // Continue with the next step if necessary
    if (step < 2) {
        if (!is_sbi_flag_set(sbi, SBI_CP_DISABLED) &&
                wbc->sync_mode == WB_SYNC_NONE && step == 1)
            goto out;
        step++;
        goto next_step;
    }

out:
    // Submit merged write if necessary
    if (nwritten)
        f2fs_submit_merged_write(sbi, DATA);

    // Check for file system corruption
    if (unlikely(f2fs_cp_error(sbi)))
        return -EIO;
    return ret;
}
/**
 * Write garbage collection pages within a given address space.
 * 
 * @param mapping - The address space to write pages from.
 * @param wbc - Writeback control structure.
 * @return - 0 on success, or an error code.
 */
static int f2dfs_write_gc_pages(struct address_space *mapping, struct writeback_control *wbc) {
	struct f2fs_sb_info *sbi = F2FS_M_SB(mapping);
	struct blk_plug plug;
	long diff;

	// Skip writing if the file system is undergoing power-on recovery
	if (unlikely(is_sbi_flag_set(sbi, SBI_POR_DOING)))
		goto skip_write;

	/* TODO: balancing f2fs's metadata in background */
	/* TODO: fatal error here! TODO: */
	// f2fs_balance_fs_bg(sbi, true);

	// Skip writing if not all pages need to be synced and the number of dirty data pages is below a threshold
	if (wbc->sync_mode != WB_SYNC_ALL &&
		get_pages(sbi, F2FS_DIRTY_DATA) <
				nr_pages_to_skip(sbi, DATA))
		goto skip_write;

	// Increment the writeback request counter if syncing all pages
	if (wbc->sync_mode == WB_SYNC_ALL)
		atomic_inc(&sbi->wb_sync_req[DATA]);
	else if (atomic_read(&sbi->wb_sync_req[DATA]))
		goto skip_write;

	// Trace the write pages operation
	trace_f2fs_writepages(mapping->host, wbc, DATA);

	// Calculate the number of pages to write
	diff = nr_pages_to_write(sbi, DATA, wbc);
	// Start a block I/O plug to batch the write operations
	blk_start_plug(&plug);
	// Synchronize the garbage collection pages
	f2dfs_sync_gc_pages(sbi, wbc, FS_DATA_IO);
	// Finish the block I/O plug
	blk_finish_plug(&plug);
	// Update the number of pages left to write
	wbc->nr_to_write = max((long)0, wbc->nr_to_write - diff);

	// Decrement the writeback request counter if syncing all pages
	if (wbc->sync_mode == WB_SYNC_ALL)
		atomic_dec(&sbi->wb_sync_req[DATA]);
	return 0;

skip_write:
	// Increment the skipped pages counter
	wbc->pages_skipped += get_pages(sbi, F2FS_DIRTY_DATA);
	// Trace the skipped write pages operation
	trace_f2fs_writepages(mapping->host, wbc, DATA);
	return 0;
}

/**
 * Set a page as dirty for garbage collection.
 * 
 * @param page - The page to mark as dirty.
 * @return - 1 if the page was not previously dirty, 0 otherwise.
 */
static int f2dfs_set_gc_page_dirty(struct page *page) {
	// Trace the set page dirty operation
	trace_f2fs_set_page_dirty(page, DATA);

	// If the page is not up-to-date, mark it as such
	if (!PageUptodate(page))
		SetPageUptodate(page);
	// If the page is not already dirty, mark it as dirty and increment the dirty page count
	if (!PageDirty(page)) {
		__set_page_dirty_nobuffers(page);
		inc_page_count(F2FS_P_SB(page), F2FS_DIRTY_DATA);
		f2fs_set_page_private(page, 0);
		f2fs_trace_pid(page);
		return 1;
	}
	return 0;
}

/**
 * Define the address space operations for garbage collection.
 */
const struct address_space_operations f2dfs_gc_aops = {
	.writepage = f2dfs_write_gc_page,          // Write a single page
	.writepages = f2dfs_write_gc_pages,        // Write multiple pages
	.set_page_dirty = f2dfs_set_gc_page_dirty, // Mark a page as dirty
	.invalidatepage = f2fs_invalidate_page,   // Invalidate a page
	.releasepage = f2fs_release_page,         // Release a page
#ifdef CONFIG_MIGRATION
	.migratepage = f2fs_migrate_page,         // Migrate a page (if CONFIG_MIGRATION is enabled)
#endif
};

/**
 * Read a garbage collection page.
 * 
 * @param page - The page to read into.
 * @param virtual_addr - The virtual address of the page to read.
 * @param op_flags - Operation flags for the read operation.
 * @return - An error code or LOCKED_PAGE if the page is already up-to-date.
 */
static int read_gc_page(struct page *page, virtual_t virtual_addr, int op_flags) {
	struct f2fs_sb_info *sbi = F2FS_P_SB(page);
	block_t block_addr;
	struct f2fs_io_info fio = {
		.sbi = sbi,
		.ino = F2FS_GC_INO(sbi),
		.type = NODE,
		.op = REQ_OP_READ,
		.op_flags = op_flags,
		.page = page,
		.encrypted_page = NULL,
	};
	int err;

	// If the page is already up-to-date, return LOCKED_PAGE
	if (PageUptodate(page))
		return LOCKED_PAGE;

	// Translate the virtual address to a block address
	block_addr = f2dfs_address_translation(sbi, virtual_addr);
	// If the block address is invalid or the page index does not match, return -ENOENT
	if (!__is_valid_data_blkaddr(block_addr) ||
			le32_to_cpu(page->index) != virtual_addr)
		return -ENOENT;

	// If the block address is NULL_ADDR or the file system is shutting down, clear the page and return -ENOENT
	if (unlikely(block_addr == NULL_ADDR) ||
			is_sbi_flag_set(sbi, SBI_IS_SHUTDOWN)) {
		ClearPageUptodate(page);
		return -ENOENT;
	}

	// Set the new and old block addresses for the I/O operation
	fio.new_blkaddr = fio.old_blkaddr = block_addr;

	// Submit the page I/O request
	err = f2fs_submit_page_bio(&fio);
	// If the I/O operation is successful, update the I/O statistics
	if (!err)
		f2fs_update_iostat(sbi, FS_DATA_READ_IO, F2FS_BLKSIZE);

	return err;
}

/**
 * Perform read-ahead for a garbage collection page.
 * 
 * @param sbi - The file system super block information.
 * @param virtual_addr - The virtual address of the page to read ahead.
 */
void f2dfs_ra_gc_page(struct f2fs_sb_info *sbi, virtual_t virtual_addr) {
	struct page *apage;
	int err;

	// If the virtual address is 0 or out of range, return
	if (!virtual_addr)
		return;
	if (f2fs_check_nid_range(sbi, virtual_addr))
		return;

	// Try to load the page from the index tree
	apage = xa_load(&GC_MAPPING(sbi)->i_pages, virtual_addr);
	if (apage)
		return;

	// Grab a cache page for the given virtual address
	apage = f2fs_grab_cache_page(GC_MAPPING(sbi), virtual_addr, false);
	if (!apage)
		return;

	// Read the page with read-ahead operation flags
	err = read_gc_page(apage, virtual_addr, REQ_RAHEAD);
	// Release the page based on the read result
	f2fs_put_page(apage, err ? 1 : 0);
}

/**
 * Get a garbage collection page.
 * 
 * @param sbi - The file system super block information.
 * @param virtual_addr - The virtual address of the page to get.
 * @return - A pointer to the page or an error pointer.
 */
static struct page *__get_gc_page(struct f2fs_sb_info *sbi, virtual_t virtual_addr) {
	struct page *page;
	int err;

	// If the virtual address is 0, return an error pointer
	if (!virtual_addr)
		return ERR_PTR(-ENOENT);
	// If the virtual address is out of range, return an error pointer
	if (f2fs_check_nid_range(sbi, virtual_addr))
		return ERR_PTR(-EINVAL);

repeat:
	// Grab a cache page for the given virtual address
	page = f2fs_grab_cache_page(GC_MAPPING(sbi), virtual_addr, false);
	if (!page)
		return ERR_PTR(-ENOMEM);

	// Read the page
	err = read_gc_page(page, virtual_addr, 0);
	// If the read operation failed, release the page and return an error pointer
	if (err < 0) {
		f2fs_put_page(page, 1);
		return ERR_PTR(err);
	} else if (err == LOCKED_PAGE) {
		err = 0;
		goto page_hit;
	}

	// Lock the page
	lock_page(page);

	// If the page mapping is invalid, release the page and repeat the operation
	if (unlikely(page->mapping != GC_MAPPING(sbi))) {
		f2fs_put_page(page, 1);
		goto repeat;
	}

	// If the page is not up-to-date, return an error pointer
	if (unlikely(!PageUptodate(page))) {
		err = -EIO;
		goto out_err;
	}

page_hit:
	// Return the page
	return page;

out_err:
	// Clear the page up-to-date flag and return an error pointer
	ClearPageUptodate(page);
	f2fs_put_page(page, 1);
	return ERR_PTR(err);
}

struct page *f2dfs_get_gc_page(struct f2fs_sb_info *sbi, virtual_t vaddr)
{
	return __get_gc_page(sbi, vaddr);
}

int f2dfs_move_gc_page(struct page *data_page,
				int gc_type, virtual_t virtual_addr)
{
	int err = 0;

	if (gc_type == FG_GC) {
		struct writeback_control wbc = {
			.sync_mode = WB_SYNC_ALL,
			.nr_to_write = 1,
			.for_reclaim = 0,
		};

		f2fs_wait_on_page_writeback(data_page, DATA, true, true);
		set_page_dirty(data_page);

		if (!clear_page_dirty_for_io(data_page)) {
			err = -EAGAIN;
			goto out_page;
		}

		if (__write_gc_page(data_page, NULL, &wbc,
					FS_DATA_IO, virtual_addr, false)) {
			err = -EAGAIN;
			unlock_page(data_page);
		}

		goto release_page;
	} else {
		/* set page dirty and write it */
		if (!PageWriteback(data_page))
			set_page_dirty(data_page);
	}

out_page:
	unlock_page(data_page);

release_page:
	f2fs_put_page(data_page, 0);
	return err;
}

/* fp_inode pages aops */

void f2dfs_clear_fpnode_from_node_mapping(
				struct f2fs_sb_info *sbi, nid_t nid)
{
	struct page *page;

	if (!__is_valid_data_blkaddr(nid) ||
				f2fs_check_nid_range(sbi, nid))
		goto direct_out;

	page = f2fs_grab_cache_page(NODE_MAPPING(sbi), nid, false);
	if (!page)
		return;

	if ((page->index != nid) ||
				(page->mapping != NODE_MAPPING(sbi))) {
		goto not_clear;
	}

	if (is_inline_node(page))
		clear_inline_node(page);

	if (PageDirty(page)) {
		f2fs_clear_page_cache_dirty_tag(page);
		clear_page_dirty_for_io(page);
		dec_page_count(sbi, F2FS_DIRTY_NODES);
	}
	ClearPageUptodate(page);

not_clear:
	f2fs_put_page(page, 1);
	invalidate_mapping_pages(NODE_MAPPING(sbi), nid, nid);

direct_out:
	return;
}

/**
 * Perform the write operation for a full page (fp) node.
 * 
 * @param nid - The node ID of the page to write.
 * @param fio - The file I/O information structure.
 */
static inline void f2dfs_do_write_fp_page(unsigned int nid, struct f2fs_io_info *fio) {
    struct f2fs_summary sum;

    // Initialize the summary structure for the page write operation
    set_summary(&sum, nid, 0, 0);
    // Perform the page write operation
    do_write_page(&sum, fio);
    // Update the I/O statistics for the file system
    f2fs_update_iostat(fio->sbi, fio->io_type, F2FS_BLKSIZE);
    // Invalidate the mapping pages to prevent further access to this page
    invalidate_mapping_pages(FP_MAPPING(fio->sbi), nid, nid);
    // Clear the fp node from the node mapping to remove it from the cache
    f2dfs_clear_fpnode_from_node_mapping(fio->sbi, nid);
}

/**
 * Determine the temperature of a full page node.
 * 
 * @param page - The page to check.
 * @return - The temperature of the page, either HOT or WARM.
 */
static inline int fpnode_tempature(struct page *page) {
    nid_t nid = nid_of_node(page);

    // Check if the page has journal tags or is a head node, and return HOT
    if (get_fpfooter_tags(page, FP_NODE_JOURNAL_TAG) ||
            IS_NID_STABLE_HEAD(nid) || IS_NID_UNSTABLE_HEAD(nid) ||
            (FP_JOURNAL_HEAD_NID == nid))
        return HOT;
    // Otherwise, return WARM
    return WARM;
}

/**
 * Write a full page node to the storage.
 * 
 * @param page - The page to write.
 * @param atomic - Flag indicating if the operation is atomic.
 * @param submitted - A pointer to a flag indicating if the page was submitted for write.
 * @param wbc - The writeback control structure.
 * @param do_balance - Flag indicating if balance operation should be performed.
 * @param io_type - The type of I/O operation.
 * @return - An error code or 0 on success.
 */
static inline int __write_fp_page(struct page *page, bool atomic, bool *submitted,
                struct writeback_control *wbc, bool do_balance, enum iostat_type io_type) {
    struct f2fs_sb_info *sbi = F2FS_P_SB(page);
    nid_t nid;
    struct node_info ni;
    struct f2fs_io_info fio = {
        .sbi = sbi,
        .ino = ino_of_node(page),
        .type = NODE,
        .temp = fpnode_tempature(page),
        .op = REQ_OP_WRITE,
        .op_flags = wbc_to_write_flags(wbc),
        .page = page,
        .encrypted_page = NULL,
        .submitted = false,
        .io_type = io_type,
        .io_wbc = wbc,
    };

    // Trace the write page operation for debugging purposes
    trace_f2fs_writepage(page, NODE);

    // Check for file system corruption or closure, and handle it
    if (unlikely(f2fs_cp_error(sbi))) {
        if (is_sbi_flag_set(sbi, SBI_IS_CLOSE)) {
            ClearPageUptodate(page);
            dec_page_count(sbi, F2FS_DIRTY_NODES);
            unlock_page(page);
            return 0;
        }
        goto redirty_out;
    }

    // If the file system is undergoing power-on recovery, redirty the page and exit
    if (unlikely(is_sbi_flag_set(sbi, SBI_POR_DOING)))
        goto redirty_out;

    // Get the node ID and old block address of the node page
    nid = nid_of_node(page);
    f2fs_bug_on(sbi, page->index != nid);

    // Retrieve node information, and if it fails, redirty the page and exit
    if (f2fs_get_node_info(sbi, nid, &ni))
        goto redirty_out;

    // Acquire the node write lock if reclaiming, otherwise block until the lock is available
    if (wbc->for_reclaim) {
        if (!down_read_trylock(&sbi->node_write))
            goto redirty_out;
    } else {
        down_read(&sbi->node_write);
    }

    // If the page is already truncated, clear the page up-to-date flag and decrement the dirty node count
    if (unlikely(ni.blk_addr == NULL_ADDR)) {
        ClearPageUptodate(page);
        dec_page_count(sbi, F2FS_DIRTY_NODES);
        up_read(&sbi->node_write);
        unlock_page(page);
        return 0;
    }

    // Validate the block address, and if it's invalid, release the lock and redirty the page
    if (__is_valid_data_blkaddr(ni.blk_addr) &&
        !f2fs_is_valid_blkaddr(sbi, ni.blk_addr, DATA_GENERIC_ENHANCE)) {
        up_read(&sbi->node_write);
        goto redirty_out;
    }

    // If the operation is atomic and no barrier is set, add flush and write flags
    if (atomic && !test_opt(sbi, NOBARRIER))
        fio.op_flags |= REQ_PREFLUSH | REQ_FUA;

    // Set the page as writeback and clear the error flag
    set_page_writeback(page);
    ClearPageError(page);

    // Perform the write operation using the full page node write function
    fio.old_blkaddr = ni.blk_addr;
    f2dfs_do_write_fp_page(nid, &fio);
    // Update the node address to the new block address after write
    set_node_addr(sbi, &ni, fio.new_blkaddr, true);
    dec_page_count(sbi, F2FS_DIRTY_NODES);
    up_read(&sbi->node_write);

    // If reclaiming, submit a merged write conditionally
    if (wbc->for_reclaim) {
        f2fs_submit_merged_write_cond(sbi, NULL, page, 0, NODE);
        submitted = NULL;
    }

    // Unlock the page
    unlock_page(page);

    // Check for file system corruption after write, and if found, submit a merged write
    if (unlikely(f2fs_cp_error(sbi))) {
        f2fs_submit_merged_write(sbi, NODE);
        submitted = NULL;
    }
    // If the submitted flag is provided, set it to the submitted status of the I/O operation
    if (submitted)
        *submitted = fio.submitted;

    // If balance is required, perform file system balancing
    if (do_balance)
        f2fs_balance_fs(sbi, false);
    return 0;

redirty_out:
    // If an error occurs, redirty the page for writeback
    redirty_page_for_writepage(wbc, page);
    return AOP_WRITEPAGE_ACTIVATE;
}

/**
 * Write a full page (fp) node to storage.
 * 
 * @param page - The page to write.
 * @param wbc - Writeback control structure.
 * @return - The result of the write operation.
 */
static int f2dfs_write_fp_page(struct page *page, struct writeback_control *wbc) {
    // Call the internal write function with specific parameters
    return __write_fp_page(page, false, NULL, wbc, false, FS_NODE_IO);
}

/**
 * Synchronize full page (fp) nodes to storage.
 * 
 * @param sbi - File system super block information.
 * @param wbc - Writeback control structure.
 * @param do_balance - Flag indicating if balance operation should be performed.
 * @param io_type - Type of I/O operation.
 * @return - The result of the synchronization operation.
 */
int f2dfs_sync_fp_pages(struct f2fs_sb_info *sbi, struct writeback_control *wbc, bool do_balance, enum iostat_type io_type) {
    pgoff_t index;
    struct pagevec pvec;
    int step = 0;
    int nwritten = 0;
    int ret = 0;
    int nr_pages, done = 0;

    // Initialize the pagevec structure for page collection
    pagevec_init(&pvec);

    // Label for the next step in the loop
next_step:
    index = 0;

    // Loop through pages with the DIRTY tag
    while (!done && (nr_pages = pagevec_lookup_tag(&pvec, FP_MAPPING(sbi), &index, PAGECACHE_TAG_DIRTY))) {
        int i;

        for (i = 0; i < nr_pages; i++) {
            struct page *page = pvec.pages[i];
            bool submitted = false;

            // Prioritize WB_SYNC threads
            if (atomic_read(&sbi->wb_sync_req[NODE]) &&
                    wbc->sync_mode == WB_SYNC_NONE) {
                done = 1;
                break;
            }

            // Lock the page based on sync mode
            if (wbc->sync_mode == WB_SYNC_ALL)
                lock_page(page);
            else if (!trylock_page(page))
                continue;

            // Check if the page mapping is valid
            if (unlikely(page->mapping != FP_MAPPING(sbi))) {
                goto continue_unlock;
            }

            // If the page is not dirty, skip it
            if (!PageDirty(page)) {
                /* someone wrote it for us */
                goto continue_unlock;
            }

            // Wait for the page writeback to complete
            f2fs_wait_on_page_writeback(page, NODE, true, true);

            // Clear the page dirty flag if possible
            if (!clear_page_dirty_for_io(page))
                goto continue_unlock;

            // Set fsync and dentry marks to 0
            set_fsync_mark(page, 0);
            set_dentry_mark(page, 0);

            // Write the page using the internal write function
            ret = __write_fp_page(page, false, &submitted, wbc, do_balance, io_type);
            if (ret)
                unlock_page(page);
            else if (submitted)
                nwritten++;

            // Break the loop if the writeback control limit is reached
            if (--wbc->nr_to_write == 0)
                break;
        }
        // Release the pagevec and allow for rescheduling
        pagevec_release(&pvec);
        cond_resched();

        // Break the loop if the writeback control limit is reached
        if (wbc->nr_to_write == 0) {
            step = 2;
            break;
        }
    }

    // Continue with the next step if necessary
    if (step < 2) {
        if (!is_sbi_flag_set(sbi, SBI_CP_DISABLED) &&
                wbc->sync_mode == WB_SYNC_NONE && step == 1)
            goto out;
        step++;
        goto next_step;
    }

out:
    // Submit merged write if necessary
    if (nwritten)
        f2fs_submit_merged_write(sbi, NODE);

    // Check for file system corruption
    if (unlikely(f2fs_cp_error(sbi)))
        return -EIO;
    return ret;
}

/**
 * Write full page (fp) nodes within a given address space.
 * 
 * @param mapping - The address space to write from.
 * @param wbc - Writeback control structure.
 * @return - The result of the write operation.
 */
static int f2dfs_write_fp_pages(struct address_space *mapping, struct writeback_control *wbc) {
    struct f2fs_sb_info *sbi = F2FS_M_SB(mapping);
    struct blk_plug plug;
    long diff;

    // Skip writing if the file system is undergoing power-on recovery
    if (unlikely(is_sbi_flag_set(sbi, SBI_POR_DOING)))
        goto skip_write;

    /* TODO: Handle background metadata balancing */

    // Collect a number of dirty node pages and write them together
    if (wbc->sync_mode != WB_SYNC_ALL &&
            get_pages(sbi, F2FS_DIRTY_NODES) <
                    nr_pages_to_skip(sbi, NODE))
        goto skip_write;

    // Increment the writeback request counter if syncing all pages
    if (wbc->sync_mode == WB_SYNC_ALL)
        atomic_inc(&sbi->wb_sync_req[NODE]);
    else if (atomic_read(&sbi->wb_sync_req[NODE]))
        goto skip_write;

    // Trace the write pages operation
    trace_f2fs_writepages(mapping->host, wbc, NODE);

    // Calculate the number of pages to write
    diff = nr_pages_to_write(sbi, NODE, wbc);
    // Start a block I/O plug to batch the write operations
    blk_start_plug(&plug);
    // Synchronize the full page nodes
    f2dfs_sync_fp_pages(sbi, wbc, true, FS_NODE_IO);
    // Finish the block I/O plug
    blk_finish_plug(&plug);
    // Update the number of pages left to write
    wbc->nr_to_write = max((long)0, wbc->nr_to_write - diff);

    // Decrement the writeback request counter if syncing all pages
    if (wbc->sync_mode == WB_SYNC_ALL)
        atomic_dec(&sbi->wb_sync_req[NODE]);
    return 0;

skip_write:
    // Increment the skipped pages counter
    wbc->pages_skipped += get_pages(sbi, F2FS_DIRTY_NODES);
    // Trace the skipped write pages operation
    trace_f2fs_writepages(mapping->host, wbc, NODE);
    return 0;
}

/**
 * Mark a full page (fp) as dirty.
 * 
 * @param page - The page to mark as dirty.
 * @return - 1 if the page was not previously dirty, 0 otherwise.
 */
static int f2dfs_set_fp_page_dirty(struct page *page) {
    // Trace the set page dirty operation for debugging purposes
    trace_f2fs_set_page_dirty(page, NODE);

    // If the page is not up-to-date, mark it as such
    if (!PageUptodate(page))
        SetPageUptodate(page);
#ifdef CONFIG_F2FS_CHECK_FS
    // If this is an inode page, set the checksum
    if (IS_INODE(page))
        f2fs_inode_chksum_set(F2FS_P_SB(page), page);
#endif
    // If the page is not already dirty, mark it as dirty and increment the dirty node count
    if (!PageDirty(page)) {
        __set_page_dirty_nobuffers(page);
        inc_page_count(F2FS_P_SB(page), F2FS_DIRTY_NODES);
        f2fs_set_page_private(page, 0);
        f2fs_trace_pid(page);
        return 1;
    }
    return 0;
}

/**
 * Define the address space operations for full page (fp) nodes.
 */
const struct address_space_operations f2dfs_fp_aops = {
    .writepage = f2dfs_write_fp_page,          // Write a single page
    .writepages = f2dfs_write_fp_pages,        // Write multiple pages
    .set_page_dirty = f2dfs_set_fp_page_dirty, // Mark a page as dirty
    .invalidatepage = f2fs_invalidate_page,   // Invalidate a page
    .releasepage = f2fs_release_page,         // Release a page
#ifdef CONFIG_MIGRATION
    .migratepage = f2fs_migrate_page,         // Migrate a page (if CONFIG_MIGRATION is enabled)
#endif
};

/**
 * Clear the dirty flag for a full page (fp) node.
 * 
 * @param page - The page to clear the dirty flag for.
 */
void f2dfs_clear_fp_page_dirty(struct page *page) {
    // If the page is dirty, clear the dirty flag and decrement the dirty node count
    if (PageDirty(page)) {
        f2fs_clear_page_cache_dirty_tag(page);
        clear_page_dirty_for_io(page);
        dec_page_count(F2FS_P_SB(page), F2FS_DIRTY_NODES);
    }
    ClearPageUptodate(page);
}

/**
 * Read a full page (fp) node from storage.
 * 
 * @param page - The page to read into.
 * @param op_flags - Operation flags for the read operation.
 * @return - An error code or LOCKED_PAGE if the page is already up-to-date.
 */
static int read_fp_page(struct page *page, int op_flags) {
    struct f2fs_sb_info *sbi = F2FS_P_SB(page);
    struct node_info ni;
    struct f2fs_io_info fio = {
        .sbi = sbi,
        .type = NODE,
        .op = REQ_OP_READ,
        .op_flags = op_flags,
        .page = page,
        .encrypted_page = NULL,
    };
    int err;

    // If the page is already up-to-date, return LOCKED_PAGE
    if (PageUptodate(page))
        return LOCKED_PAGE;

    // Retrieve node information and return an error if not found
    err = f2fs_get_node_info(sbi, page->index, &ni);
    if (err)
        return err;

    // If the node address is NULL or the file system is shutting down, clear the page and return an error
    if (unlikely(ni.blk_addr == NULL_ADDR) ||
            is_sbi_flag_set(sbi, SBI_IS_SHUTDOWN)) {
        ClearPageUptodate(page);
        return -ENOENT;
    }

    // Set the block address for the I/O operation and submit the read request
    fio.new_blkaddr = fio.old_blkaddr = ni.blk_addr;
    err = f2fs_submit_page_bio(&fio);
    if (!err)
        f2fs_update_iostat(sbi, FS_NODE_READ_IO, F2FS_BLKSIZE);

    return err;
}
/**
 * Perform read-ahead for a full page (fp) node.
 * 
 * @param sbi - The file system super block information.
 * @param nid - The node ID to read ahead.
 */
void f2dfs_ra_fp_page(struct f2fs_sb_info *sbi, nid_t nid) {
    struct page *apage;
    int err;

    // If the node ID is 0 or out of range, return immediately
    if (!nid)
        return;
    if (f2fs_check_nid_range(sbi, nid))
        return;

    // Try to load the page from the index tree
    apage = xa_load(&FP_MAPPING(sbi)->i_pages, nid);
    if (apage)
        return;

    // Grab a cache page for the given node ID
    apage = f2fs_grab_cache_page(FP_MAPPING(sbi), nid, false);
    if (!apage)
        return;

    // Read the page with read-ahead operation flags
    err = read_fp_page(apage, REQ_RAHEAD);
    // Release the page based on the read result
    f2fs_put_page(apage, err ? 1 : 0);
}

/**
 * Get a full page (fp) node.
 * 
 * @param sbi - The file system super block information.
 * @param nid - The node ID to get.
 * @return - A pointer to the page or an error pointer.
 */
static struct page *__get_fp_page(struct f2fs_sb_info *sbi, nid_t nid) {
    struct page *page;
    int err;

    // If the node ID is 0, return an error pointer
    if (!nid)
        return ERR_PTR(-ENOENT);
    // If the node ID is out of range, return an error pointer
    if (f2fs_check_nid_range(sbi, nid))
        return ERR_PTR(-EINVAL);

repeat:
    // Grab a cache page for the given node ID
    page = f2fs_grab_cache_page(FP_MAPPING(sbi), nid, false);
    if (!page)
        return ERR_PTR(-ENOMEM);

    // Read the page
    err = read_fp_page(page, 0);
    // If the read operation failed, release the page and return an error pointer
    if (err < 0) {
        f2fs_put_page(page, 1);
        return ERR_PTR(err);
    } else if (err == LOCKED_PAGE) {
        err = 0;
        goto page_hit;
    }

    // Lock the page
    lock_page(page);

    // If the page mapping is invalid, release the page and repeat the operation
    if (unlikely(page->mapping != FP_MAPPING(sbi))) {
        f2fs_put_page(page, 1);
        goto repeat;
    }

    // If the page is not up-to-date, return an error pointer
    if (unlikely(!PageUptodate(page))) {
        err = -EIO;
        goto out_err;
    }

page_hit:
    // Check if the node ID is consistent with the page's node ID
    if(unlikely(nid != nid_of_node(page))) {
        f2fs_warn(sbi, "inconsistent fp node block, nid:%u, "
                "node_footer[nid:%u,ino:%u,ofs:%u,cpver:%llu,blkaddr:%u]",
                nid, nid_of_node(page), ino_of_node(page),
                ofs_of_node(page), cpver_of_node(page),
                next_blkaddr_of_node(page));
        err = -EINVAL;
    }

out_err:
    // If there's an error, clear the page up-to-date flag and return an error pointer
    ClearPageUptodate(page);
    f2fs_put_page(page, 1);
    return ERR_PTR(err);
}

/**
 * Retrieve a full page (fp) node.
 * 
 * @param sbi - The file system super block information.
 * @param nid - The node ID to retrieve.
 * @return - A pointer to the page.
 */
struct page *f2dfs_get_fp_page(struct f2fs_sb_info *sbi, nid_t nid) {
    return __get_fp_page(sbi, nid);
}

/**
 * Move a full page (fp) node.
 * 
 * @param fp_page - The page to move.
 * @param umount - Flag indicating if the file system is unmounting.
 * @return - An error code.
 */
int f2dfs_move_fp_page(struct page *fp_page, bool umount) {
    int err = 0;

    // If the file system is unmounting, perform the move operation
    if (umount) {
        struct writeback_control wbc = {
            .sync_mode = WB_SYNC_ALL,
            .nr_to_write = 1,
            .for_reclaim = 0,
        };

        // Wait for any ongoing writeback to complete
        f2fs_wait_on_page_writeback(fp_page, NODE, true, true);

        // Mark the page as dirty
        set_page_dirty(fp_page);

        // If the page is not clear for I/O, return an error
        if (!clear_page_dirty_for_io(fp_page)) {
            err = -EAGAIN;
            goto out_page;
        }

        // Write the page and handle any errors
        if (__write_fp_page(fp_page, false, NULL, &wbc, false, FS_NODE_IO)) {
            err = -EAGAIN;
            unlock_page(fp_page);
        }

        goto release_page;
    } else {
        // For other cases, set the page dirty and write it
        if (!PageWriteback(fp_page))
            set_page_dirty(fp_page);
    }

out_page:
    // Unlock the page
    unlock_page(fp_page);

release_page:
    // Release the page
    f2fs_put_page(fp_page, 0);
    return err;
}
struct page *f2dfs_create_new_fp_page(struct dnode_of_data *dn)
{
	struct f2fs_sb_info *sbi = F2FS_I_SB(dn->inode);
	struct node_info new_ni;
	struct page *page;
	int err;

	page = f2fs_grab_cache_page(FP_MAPPING(sbi), dn->nid, false);
	if (!page)
		return ERR_PTR(-ENOMEM);

	if (unlikely((err = inc_valid_node_count(sbi, dn->inode, true))))
		goto fail;

#ifdef CONFIG_F2FS_CHECK_FS
	err = f2fs_get_node_info(sbi, dn->nid, &new_ni);
	if (err) {
		dec_valid_node_count(sbi, dn->inode, true);
		goto fail;
	}
	f2fs_bug_on(sbi, new_ni.blk_addr != NULL_ADDR);
#endif

	new_ni.nid = dn->nid;
	new_ni.ino = dn->nid; /* treat as inode(nid=ino), but use fp_inode */
	new_ni.blk_addr = NULL_ADDR;
	new_ni.flag = 0;
	new_ni.version = 0;
	set_node_addr(sbi, &new_ni, NEW_ADDR, false);
	f2fs_wait_on_page_writeback(page, NODE, true, true);
	fill_node_footer(page, dn->nid, dn->nid, 0, true);

	if (!PageUptodate(page))
		SetPageUptodate(page);
	if (set_page_dirty(page))
		dn->node_changed = true;

	inc_valid_inode_count(sbi);
	return page;

fail:
	f2dfs_clear_fp_page_dirty(page);
	f2fs_put_page(page, 1);
	return ERR_PTR(err);
}

struct page *f2dfs_get_new_fp_page_head(
				struct f2fs_sb_info *sbi, nid_t fp_nid)
{
	struct page *page = NULL;
	struct dnode_of_data dn;

	// set_new_dnode(&dn, sbi->fp_inode, NULL, NULL, fp_nid);
	set_new_dnode(&dn, sbi->node_inode, NULL, NULL, fp_nid);
	page = f2dfs_create_new_fp_page(&dn);

	if (PTR_ERR(page) > 0)
		return NULL;
	return page;
}

struct page *f2dfs_get_new_fp_page(struct f2fs_sb_info *sbi)
{
	struct page *page = NULL;
	struct dnode_of_data dn;
	nid_t nid;

	if (!f2fs_alloc_nid(sbi, &nid))
		return ERR_PTR(-ENOSPC);

	// set_new_dnode(&dn, sbi->fp_inode, NULL, NULL, nid);
	set_new_dnode(&dn, sbi->node_inode, NULL, NULL, nid);
	page = f2dfs_create_new_fp_page(&dn);
	if (PTR_ERR(page) > 0) {
		f2fs_alloc_nid_failed(sbi, nid);
		return NULL;
	}
	return page;
}

/**
 * Truncate a full page (fp) node.
 * 
 * @param sbi - The file system super block information.
 * @param dn - The dnode of the data which contains the node page to truncate.
 * @return - An error code or 0 on success.
 */
static inline int truncate_fpnode(struct f2fs_sb_info *sbi, struct dnode_of_data *dn) {
    struct node_info ni;
    int err;
    pgoff_t index = dn->node_page->index;

    // Ensure the index matches the nid
    f2fs_bug_on(sbi, index != dn->nid);
    // Retrieve node information
    err = f2fs_get_node_info(sbi, dn->nid, &ni);
    // If there's an error or the block address is invalid, return the error
    if (err || (!__is_valid_data_blkaddr(ni.blk_addr)))
        return err;

    // Invalidate the blocks and decrement valid node and inode counts
    f2fs_invalidate_blocks(sbi, ni.blk_addr);
    dec_valid_node_count(sbi, dn->inode, true);
    set_node_addr(sbi, &ni, NULL_ADDR, false);
    dec_valid_inode_count(sbi);

    // Clear the fp page dirty and remove it from the node mapping
    f2dfs_clear_fp_page_dirty(dn->node_page);
    f2dfs_clear_fpnode_from_node_mapping(sbi, dn->nid);
    set_sbi_flag(sbi, SBI_IS_DIRTY);
    f2fs_put_page(dn->node_page, 1);
    invalidate_mapping_pages(FP_MAPPING(sbi), index, index);
    invalidate_mapping_pages(NODE_MAPPING(sbi), index, index);

    // Set the node page to NULL
    dn->node_page = NULL;
    // Trace the truncate node event
    trace_f2fs_truncate_node(dn->inode, dn->nid, ni.blk_addr);
    return 0;
}

/**
 * Truncate a full page (fp) page.
 * 
 * @param sbi - The file system super block information.
 * @param page - The page to truncate.
 * @param alloc - Flag indicating if this is an allocation operation.
 * @return - An error code or 0 on success.
 */
int f2dfs_truncate_fp_page(struct f2fs_sb_info *sbi, struct page *page, __u8 alloc) {
    struct dnode_of_data dn;
    // struct inode *inode = sbi->fp_inode;
    struct inode *inode = sbi->node_inode;
    int err = 0;

    // Get the nid from the page
    dn.nid = nid_of_node(page);
    // If this is an allocation operation, clear the page dirty and decrement counts
    if (alloc == 1) {
        f2dfs_clear_fp_page_dirty(page);
        f2fs_put_page(page, 1);
        dec_valid_node_count(sbi, inode, true);
        dec_valid_inode_count(sbi);
        f2fs_alloc_nid_failed(sbi, dn.nid);
        return err;
    }

    // Initialize the dnode and call truncate_fpnode to perform the truncation
    dn.inode = inode;
    dn.node_page = page;
    dn.ofs_in_node = 0;
    err = truncate_fpnode(sbi, &dn);
    return err;
}
